@@ -6,7 +6,7 @@
 
 /* #[ License : */
 /*
- *   Copyright (C) 1984-2013 J.A.M. Vermaseren
+ *   Copyright (C) 1984-2023 J.A.M. Vermaseren
  *   When using this file you are requested to refer to the publication
  *   J.A.M.Vermaseren "New features of FORM" math-ph/0010025
  *   This is considered a matter of courtesy as the development was paid
@@ -48,7 +48,7 @@
 		level = C->lhs[level][2];
 		goto SkipCount;
 	}
-	Note that there will be cases that extra space is needed.
+	Note that there will be cases in which extra space is needed.
 	In addition the compare with C->numlhs isn't very fine, because we
 	need to insert a different value (C->lhs[level][2]).
 */
@@ -65,8 +65,8 @@ WORD execarg(PHEAD WORD *term, WORD level)
 	LONG oldcpointer = CC->Pointer - CC->Buffer, oldppointer = AT.pWorkPointer, lp;
 	WORD *oldwork = AT.WorkPointer, *oldwork2, scale, renorm;
 	WORD kLCM = 0, kGCD = 0, kGCD2, kkLCM = 0, jLCM = 0, jGCD, sign = 1;
-	int ii;
-	UWORD *EAscrat, *GCDbuffer, *GCDbuffer2, *LCMbuffer, *LCMb, *LCMc;
+	int ii, didpolyratfun;
+	UWORD *EAscrat, *GCDbuffer = 0, *GCDbuffer2, *LCMbuffer, *LCMb, *LCMc;
 	AT.WorkPointer += *term;
 	start = C->lhs[level];
 	AR.Cnumlhs = start[2];
@@ -124,8 +124,9 @@ WORD execarg(PHEAD WORD *term, WORD level)
 /*
   	#[ Argument detection : + argument statement
 */
+	didpolyratfun = 0;
 	while ( t < rstop ) {
-		if ( *t >= FUNCTION && functions[*t-FUNCTION].spec == 0 ) {
+		if ( *t >= FUNCTION && functions[*t-FUNCTION].spec <= 0 ) {
 /*
 			We have a function. First count the number of arguments.
 			Tensors are excluded.
@@ -168,7 +169,7 @@ RightNum:
 							if ( *m == CSET ) {
 								r1 = SetElements + Sets[m[1]].first;
 								r2 = SetElements + Sets[m[1]].last;
-								while ( r1 <= r2 ) {
+								while ( r1 < r2 ) {
 									if ( *r1++ == *t ) goto HaveTodo;
 								}
 							}
@@ -185,6 +186,7 @@ HaveTodo:
 */
 				sign = 1;
 				action = 1;
+				if ( *t == AR.PolyFun ) didpolyratfun = 1;
 				v[2] |= DIRTYFLAG;
 				r = t + FUNHEAD;
 				j = i;
@@ -277,17 +279,20 @@ HaveTodo:
 				}
 				else if ( type == TYPENORM || type == TYPENORM2 || type == TYPENORM3 || type == TYPENORM4 ) {
 					if ( *r < 0 ) {
-						if ( *r != -SNUMBER || r[1] == 1 || r[1] == 0 ) continue;
+						WORD rone;
+						if ( *r == -MINVECTOR ) { rone = -1; *r = -INDEX; }
+						else if ( *r != -SNUMBER || r[1] == 1 || r[1] == 0 ) continue;
+						else { rone = r[1]; r[1] = 1; }
 /*
 						Now we must multiply the general coefficient by r[1]
 */
 						if ( scale && ( factor == 0 || *factor ) ) {
 							action = 1;
 							v[2] |= DIRTYFLAG;
-							if ( r[1] < 0 ) {
+							if ( rone < 0 ) {
 								if ( type == TYPENORM3 ) k = 1;
 								else k = -1;
-								r[1] = -r[1];
+								rone = -rone;
 							}
 							else k = 1;
 							r1 = term + *term;
@@ -295,13 +300,13 @@ HaveTodo:
 							size = REDLENG(size);
 							if ( scale > 0 ) {
 								for ( jj = 0; jj < scale; jj++ ) {
-									if ( Mully(BHEAD (UWORD *)rstop,&size,(UWORD *)(r+1),k) )
+									if ( Mully(BHEAD (UWORD *)rstop,&size,(UWORD *)(&rone),k) )
 										goto execargerr;
 								}
 							}
 							else {
 								for ( jj = 0; jj > scale; jj-- ) {
-									if ( Divvy(BHEAD (UWORD *)rstop,&size,(UWORD *)(r+1),k) )
+									if ( Divvy(BHEAD (UWORD *)rstop,&size,(UWORD *)(&rone),k) )
 										goto execargerr;
 								}
 							}
@@ -310,7 +315,6 @@ HaveTodo:
 							rstop[k-1] = size;
 							*term = (WORD)(rstop - term) + k;
 						}
-						r[1] = 1;
 						continue;
 					}
 /*
@@ -430,8 +434,10 @@ HaveTodo:
 								r3[jGCD+kLCM] = LCMbuffer[jGCD];
 							k = kLCM;
 						}
-						NumberFree(GCDbuffer,"execarg"); NumberFree(GCDbuffer2,"execarg");
-						NumberFree(LCMbuffer,"execarg"); NumberFree(LCMb,"execarg"); NumberFree(LCMc,"execarg");
+/*						NumberFree(GCDbuffer,"execarg"); GCDbuffer = 0; */
+						NumberFree(GCDbuffer2,"execarg");
+						NumberFree(LCMbuffer,"execarg");
+						NumberFree(LCMb,"execarg"); NumberFree(LCMc,"execarg");
 						j = 2*k+1;
 /*
 						Now we have to correct the overal factor
@@ -519,12 +525,12 @@ ScaledVariety:;
 					}
 /*
                   	We generate a statement for adapting all terms in the
-					argument sucessively
+					argument successively
 */
 					r4 = AddRHS(AT.ebufnum,1);
-					while ( (r4+j+12) > CC->Top ) r4 = DoubleCbuffer(AT.ebufnum,r4);
+					while ( (r4+j+12) > CC->Top ) r4 = DoubleCbuffer(AT.ebufnum,r4,3);
 					*r4++ = j+1;
-					i = (j-1)>>1;
+					i = (j-1)/2;  /* was (j-1)*2  ????? 17-oct-2017 */
 					for ( k = 0; k < i; k++ ) *r4++ = r3[i+k];
 					for ( k = 0; k < i; k++ ) *r4++ = r3[k];
 					if ( ( type == TYPENORM3 ) || ( type == TYPENORM4 ) ) *r4++ = j*sign;
@@ -535,6 +541,102 @@ ScaledVariety:;
 					AT.mulpat[5] = CC->numrhs;
 					AT.mulpat[7] = AT.ebufnum;
 				}
+				else if ( type == TYPEARGTOEXTRASYMBOL ) {
+					WORD n;
+					if ( r[0] < 0 ) {
+						/* The argument is in the fast notation. */
+						WORD tmp[MaX(9,FUNHEAD+5)];
+						switch ( r[0] ) {
+							case -SNUMBER:
+								if ( r[1] == 0 ) {
+									tmp[0] = 0;
+								}
+								else {
+									tmp[0] = 4;
+									tmp[1] = ABS(r[1]);
+									tmp[2] = 1;
+									tmp[3] = r[1] > 0 ? 3 : -3;
+									tmp[4] = 0;
+								}
+								break;
+							case -SYMBOL:
+								tmp[0] = 8;
+								tmp[1] = SYMBOL;
+								tmp[2] = 4;
+								tmp[3] = r[1];
+								tmp[4] = 1;
+								tmp[5] = 1;
+								tmp[6] = 1;
+								tmp[7] = 3;
+								tmp[8] = 0;
+								break;
+							case -INDEX:
+							case -VECTOR:
+							case -MINVECTOR:
+								tmp[0] = 7;
+								tmp[1] = INDEX;
+								tmp[2] = 3;
+								tmp[3] = r[1];
+								tmp[4] = 1;
+								tmp[5] = 1;
+								tmp[6] = r[0] != -MINVECTOR ? 3 : -3;
+								tmp[7] = 0;
+								break;
+							default:
+								if ( r[0] <= -FUNCTION ) {
+									tmp[0] = FUNHEAD+4;
+									tmp[1] = -r[0];
+									tmp[2] = FUNHEAD;
+									ZeroFillRange(tmp,3,1+FUNHEAD);
+									tmp[FUNHEAD+1] = 1;
+									tmp[FUNHEAD+2] = 1;
+									tmp[FUNHEAD+3] = 3;
+									tmp[FUNHEAD+4] = 0;
+									break;
+								}
+								else {
+									MLOCK(ErrorMessageLock);
+									MesPrint("Unknown fast notation found (TYPEARGTOEXTRASYMBOL)");
+									MUNLOCK(ErrorMessageLock);
+									return(-1);
+								}
+						}
+						n = FindSubexpression(tmp);
+					}
+					else {
+						/*
+						 * NOTE: writing to r[r[0]] is legal. As long as we work
+						 * in a part of the term, at least the coefficient of
+						 * the term must follow.
+						 */
+						WORD old_rr0 = r[r[0]];
+						r[r[0]] = 0;  /* zero-terminated */
+						n = FindSubexpression(r+ARGHEAD);
+						r[r[0]] = old_rr0;
+					}
+					/* Put the new argument in the work space. */
+					if ( AT.WorkPointer+2 > AT.WorkTop ) {
+						MLOCK(ErrorMessageLock);
+						MesWork();
+						MUNLOCK(ErrorMessageLock);
+						return(-1);
+					}
+					r1 = AT.WorkPointer;
+					if ( scale ) {  /* means "tonumber" */
+						r1[0] = -SNUMBER;
+						r1[1] = n;
+					}
+					else {
+						r1[0] = -SYMBOL;
+						r1[1] = MAXVARIABLES-n;
+					}
+					/* We need r2, r3, m and k to shift the data. */
+					r2 = r + (r[0] > 0 ? r[0] : r[0] <= -FUNCTION ? 1 : 2);
+					r3 = r;
+					m = r1+ARGHEAD+2;
+					k = 2;
+					goto do_shift;
+				}
 				r3 = r;
 				AR.DeferFlag = 0;
 				if ( *r > 0 ) {
@@ -574,7 +676,7 @@ ScaledVariety:;
 						if ( MultDo(BHEAD m,AT.mulpat) ) goto execargerr;
 						AT.WorkPointer = m + *m;
 					}
-					if ( Generator(BHEAD m,level) ) goto execargerr;
+					if ( (*m != 0 ) && Generator(BHEAD m,level) ) goto execargerr;
 					AT.WorkPointer = r1;
 				}
 				if ( EndSort(BHEAD AT.WorkPointer+ARGHEAD,1) < 0 ) goto execargerr;
@@ -591,7 +693,7 @@ ScaledVariety:;
 					if ( *AT.WorkPointer <= -FUNCTION ) k = 1;
 					else k = 2;
 				}
-				
+do_shift:
 				if ( *r3 > 0 ) j = k - *r3;
 				else if ( *r3 <= -FUNCTION ) j = k - 1;
 				else j = k - 2;
@@ -623,6 +725,10 @@ ScaledVariety:;
 		}
 		t += t[1];
 	}
+	if ( didpolyratfun ) {
+		PolyFunDirty(BHEAD term);
+		didpolyratfun = 0;
+	}
 /*
   	#] Argument detection : 
   	#[ SplitArg : + varieties
@@ -1031,20 +1137,37 @@ ScaledVariety:;
 */
 						t = r5;
 						EAscrat = (UWORD *)(TermMalloc("execarg"));
-						if ( t + *t == r3 ) goto onetermnew;
-						GETSTOP(t,r6);
-						ngcd = t[t[0]-1];
-						i = abs(ngcd)-1;
-						while ( --i >= 0 ) EAscrat[i] = r6[i];
-						t += *t;
-						while ( t < r3 ) {
+						if ( t + *t == r3 ) {
+							if ( factor == 0 || *factor > 2 ) {
+							if ( pow > 0 ) {
+								*r1++ = -SNUMBER; *r1++ = -1;
+								t = r5;
+								while ( t < r3 ) {
+									t += *t; t[-1] = -t[-1];
+								}
+							}
+							t = rr; *r1++ = *t++; *r1++ = 1; t++;
+							COPYARG(r1,t);
+							while ( t < m ) *r1++ = *t++;
+							}
+						}
+						else {
+						 GETSTOP(t,r6);
+						 ngcd = t[t[0]-1];
+						 i = abs(ngcd)-1;
+						 while ( --i >= 0 ) EAscrat[i] = r6[i];
+						 t += *t;
+						 while ( t < r3 ) {
 							GETSTOP(t,r6);
 							i = t[t[0]-1];
 							if ( AccumGCD(BHEAD EAscrat,&ngcd,(UWORD *)r6,i) ) goto execargerr;
 							if ( ngcd == 3 && EAscrat[0] == 1 && EAscrat[1] == 1 ) break;
 							t += *t;
-						}
-	 					if ( ngcd != 3 || EAscrat[0] != 1 || EAscrat[1] != 1 ) {
+						 }
+/*
+	 					 if ( ngcd != 3 || EAscrat[0] != 1 || EAscrat[1] != 1 ) {
+*/
+						 {
 							if ( pow ) ngcd = -ngcd;
 							t = r5; r9 = r1; *r1++ = t[-ARGHEAD]; *r1++ = 1;
 							FILLARG(r1); ngcd = REDLENG(ngcd);
@@ -1092,24 +1215,10 @@ ScaledVariety:;
 									if ( ToFast(r9,r9) ) r1 = r9+2;
 								}
 							}
-	 					}
+	 					 }
 /*
-			#] Numerical factor : 
+			#] Numerical factor :
 */
-						else {
-onetermnew:;
-							if ( factor == 0 || *factor > 2 ) {
-							if ( pow > 0 ) {
-								*r1++ = -SNUMBER; *r1++ = -1;
-								t = r5;
-								while ( t < r3 ) {
-									t += *t; t[-1] = -t[-1];
-								}
-							}
-							t = rr; *r1++ = *t++; *r1++ = 1; t++;
-							COPYARG(r1,t);
-							while ( t < m ) *r1++ = *t++;
-							}
 						}
 						TermFree(EAscrat,"execarg");
 					}
@@ -1598,13 +1707,14 @@ oneterm:;
 		if ( AT.WorkPointer < t ) AT.WorkPointer = t;
 	}
 /*
-  	#] FACTARG : 
+  	#] FACTARG :
 */
 	AR.Cnumlhs = oldnumlhs;
 	if ( action && Normalize(BHEAD term) ) goto execargerr;
 	AT.WorkPointer = oldwork;
 	if ( AT.WorkPointer < term + *term ) AT.WorkPointer = term + *term;
 	AT.pWorkPointer = oldppointer;
+	if ( GCDbuffer ) NumberFree(GCDbuffer,"execarg");
 	return(action);
 execargerr:
 	AT.WorkPointer = oldwork;
@@ -1616,7 +1726,7 @@ execargerr:
 }
 
 /*
-  	#] execarg : 
+  	#] execarg :
   	#[ execterm :
 */
 
@@ -1629,11 +1739,23 @@ WORD execterm(PHEAD WORD *term, WORD level)
 	WORD *buffer1 = 0;
 	WORD *oldworkpointer = AT.WorkPointer;
 	WORD *t1, i;
-	WORD olddeferflag = AR.DeferFlag;
+	WORD olddeferflag = AR.DeferFlag, tryterm = 0;
 	AR.DeferFlag = 0;
 	do {
 		AR.Cnumlhs = C->lhs[level][3];
 		NewSort(BHEAD0);
+/*
+		Normally for function arguments we do not use PolyFun/PolyRatFun.
+		Hence NewSort sets the corresponding variables to zero.
+		Here we overwrite that.
+*/
+		AN.FunSorts[AR.sLevel]->PolyFlag = ( AR.PolyFun != 0 ) ? AR.PolyFunType: 0;
+		if ( AR.PolyFun == 0 ) { AN.FunSorts[AR.sLevel]->PolyFlag = 0; }
+		else if ( AR.PolyFunType == 1 ) { AN.FunSorts[AR.sLevel]->PolyFlag = 1; }
+		else if ( AR.PolyFunType == 2 ) {
+			if ( AR.PolyFunExp == 2 ) AN.FunSorts[AR.sLevel]->PolyFlag = 1;
+			else                      AN.FunSorts[AR.sLevel]->PolyFlag = 2;
+		}
 		if ( buffer1 ) {
 			term = buffer1;
 			while ( *term ) {
@@ -1647,10 +1769,13 @@ WORD execterm(PHEAD WORD *term, WORD level)
 			if ( Generator(BHEAD term,level) ) goto exectermerr;
 		}
 		if ( buffer1 ) {
-			M_free((void *)buffer1,"buffer in sort statement");
+			if ( tryterm ) { TermFree(buffer1,"buffer in sort statement"); tryterm = 0; }
+			else { M_free((void *)buffer1,"buffer in sort statement"); }
 			buffer1 = 0;
 		}
+		AN.tryterm = 1;
 		if ( EndSort(BHEAD (WORD *)((VOID *)(&buffer1)),2) < 0 ) goto exectermerr;
+		tryterm = AN.tryterm; AN.tryterm = 0;
 		level = AR.Cnumlhs;
 	} while ( AR.Cnumlhs < maxisat );
 	AR.Cnumlhs = oldnumlhs;
@@ -1662,7 +1787,9 @@ WORD execterm(PHEAD WORD *term, WORD level)
 		AT.WorkPointer = t1;
 		if ( Generator(BHEAD oldworkpointer,level) ) goto exectermerr;
 	}
-	M_free(buffer1,"buffer in term statement");
+	if ( tryterm ) { TermFree(buffer1,"buffer in term statement"); tryterm = 0; }
+	else { M_free(buffer1,"buffer in term statement"); }
+	buffer1 = 0;
 	AT.WorkPointer = oldworkpointer;
 	return(0);
 exectermerr:
@@ -2381,11 +2508,11 @@ WORD InsertArg(PHEAD WORD *argin, WORD *argout,int par)
 	}
 	else { return(-1); }
 	AddRHS(bufnum,1);
-	AddNtoC(bufnum,*argin,argin);
+	AddNtoC(bufnum,*argin,argin,1);
 	AddToCB(C,0)
 	a = argout; while ( *a ) NEXTARG(a);
 	i = a - argout;
-	AddNtoC(bufnum,i,argout);
+	AddNtoC(bufnum,i,argout,2);
 	AddToCB(C,0)
 	return(InsTree(bufnum,C->numrhs));
 }
@@ -3412,4 +3539,3 @@ void SortWeights(LONG *weights,LONG *extraspace,WORD number)
 /*
   	#] SortWeights : 
 */
-
