@@ -53,7 +53,7 @@
   	#] Explanations : 
   	#[ License :
  *
- *   Copyright (C) 1984-2013 J.A.M. Vermaseren
+ *   Copyright (C) 1984-2023 J.A.M. Vermaseren
  *   When using this file you are requested to refer to the publication
  *   J.A.M.Vermaseren "New features of FORM" math-ph/0010025
  *   This is considered a matter of courtesy as the development was paid
@@ -757,6 +757,7 @@ static void print_M()
 	MesPrint("%d", AM.hparallelflag);
 	MesPrint("%d", AM.gparallelflag);
 	MesPrint("%d", AM.totalnumberofthreads);
+	MesPrint("%d", AM.gSizeCommuteInSet);
 	MesPrint("%d", AM.gThreadStats);
 	MesPrint("%d", AM.ggThreadStats);
 	MesPrint("%d", AM.gFinalStats);
@@ -771,6 +772,8 @@ static void print_M()
 	MesPrint("%d", AM.ggProcessStats);
 	MesPrint("%d", AM.gOldParallelStats);
 	MesPrint("%d", AM.ggOldParallelStats);
+	MesPrint("%d", AM.gWTimeStatsFlag);
+	MesPrint("%d", AM.ggWTimeStatsFlag);
 	MesPrint("%d", AM.maxFlevels);
 	MesPrint("--MARK  6");
 	MesPrint("%d", AM.resetTimeOnClear);
@@ -821,7 +824,11 @@ static void print_M()
 	MesPrint("%d", AM.matchfunnum);
 	MesPrint("%d", AM.countfunnum);
 	MesPrint("%d", AM.gPolyFun);
+	MesPrint("%d", AM.gPolyFunInv);
 	MesPrint("%d", AM.gPolyFunType);
+	MesPrint("%d", AM.gPolyFunExp);
+	MesPrint("%d", AM.gPolyFunVar);
+	MesPrint("%d", AM.gPolyFunPow);
 	MesPrint("--MARK 10");
 	MesPrint("%d", AM.dollarzero);
 	MesPrint("%d", AM.atstartup);
@@ -831,6 +838,8 @@ static void print_M()
 	MesPrint("%d", AM.ggIndentSpace);
 	MesPrint("%d", AM.gShortStatsMax);
 	MesPrint("%d", AM.ggShortStatsMax);
+	MesPrint("--MARK 11");
+	MesPrint("%d", AM.FromStdin);
 	MesPrint("%%%% END M_const");
 /*	fflush(0); */
 }
@@ -1060,6 +1069,7 @@ static void print_C()
 	MesPrint("%d", AC.ThreadSortFileSynch);
 	MesPrint("%d", AC.ProcessStats);
 	MesPrint("%d", AC.OldParallelStats);
+	MesPrint("%d", AC.WTimeStatsFlag);
 	MesPrint("%d", AC.BracketNormalize);
 	MesPrint("%d", AC.maxtermlevel);
 	MesPrint("%d", AC.dumnumflag);
@@ -1115,7 +1125,11 @@ static void print_C()
 	MesPrint("%d", AC.StoreHandle);
 	MesPrint("%d", AC.HideLevel);
 	MesPrint("%d", AC.lPolyFun);
+	MesPrint("%d", AC.lPolyFunInv);
 	MesPrint("%d", AC.lPolyFunType);
+	MesPrint("%d", AC.lPolyFunExp);
+	MesPrint("%d", AC.lPolyFunVar);
+	MesPrint("%d", AC.lPolyFunPow);
 	MesPrint("%d", AC.SymChangeFlag);
 	MesPrint("%d", AC.CollectPercentage);
 	MesPrint("%d", AC.ShortStatsMax);
@@ -1172,7 +1186,11 @@ static void print_R()
 	MesPrint("%d", AR.Stage4Name);
 	MesPrint("%d", AR.GetOneFile);
 	MesPrint("%d", AR.PolyFun);
+	MesPrint("%d", AR.PolyFunInv);
 	MesPrint("%d", AR.PolyFunType);
+	MesPrint("%d", AR.PolyFunExp);
+	MesPrint("%d", AR.PolyFunVar);
+	MesPrint("%d", AR.PolyFunPow);
 	MesPrint("%d", AR.Eside);
 	MesPrint("%d", AR.MaxDum);
 	MesPrint("%d", AR.level);
@@ -1437,13 +1455,18 @@ int DoRecovery(int *moduletype)
 	R_SET(AM.gOutNumberType, WORD);
 	R_SET(AM.gfunpowers, int);
 	R_SET(AM.gPolyFun, WORD);
+	R_SET(AM.gPolyFunInv, WORD);
 	R_SET(AM.gPolyFunType, WORD);
+	R_SET(AM.gPolyFunExp, WORD);
+	R_SET(AM.gPolyFunVar, WORD);
+	R_SET(AM.gPolyFunPow, WORD);
 	R_SET(AM.gProcessBucketSize, LONG);
 	R_SET(AM.OldChildTime, LONG);
 	R_SET(AM.OldSecTime, LONG);
 	R_SET(AM.OldMilliTime, LONG);
 	R_SET(AM.gproperorderflag, int);
 	R_SET(AM.gThreadBucketSize, LONG);
+	R_SET(AM.gSizeCommuteInSet, int);
 	R_SET(AM.gThreadStats, int);
 	R_SET(AM.gFinalStats, int);
 	R_SET(AM.gThreadsFlag, int);
@@ -1467,6 +1490,20 @@ int DoRecovery(int *moduletype)
 	R_SET(AM.gOldFactArgFlag,int);
 	R_SET(AM.ggOldFactArgFlag,int);
 
+    R_SET(AM.gnumextrasym,int);
+    R_SET(AM.ggnumextrasym,int);
+	R_SET(AM.NumSpectatorFiles,int);
+	R_SET(AM.SizeForSpectatorFiles,int);
+    R_SET(AM.gOldGCDflag,int);
+    R_SET(AM.ggOldGCDflag,int);
+	R_SET(AM.gWTimeStatsFlag, int);
+
+	R_FREE(AM.Path);
+	R_SET(AM.Path,UBYTE *);
+	R_COPY_S(AM.Path,UBYTE *);
+
+	R_SET(AM.FromStdin, BOOL);
+
 #ifdef PRINTDEBUG
 	print_M();
 #endif
@@ -1587,6 +1624,7 @@ int DoRecovery(int *moduletype)
 	R_FREE(AC.inputnumbers);
 #endif
 	R_FREE(AC.IfSumCheck);
+	R_FREE(AC.CommuteInSet);
 	R_FREE(AC.CheckpointRunAfter);
 	R_FREE(AC.CheckpointRunBefore);
 
@@ -1980,6 +2018,9 @@ int DoRecovery(int *moduletype)
 	if ( AC.IfSumCheck ) {
 		R_COPY_B(AC.IfSumCheck, (LONG)sizeof(WORD)*(AC.MaxIf+1), WORD*);
 	}
+	if ( AC.CommuteInSet ) {
+		R_COPY_B(AC.CommuteInSet, (LONG)sizeof(WORD)*(AC.SizeCommuteInSet+1), WORD*);
+	}
 
 	AC.LogHandle = oldLogHandle;
 
@@ -2125,6 +2166,7 @@ int DoRecovery(int *moduletype)
 	R_COPY_S(AP.procedureExtension,UBYTE*);
 	R_COPY_S(AP.cprocedureExtension,UBYTE*);
 
+	R_COPY_B(AP.PreAssignStack,AP.MaxPreAssignLevel*(LONG)sizeof(LONG),LONG*);
 	R_COPY_B(AP.PreIfStack, AP.MaxPreIfLevel*(LONG)sizeof(int), int*);
 	R_COPY_B(AP.PreSwitchModes, (AP.NumPreSwitchStrings+1)*(LONG)sizeof(int), int*);
 	R_COPY_B(AP.PreTypes, (AP.MaxPreTypes+1)*(LONG)sizeof(int), int*);
@@ -2287,7 +2329,11 @@ int DoRecovery(int *moduletype)
 	R_SET(AR.Stage4Name, WORD);
 	R_SET(AR.GetOneFile, WORD);
 	R_SET(AR.PolyFun, WORD);
+	R_SET(AR.PolyFunInv, WORD);
 	R_SET(AR.PolyFunType, WORD);
+	R_SET(AR.PolyFunExp, WORD);
+	R_SET(AR.PolyFunVar, WORD);
+	R_SET(AR.PolyFunPow, WORD);
 	R_SET(AR.Eside, WORD);
 	R_SET(AR.MaxDum, WORD);
 	R_SET(AR.level, WORD);
@@ -2297,9 +2343,9 @@ int DoRecovery(int *moduletype)
 	R_SET(AR.SortType, WORD);
 	R_SET(AR.ShortSortCount, WORD);
 
-	/* this is usually done in Process(), but sometimes FORM doesn't
+	/* this is usually done in Process(), but sometimes FORM does not
 	   end up executing Process() before it uses the AR.CompressPointer,
-	   so we need to explicitely set it here. */
+	   so we need to explicitly set it here. */
 	AR.CompressPointer = AR.CompressBuffer;
 
 #ifdef WITHPTHREADS
@@ -2345,7 +2391,19 @@ int DoRecovery(int *moduletype)
 		R_COPY_B(A.O.OptimizeResult.code,A.O.OptimizeResult.codesize*sizeof(WORD),WORD *);
 	}
 	R_COPY_S(A.O.OptimizeResult.nameofexpr,UBYTE *);
-
+/*
+	And now the dictionaries. We know how many there are. We also know
+	how many elements the array AO.Dictionaries should have.
+*/
+	if ( AO.SizeDictionaries > 0 ) {
+		AO.Dictionaries = (DICTIONARY **)Malloc1(AO.SizeDictionaries*sizeof(DICTIONARY *),
+					"Dictionaries");
+		for ( i = 0; i < AO.NumDictionaries; i++ ) {
+			R_SET(l,LONG)
+			AO.Dictionaries[i] = DictFromBytes(p);
+			p = (char *)p + l;
+		}
+	}
 	/*#] AO :*/ 
 #ifdef WITHMPI
 	/*#[ PF : */
@@ -2469,13 +2527,18 @@ static int DoSnapshot(int moduletype)
 	S_WRITE_B(&AM.gOutNumberType, sizeof(WORD));
 	S_WRITE_B(&AM.gfunpowers, sizeof(int));
 	S_WRITE_B(&AM.gPolyFun, sizeof(WORD));
+	S_WRITE_B(&AM.gPolyFunInv, sizeof(WORD));
 	S_WRITE_B(&AM.gPolyFunType, sizeof(WORD));
+	S_WRITE_B(&AM.gPolyFunExp, sizeof(WORD));
+	S_WRITE_B(&AM.gPolyFunVar, sizeof(WORD));
+	S_WRITE_B(&AM.gPolyFunPow, sizeof(WORD));
 	S_WRITE_B(&AM.gProcessBucketSize, sizeof(LONG));
 	S_WRITE_B(&AM.OldChildTime, sizeof(LONG));
 	S_WRITE_B(&AM.OldSecTime, sizeof(LONG));
 	S_WRITE_B(&AM.OldMilliTime, sizeof(LONG));
 	S_WRITE_B(&AM.gproperorderflag, sizeof(int));
 	S_WRITE_B(&AM.gThreadBucketSize, sizeof(LONG));
+	S_WRITE_B(&AM.gSizeCommuteInSet, sizeof(int));
 	S_WRITE_B(&AM.gThreadStats, sizeof(int));
 	S_WRITE_B(&AM.gFinalStats, sizeof(int));
 	S_WRITE_B(&AM.gThreadsFlag, sizeof(int));
@@ -2499,6 +2562,19 @@ static int DoSnapshot(int moduletype)
 	S_WRITE_B(&AM.gOldFactArgFlag,sizeof(int));
 	S_WRITE_B(&AM.ggOldFactArgFlag,sizeof(int));
 
+    S_WRITE_B(&AM.gnumextrasym,sizeof(int));
+    S_WRITE_B(&AM.ggnumextrasym,sizeof(int));
+	S_WRITE_B(&AM.NumSpectatorFiles,sizeof(int));
+	S_WRITE_B(&AM.SizeForSpectatorFiles,sizeof(int));
+    S_WRITE_B(&AM.gOldGCDflag,sizeof(int));
+    S_WRITE_B(&AM.ggOldGCDflag,sizeof(int));
+	S_WRITE_B(&AM.gWTimeStatsFlag, sizeof(int));
+
+	S_WRITE_B(&AM.Path,sizeof(UBYTE *));
+	S_WRITE_S(AM.Path);
+
+	S_WRITE_B(&AM.FromStdin,sizeof(BOOL));
+
 	/*#] AM :*/ 
 	/*#[ AC :*/
 
@@ -2726,6 +2802,9 @@ static int DoSnapshot(int moduletype)
 	if ( AC.IfSumCheck ) {
 		S_WRITE_B(AC.IfSumCheck, (LONG)sizeof(WORD)*(AC.MaxIf+1));
 	}
+	if ( AC.CommuteInSet ) {
+		S_WRITE_B(AC.CommuteInSet, (LONG)sizeof(WORD)*(AC.SizeCommuteInSet+1));
+	}
 
 	S_WRITE_S(AC.CheckpointRunAfter);
 	S_WRITE_S(AC.CheckpointRunBefore);
@@ -2812,6 +2891,7 @@ static int DoSnapshot(int moduletype)
 	S_WRITE_S(AP.procedureExtension);
 	S_WRITE_S(AP.cprocedureExtension);
 
+	S_WRITE_B(AP.PreAssignStack, AP.MaxPreAssignLevel*(LONG)sizeof(LONG));
 	S_WRITE_B(AP.PreIfStack, AP.MaxPreIfLevel*(LONG)sizeof(int));
 	S_WRITE_B(AP.PreSwitchModes, (AP.NumPreSwitchStrings+1)*(LONG)sizeof(int));
 	S_WRITE_B(AP.PreTypes, (AP.MaxPreTypes+1)*(LONG)sizeof(int));
@@ -2866,7 +2946,11 @@ static int DoSnapshot(int moduletype)
 	S_WRITE_B(&AR.Stage4Name, sizeof(WORD));
 	S_WRITE_B(&AR.GetOneFile, sizeof(WORD));
 	S_WRITE_B(&AR.PolyFun, sizeof(WORD));
+	S_WRITE_B(&AR.PolyFunInv, sizeof(WORD));
 	S_WRITE_B(&AR.PolyFunType, sizeof(WORD));
+	S_WRITE_B(&AR.PolyFunExp, sizeof(WORD));
+	S_WRITE_B(&AR.PolyFunVar, sizeof(WORD));
+	S_WRITE_B(&AR.PolyFunPow, sizeof(WORD));
 	S_WRITE_B(&AR.Eside, sizeof(WORD));
 	S_WRITE_B(&AR.MaxDum, sizeof(WORD));
 	S_WRITE_B(&AR.level, sizeof(WORD));
@@ -2914,6 +2998,16 @@ static int DoSnapshot(int moduletype)
 		S_WRITE_B(A.O.OptimizeResult.code,A.O.OptimizeResult.codesize*sizeof(WORD));
 	}
 	S_WRITE_S(A.O.OptimizeResult.nameofexpr);
+/*
+	And now the dictionaries.
+	We write each dictionary to a buffer and get the size of that buffer.
+	Then we write the size and the buffer.
+*/
+	for ( i = 0; i < AO.NumDictionaries; i++ ) {
+		l = DictToBytes(AO.Dictionaries[i],(UBYTE *)(AT.WorkPointer));
+		S_WRITE_B(&l,sizeof(LONG));
+		S_WRITE_B(AT.WorkPointer,l);
+	}
 
 	/*#] AO :*/ 
 	/*#[ PF :*/
