@@ -7,7 +7,7 @@
 
 /* #[ License : */
 /*
- *   Copyright (C) 1984-2013 J.A.M. Vermaseren
+ *   Copyright (C) 1984-2023 J.A.M. Vermaseren
  *   When using this file you are requested to refer to the publication
  *   J.A.M.Vermaseren "New features of FORM" math-ph/0010025
  *   This is considered a matter of courtesy as the development was paid
@@ -55,6 +55,7 @@ static struct id_options {
 	,{(UBYTE *)"ifnotmatch", SUBAFTERNOT ,0}
 	,{(UBYTE *)"disorder",   SUBDISORDER ,0}
 	,{(UBYTE *)"select",     SUBSELECT   ,0}
+	,{(UBYTE *)"all",        SUBALL      ,0}
 };
 
 /*
@@ -138,6 +139,10 @@ int DoExpr(UBYTE *inp, int type, int par)
 								*w = DROPLEXPRESSION;
 							else if ( *w == GLOBALEXPRESSION || *w == SKIPGEXPRESSION )
 								*w = DROPGEXPRESSION;
+							else if ( *w == HIDDENLEXPRESSION )
+								*w = DROPHLEXPRESSION;
+							else if ( *w == HIDDENGEXPRESSION )
+								*w = DROPHGEXPRESSION;
 						}
 						AC.TransEname = Expressions[c2].name;
 						j = EntVar(CEXPRESSION,0,type,0,0,0);
@@ -214,6 +219,13 @@ int DoExpr(UBYTE *inp, int type, int par)
 				}
 				AC.ProtoType[1] = w - AC.ProtoType;
 			}
+			else if ( c != '=' ) {
+/*
+				The dummy accepted L F := RHS;
+*/
+				MesPrint("&Illegal LHS for expression definition");
+				error = 1;
+			}
 			*w++ = 1;
 			*w++ = 1;
 			*w++ = 3;
@@ -236,6 +248,7 @@ int DoExpr(UBYTE *inp, int type, int par)
 		inp = p+1;
 		ClearWildcardNames();
 		osize = AC.ProtoType[1]; AC.ProtoType[1] = SUBEXPSIZE;
+		PutInVflags(jold);
 		if ( ( i = CompileAlgebra(inp,RHSIDE,AC.ProtoType) ) < 0 ) {
 			AC.ProtoType[1] = osize;
 			error = 1;
@@ -243,12 +256,12 @@ int DoExpr(UBYTE *inp, int type, int par)
 		else if ( error == 0 ) {
 			AC.ProtoType[1] = osize;
 			AC.ProtoType[2] = i;
-/*			AR.outfile->POfill = AR.outfile->POfull; */
 			if ( PutOut(BHEAD OldWork+2,&pos,AR.outfile,0) < 0 ) {
 				MesPrint("&Cannot create expression");
 				error = -1;
 			}
 			else {
+				Expressions[j].sizeprototype = OldWork[2];
 				OldWork[2] = 4+SUBEXPSIZE;
 				OldWork[4] = SUBEXPSIZE;
 				OldWork[5] = i;
@@ -261,22 +274,17 @@ int DoExpr(UBYTE *inp, int type, int par)
 					MesPrint("&Cannot create expression");
 					error = -1;
 				}
-/*
-				The next statement repairs a bug that occurs with
-				L  F2 = .....
-				#$d = F1[x];
-				in which the POfull was less than POfill and the setting
-				of the Scratch file in store.c goes wrong.
-				15-oct-2006 JV
-*/
 				AR.outfile->POfull = AR.outfile->POfill;
 			}
 			OldWork[2] = j;
+/*
+			Seems unnecessary (13-feb-2018)
+
 			AddNtoL(OldWork[1],OldWork);
+*/
 			AT.WorkPointer = OldWork;
 			if ( AC.dumnumflag ) Add2Com(TYPEDETCURDUM)
 		}
-		PutInVflags(jold);
 		AC.ToBeInFactors = 0;
 	}
 	else {	/* Variety in which expressions change property */
@@ -319,6 +327,9 @@ int DoExpr(UBYTE *inp, int type, int par)
 						case INTOHIDEGEXPRESSION:
 							*w = INTOHIDELEXPRESSION;
 							break;
+						case DROPHGEXPRESSION:
+							*w = DROPHLEXPRESSION;
+							break;
 					}
 				}
 				else if ( type == GLOBALEXPRESSION ) {
@@ -344,6 +355,9 @@ int DoExpr(UBYTE *inp, int type, int par)
 						case INTOHIDELEXPRESSION:
 							*w = INTOHIDEGEXPRESSION;
 							break;
+						case DROPHLEXPRESSION:
+							*w = DROPHGEXPRESSION;
+							break;
 					}
 				}
 /*
@@ -498,7 +512,7 @@ int CoIdExpression(UBYTE *inp, int type)
 		 oldnumrhs, *ow, oldEside;
 	UBYTE *p, *pp, c;
 	CBUF *C = cbuf+AC.cbufnum;
-	LONG oldcpointer;
+	LONG oldcpointer, x;
 	FirstWork = OldWork = AT.WorkPointer;
 /*
 	Don't forget to change in StudyPattern if we change/add_to the
@@ -671,6 +685,56 @@ readlabel:
 				OldWork[3] = GetLabel(inp);
 				*p++ = c; inp = p;
 				break;
+			case SUBALL:
+				x = 0;
+				if ( *pp == '(' ) {
+					if ( FG.cTable[*inp] == 1 ) {
+						while ( *inp >= '0' && *inp <= '9' ) x = 10*x+*inp++ - '0';
+					}
+					else {
+						pp++;
+						while ( FG.cTable[*inp] == 0 ) inp++;
+						c = *inp; *inp = 0;
+						if ( StrICont(pp,(UBYTE *)"normalize") != 0 ) goto IllOpt;
+						*inp = c;
+						OldWork[4] |= NORMALIZEFLAG;
+					}
+					if ( *inp != ')' || inp+1 != p ) {
+						c = *inp; *inp = 0;
+IllOpt:
+						MesPrint("&Illegal ALL option in id-statement: ",pp);
+						*inp++ = c;
+						error = 1;
+						continue;
+					}
+					pp = inp;
+					inp = pp+1;
+				}
+/*
+				Note that the following statement limits x to 
+*/
+				if ( x > MAXPOSITIVE ) {
+					MesPrint("&Requested maximum number of matches %l in ALL option in id-statement is greater than %l ",x,MAXPOSITIVE);
+					error = 1;
+				}
+				OldWork[5] = x;
+				if ( type != TYPEIDNEW ) {
+				  if ( type == TYPEIDOLD ) {
+					MesPrint("&Requested ALL option not allowed in idold/also statement.");
+					error = 1;
+				  }
+				  else if ( type == TYPEIF ) {
+					MesPrint("&Requested ALL option not allowed in if(match())");
+					error = 1;
+				  }
+				  else {
+					MesPrint("&ALL option only allowed in regular id-statement.");
+					error = 1;
+				  }
+				}
+				p++; inp = p;
+				AC.idoption = opt;
+				break;
 			default:
 				if ( pp != p ) {
 IllField:			c = *p; *p = 0;
@@ -707,6 +771,7 @@ IllField:			c = *p; *p = 0;
 /*
 	Now read the LHS
 */
+	ClearWildcardNames();
 	oldcpointer = AddLHS(AC.cbufnum) - C->Buffer;
 
 	*p = 0;
@@ -755,6 +820,7 @@ IllLeft:MesPrint("&Illegal LHS");
 		return(1);
 	}
 	if ( error == 0 ) {
+		WORD oldpolyfun = AR.PolyFun;
 		if ( NewSort(BHEAD0) || NewSort(BHEAD0) ) {
 			if ( !error ) error = 1;
 			return(error);
@@ -765,13 +831,15 @@ IllLeft:MesPrint("&Illegal LHS");
 		while ( --i >= 0 ) *ww++ = *mm++; AT.WorkPointer = ww;
 		AC.lhdollarflag = 0; oldEside = AR.Eside; AR.Eside = LHSIDE;
 		AR.Cnumlhs = C->numlhs;
+		AR.PolyFun = 0;
 		if ( Generator(BHEAD ow,C->numlhs) ) {
 			AR.Eside = oldEside;
-			LowerSortLevel(); LowerSortLevel(); goto IllLeft;
+			LowerSortLevel(); LowerSortLevel(); AR.PolyFun = oldpolyfun; goto IllLeft;
 		}
 		AR.Eside = oldEside;
 		AT.WorkPointer = w;
-		if ( EndSort(BHEAD w,0) < 0 ) { LowerSortLevel(); goto IllLeft; }
+		if ( EndSort(BHEAD w,0) < 0 ) { LowerSortLevel(); AR.PolyFun = oldpolyfun; goto IllLeft; }
+		AR.PolyFun = oldpolyfun;
 		if ( *w == 0 || *(w+*w) != 0 ) {
 			MesPrint("&LHS must be one term");
 			AC.lhdollarflag = 0;
@@ -793,6 +861,7 @@ IllLeft:MesPrint("&Illegal LHS");
 	C->numlhs--;
 
 	m = w + *w - 3;
+	AC.vectorlikeLHS = 0;
 	if ( !error ) {
 	  if ( m[2] != 3 || m[1] != 1 || *m != 1 ) {
 		if ( *m == 1 && m[1] == 1 && m[2] == -3 ) {
@@ -812,7 +881,7 @@ IllLeft:MesPrint("&Illegal LHS");
 			error = 1;
 		}
 		AC.DumNum = AM.IndDum;
-		OldWork[2] = ( OldWork[2] - ( OldWork[2] & SUBMASK ) ) | SUBALL;
+		OldWork[2] = ( OldWork[2] - ( OldWork[2] & SUBMASK ) ) | SUBVECTOR;
 		c1 = w[3];
 			/* We overwrite the LHS */
 		*w++ = INDTOIND;
@@ -825,6 +894,7 @@ IllLeft:MesPrint("&Illegal LHS");
 		w[3] = c1;
 		w[4] = AC.DumNum + WILDOFFSET;
 		OldWork[idhead+1] = w - OldWork - idhead;
+		AC.vectorlikeLHS = 1;
 	  }
 	  else {
 		AC.DumNum = 0;
@@ -901,8 +971,30 @@ IllLeft:MesPrint("&Illegal LHS");
 		if ( StudyPattern(OldWork) ) error = 1;
 	}
 	AT.WorkPointer = OldWork + OldWork[1];
-	OldWork[4] = AC.lhdollarflag;
+	if ( AC.lhdollarflag ) OldWork[4] |= DOLLARFLAG;
 	AC.lhdollarflag = 0;
+/*
+	Test whether the id/idold configuration is fine.
+*/
+	if ( type == TYPEIDOLD ) {
+		WORD ci = C->numlhs;
+		while ( ci >= 1 ) {
+			if ( C->lhs[ci][0] == TYPEIDNEW ) {
+				if ( (C->lhs[ci][2] & SUBMASK) == SUBALL ) {
+					MesPrint("&Idold/also cannot follow an id,all statement.");
+					error = 1;
+				}
+				break;
+			}
+			else if ( C->lhs[ci][0] == TYPEDETCURDUM ) { ci--; continue; }
+			else if ( C->lhs[ci][0] == TYPEIDOLD ) { ci--; continue; }
+			else ci = 0;
+		}
+		if ( ci < 1 ) {
+			MesPrint("&Idold/also should follow an id/idnew statement.");
+			error = 1;
+		}
+	}
 /*
 	Now the right hand side.
 */
@@ -921,7 +1013,7 @@ IllLeft:MesPrint("&Illegal LHS");
 /*
 	Actual adding happens only now after numrhs insertion
 */
-	/* if ( !error ) */ { AddNtoL(OldWork[1],OldWork); }
+	if ( !error ) { AddNtoL(OldWork[1],OldWork); }
 AllDone:
 	AC.lhdollarflag = 0;
 	AT.WorkPointer = FirstWork;
@@ -955,6 +1047,7 @@ int CoMultiply(UBYTE *inp)
 		*p = ',';
 		inp = p + 1;
 	}
+	ClearWildcardNames();
 	while ( *inp == ',' ) inp++;
 	AC.ProtoType = mularray+3;
 	mularray[7] = AC.cbufnum;
@@ -994,12 +1087,26 @@ int CoFill(UBYTE *inp)
 	p2 = p;
 	c = *p; *p = 0;
 	if ( ( GetVar(inp,&type,&funnum,CFUNCTION,WITHAUTO) == NAMENOTFOUND )
-	|| ( T = functions[funnum].tabl ) == 0 || c != '(' ) {
+	|| ( T = functions[funnum].tabl ) == 0 || ( T->numind > 0 && c != '(' ) ) {
 		MesPrint("&%s should be a table with argument(s)",inp);
 		*p = c; return(1);
 	}
 	oldT = T;
 	*p++ = c;
+	if ( T->numind == 0 ) {
+		if ( c == '(' ) {
+			if ( *p != ')' ) {
+				c = *p; *p = 0;
+				MesPrint("&%s should be a table without arguments",inp);
+				*p = c; return(1);
+			}
+			else { p++; }
+		}
+		else { p--; }
+		sum = 0;
+		p3 = p;
+		goto andagain;
+	}
 	for ( sum = 0, i = 0, w = oldwp; i < T->numind; i++ ) {
 		ParseSignedNumber(x,p);
 		if ( FG.cTable[p[-1]] != 1 || ( *p != ',' && *p != ')' ) ) {
@@ -1099,7 +1206,8 @@ redef:;
 		T->tablepointers[sum+5] = 0;
 #endif
 	}
-	p++; if ( *p != '=' ) {
+	if ( T->numind ) { p++; }
+	if ( *p != '=' ) {
 		MesPrint("&Fill statement misses = sign after the table element");
 		AC.cbufnum = oldcbufnum;
 		AT.WorkPointer = oldwp;
@@ -1146,8 +1254,10 @@ redef:;
 		inp1 = fake;
 /*		AT.WorkPointer += T->numind; */
 	}
-	else
-	inp1 = ++p; c = 0;
+	else {
+		inp1 = ++p;
+	}
+	c = 0;
 /*
 	Now we have the indices and p points to the rhs.
 */
@@ -1361,6 +1471,20 @@ int CoFillExpression(UBYTE *inp)
 		goto noway;
 	}
 	EXCHINOUT
+#ifdef WITHMPI
+	/*
+	 * The workers can't access to the data of the input expression. We need to
+	 * broadcast it to all the workers.
+	 */
+	PF_BroadcastExpr(&Expressions[expnum], AR.infile);
+	if ( PF.me == MASTER ) {
+		/*
+		 * Restore the file position on the master.
+		 */
+		POSITION pos;
+		SetEndScratch(AR.infile, &pos);
+	}
+#endif
 	fi = AR.infile;
 	if ( fi->handle >= 0 ) {
 		PUTZERO(oldposition);
@@ -1406,7 +1530,7 @@ int CoFillExpression(UBYTE *inp)
 					if ( weneedit ) {
 						m += m[1] - 1;
 						*m = *term - (m-term);
-						AddNtoC(AC.cbufnum,*m,m);
+						AddNtoC(AC.cbufnum,*m,m,3);
 						numdummies = DetCurDum(BHEAD term) - AM.IndDum;
 						if ( numdummies > T->numdummies ) T->numdummies = numdummies;
 					}
@@ -1414,7 +1538,7 @@ int CoFillExpression(UBYTE *inp)
 				}
 			}
 			if ( weneedit ) {
-				AddNtoC(AC.cbufnum,1,&zero);	/* Terminate old bracket */
+				AddNtoC(AC.cbufnum,1,&zero,4);	/* Terminate old bracket */
 				numcommu = numcommute(C->rhs[curelement],&(C->NumTerms[curelement]));
 				C->CanCommu[curelement] = numcommu;
 			}
@@ -1566,12 +1690,12 @@ Wrong!!!!			C->rhs[T->tablepointers[sum]] = C->Pointer;
 newentry:	if ( *m == HAAKJE ) { m += m[1] - 1; }
 			else m--;
 			*m = *term - (m-term);
-			AddNtoC(AC.cbufnum,*m,m);
+			AddNtoC(AC.cbufnum,*m,m,5);
 			curelement = T->tablepointers[sum];
 nextterm:;
 		}
 		if ( weneedit ) {
-			AddNtoC(AC.cbufnum,1,&zero);	/* Terminate old bracket */
+			AddNtoC(AC.cbufnum,1,&zero,6);	/* Terminate old bracket */
 			numcommu = numcommute(C->rhs[curelement],&(C->NumTerms[curelement]));
 			C->CanCommu[curelement] = numcommu;
 		}
