@@ -14,7 +14,7 @@
  */
 /* #[ License : */
 /*
- *   Copyright (C) 1984-2013 J.A.M. Vermaseren
+ *   Copyright (C) 1984-2023 J.A.M. Vermaseren
  *   When using this file you are requested to refer to the publication
  *   J.A.M.Vermaseren "New features of FORM" math-ph/0010025
  *   This is considered a matter of courtesy as the development was paid
@@ -106,18 +106,26 @@ static KEYWORD com2commands[] = {
 	,{"aputinside",     (TFUN)CoAntiPutInside,    STATEMENT,    PARTEST}
 	,{"argexplode",     (TFUN)CoArgExplode,       STATEMENT,    PARTEST}
 	,{"argimplode",     (TFUN)CoArgImplode,       STATEMENT,    PARTEST}
+	,{"argtoextrasymbol",(TFUN)CoArgToExtraSymbol,STATEMENT,    PARTEST}
 	,{"argument",       (TFUN)CoArgument,         STATEMENT,    PARTEST}
 	,{"assign",         (TFUN)CoAssign,           STATEMENT,    PARTEST}
 	,{"auto",           (TFUN)CoAuto,             DECLARATION,  PARTEST}
 	,{"autodeclare",    (TFUN)CoAuto,             DECLARATION,  PARTEST}
+	,{"break",          (TFUN)CoBreak,            STATEMENT,    PARTEST}
+	,{"canonicalize",   (TFUN)CoCanonicalize,     STATEMENT,    PARTEST}
+	,{"case",           (TFUN)CoCase,             STATEMENT,    PARTEST}
 	,{"chainin",        (TFUN)CoChainin,          STATEMENT,    PARTEST}
 	,{"chainout",       (TFUN)CoChainout,         STATEMENT,    PARTEST}
 	,{"chisholm",       (TFUN)CoChisholm,         STATEMENT,    PARTEST}
 	,{"cleartable",     (TFUN)CoClearTable,       DECLARATION,  PARTEST}
 	,{"collect",        (TFUN)CoCollect,          SPECIFICATION,PARTEST}
+	,{"commuteinset",   (TFUN)CoCommuteInSet,     DECLARATION,  PARTEST}
 	,{"contract",       (TFUN)CoContract,         STATEMENT,    PARTEST}
+	,{"copyspectator"  ,(TFUN)CoCopySpectator,    DEFINITION,   PARTEST}
+	,{"createspectator",(TFUN)CoCreateSpectator,  DECLARATION,  PARTEST}
 	,{"ctable",         (TFUN)CoCTable,           DECLARATION,  PARTEST}
 	,{"deallocatetable",(TFUN)CoDeallocateTable,  DECLARATION,  PARTEST}
+	,{"default",        (TFUN)CoDefault,          STATEMENT,    PARTEST}
 	,{"delete",         (TFUN)CoDelete,           SPECIFICATION,PARTEST}
 	,{"denominators",   (TFUN)CoDenominators,     STATEMENT,    PARTEST}
 	,{"disorder",       (TFUN)CoDisorder,         STATEMENT,    PARTEST}
@@ -127,12 +135,14 @@ static KEYWORD com2commands[] = {
 	,{"dropsymbols",    (TFUN)CoDropSymbols,      STATEMENT,    PARTEST}
 	,{"else",           (TFUN)CoElse,             STATEMENT,    PARTEST}
 	,{"elseif",         (TFUN)CoElseIf,           STATEMENT,    PARTEST}
+	,{"emptyspectator", (TFUN)CoEmptySpectator,   SPECIFICATION,PARTEST}
 	,{"endargument",    (TFUN)CoEndArgument,      STATEMENT,    PARTEST}
 	,{"enddo",          (TFUN)CoEndDo,            STATEMENT,    PARTEST}
 	,{"endif",          (TFUN)CoEndIf,            STATEMENT,    PARTEST}
 	,{"endinexpression",(TFUN)CoEndInExpression,  STATEMENT,    PARTEST}
 	,{"endinside",      (TFUN)CoEndInside,        STATEMENT,    PARTEST}
 	,{"endrepeat",      (TFUN)CoEndRepeat,        STATEMENT,    PARTEST}
+	,{"endswitch",      (TFUN)CoEndSwitch,        STATEMENT,    PARTEST}
 	,{"endterm",        (TFUN)CoEndTerm,          STATEMENT,    PARTEST}
 	,{"endwhile",       (TFUN)CoEndWhile,         STATEMENT,    PARTEST}
 	,{"exit",           (TFUN)CoExit,             STATEMENT,    PARTEST}
@@ -185,6 +195,7 @@ static KEYWORD com2commands[] = {
 	,{"pushhide",       (TFUN)CoPushHide,         SPECIFICATION,PARTEST}
 	,{"putinside",      (TFUN)CoPutInside,        STATEMENT,    PARTEST}
 	,{"ratio",          (TFUN)CoRatio,            STATEMENT,    PARTEST}
+	,{"removespectator",(TFUN)CoRemoveSpectator,  SPECIFICATION,PARTEST}
 	,{"renumber",       (TFUN)CoRenumber,         STATEMENT,    PARTEST}
 	,{"repeat",         (TFUN)CoRepeat,           STATEMENT,    PARTEST}
 	,{"replaceloop",    (TFUN)CoReplaceLoop,      STATEMENT,    PARTEST}
@@ -199,6 +210,7 @@ static KEYWORD com2commands[] = {
 	,{"splitlastarg",   (TFUN)CoSplitLastArg,     STATEMENT,    PARTEST}
 	,{"stuffle",        (TFUN)CoStuffle,          STATEMENT,    PARTEST}
 	,{"sum",            (TFUN)CoSum,              STATEMENT,    PARTEST}
+	,{"switch",         (TFUN)CoSwitch,           STATEMENT,    PARTEST}
 	,{"table",          (TFUN)CoTable,            DECLARATION,  PARTEST}
 	,{"tablebase",      (TFUN)CoTableBase,        DECLARATION,  PARTEST}
 	,{"tb",             (TFUN)CoTableBase,        DECLARATION,  PARTEST}
@@ -206,6 +218,7 @@ static KEYWORD com2commands[] = {
 	,{"testuse",        (TFUN)CoTestUse,          STATEMENT,    PARTEST}
 	,{"threadbucketsize",(TFUN)CoThreadBucket,    DECLARATION,  PARTEST}
 	,{"topolynomial",   (TFUN)CoToPolynomial,     STATEMENT,    PARTEST}
+	,{"tospectator",    (TFUN)CoToSpectator,      STATEMENT,    PARTEST}
 	,{"totensor",       (TFUN)CoToTensor,         STATEMENT,    PARTEST}
 	,{"tovector",       (TFUN)CoToVector,         STATEMENT,    PARTEST}
 	,{"trace4",         (TFUN)CoTrace4,           STATEMENT,    PARTEST}
@@ -752,8 +765,8 @@ int CompileSubExpressions(SBYTE *tokens)
 			this minimum. Ignoring this might lead to really rare and
 			hard to find errors, years from now.
 */
-			if ( insubexpbuffers >= 0x3FFFFFL ) {
-				MesPrint("&More than 2^22 subexpressions inside one expression");
+			if ( insubexpbuffers >= MAXSUBEXPRESSIONS ) {
+				MesPrint("&More than %d subexpressions inside one expression",(WORD)MAXSUBEXPRESSIONS);
 				Terminate(-1);
 			}
 			if ( subexpbuffers+insubexpbuffers >= topsubexpbuffers ) {
@@ -1020,6 +1033,8 @@ dovector:		if ( inset == 0 ) x1 += AM.OffsetVector;
 					if ( s[0] == TFUNOPEN && s[1] == TEXPRESSION ) {
 doexpr:					s += 2;
 						*t++ = x1; *t++ = FUNHEAD+2; *t++ = 0;
+						if ( x1 == AR.PolyFun && AR.PolyFunType == 2 && AR.Eside != LHSIDE )
+								t[-1] |= MUSTCLEANPRF;
 						FILLFUN3(t)
 						x2 = 0; while ( *s >= 0 ) { x2 = x2*128 + *s++; }
 						*t++ = -EXPRESSION; *t++ = x2;
@@ -1070,6 +1085,8 @@ doexpr:					s += 2;
 								MesPrint("&Problems with argument of FirstTerm_");
 							else if ( x1 == TERMSINEXPR )
 								MesPrint("&Problems with argument of TermsIn_");
+							else if ( x1 == SIZEOFFUNCTION )
+								MesPrint("&Problems with argument of SizeOf_");
 							else if ( x1 == NUMFACTORS )
 								MesPrint("&Problems with argument of NumFactors_");
 							else
@@ -1081,7 +1098,7 @@ doexpr:					s += 2;
 						goto fin;
 					}
 				}
-				else if ( x1 == TERMSINEXPR || x1 == FACTORIN
+				else if ( x1 == TERMSINEXPR || x1 == SIZEOFFUNCTION || x1 == FACTORIN
 				 || x1 == NUMFACTORS || x1 == FIRSTTERM || x1 == CONTENTTERM ) {
 					if ( s[0] == TFUNOPEN && s[1] == TEXPRESSION ) goto doexpr;
 					if ( s[0] == TFUNOPEN && s[1] == TDOLLAR ) {
@@ -1093,6 +1110,8 @@ doexpr:					s += 2;
 						if ( *s != TFUNCLOSE ) {
 							if ( x1 == TERMSINEXPR )
 								MesPrint("&Problems with argument of TermsIn_");
+							else if ( x1 == SIZEOFFUNCTION )
+								MesPrint("&Problems with argument of SizeOf_");
 							else if ( x1 == NUMFACTORS )
 								MesPrint("&Problems with argument of NumFactors_");
 							else
@@ -1110,9 +1129,14 @@ doexpr:					s += 2;
 				if ( functions[x3-FUNCTION].commute ) cc = 1;
 				if ( *s != TFUNOPEN ) {
 					*t++ = x1; *t++ = FUNHEAD; *t++ = 0;
+					if ( x1 == AR.PolyFun && AR.PolyFunType == 2 && AR.Eside != LHSIDE )
+							t[-1] |= MUSTCLEANPRF;
 					FILLFUN3(t) sumlevel = 0; goto fin;
 				}
-				v = t; *t++ = x1; *t++ = FUNHEAD; *t++ = DIRTYFLAG; FILLFUN3(t)
+				v = t; *t++ = x1; *t++ = FUNHEAD; *t++ = DIRTYFLAG;
+				if ( x1 == AR.PolyFun && AR.PolyFunType == 2 && AR.Eside != LHSIDE )
+						t[-1] |= MUSTCLEANPRF;
+				FILLFUN3(t)
 				needarg = -1;
 				if ( !inset && functions[x3-FUNCTION].spec >= TENSORFUNCTION ) {
 dotensor:
@@ -1161,6 +1185,7 @@ dotensor:
 									if ( x2 >= 0 && x2 < AM.OffsetIndex ) {
 										*t++ = x2; break;
 									}
+									/* fall through */
 								default:
 									MesPrint("&Illegal type of set inside tensor");
 									error = 1;
@@ -1221,6 +1246,7 @@ dotensor:
 				else {
 dofunction:			firstsumarg = 1;
 					do {
+						unsigned int ux2;
 						s++;
 						c = *s++;
 						if ( c == TMINUS && ( *s == TVECTOR || *s == TNUMBER
@@ -1229,7 +1255,8 @@ dofunction:			firstsumarg = 1;
 						}
 						else minus = 0;
 						base = ( c == TNUMBER ) ? 100: 128;
-						x2 = 0; while ( *s >= 0 ) { x2 = base*x2 + *s++; }
+						ux2 = 0; while ( *s >= 0 ) { ux2 = base*ux2 + *s++; }
+						x2 = ux2;  /* may cause an implementation-defined behaviour */
 /*
 		!!!!!!!!  What if it does not fit?
 */
@@ -1325,10 +1352,15 @@ dofunction:			firstsumarg = 1;
 							case TFUNCTION:
 								*t++ = -x2-FUNCTION;
 								break;
+							case TSET:
+								*t++ = -SETSET;
+								*t++ = x2;
+								break;
 							case TWILDARG:
 								*t++ = -ARGWILD; *t++ = x2; break;
 							case TSETDOL:
 								x2 = -x2;
+								/* fall through */
 							case TSETNUM:
 								if ( inset == 0 ) {
 									w1 = t; t += 2; w2 = t;
@@ -1385,7 +1417,7 @@ dofunction:			firstsumarg = 1;
 								break;
 							case TNUMBER:
 							case TNUMBER1:
-								if ( minus ) x2 = -x2;
+								if ( minus ) x2 = UnsignedToInt(-IntAbs(x2));
 								*t++ = -SNUMBER;
 								*t++ = x2;
 								break;
@@ -1418,6 +1450,8 @@ dofunction:			firstsumarg = 1;
 					x1 += FUNCTION;
 					cc = 1;
 					v = t; *t++ = x1; *t++ = FUNHEAD; *t++ = DIRTYFLAG;
+					if ( x1 == AR.PolyFun && AR.PolyFunType == 2 && AR.Eside != LHSIDE )
+							t[-1] |= MUSTCLEANPRF;
 					FILLFUN3(t)
 					needarg = -1; goto dofunction;
 				}
@@ -1491,6 +1525,7 @@ dofunction:			firstsumarg = 1;
 								*t++ = INDEX; *t++ = 3; *t++ = x2; break;
 							case TSETDOL:
 								x2 = -x2;
+								/* fall through */
 							case TSETNUM:
 								if ( inset == 0 ) {
 									w1 = t; t += 2; w2 = t;
@@ -1850,7 +1885,7 @@ int CompleteTerm(WORD *term, UWORD *numer, UWORD *denom, WORD nnum, WORD nden, i
 	for ( ; i < nsize; i++ ) *t++ = 0;
 	*t++ = (2*nsize+1)*sign;
 	*term = t - term;
-	AddNtoC(AC.cbufnum,*term,term);
+	AddNtoC(AC.cbufnum,*term,term,7);
 	return(0);
 }
 
@@ -1901,8 +1936,8 @@ int CodeFactors(SBYTE *tokens)
 			|| t[-1] == TSETCLOSE || t[-1] == TFUNCLOSE ) {
 				subexp = CodeGenerator(tokens);
 				if ( subexp < 0 ) error = -1;
-				if ( insubexpbuffers >= 0x3FFFFFL ) {
-					MesPrint("&More than 2^22 subexpressions inside one expression");
+				if ( insubexpbuffers >= MAXSUBEXPRESSIONS ) {
+					MesPrint("&More than %d subexpressions inside one expression",(WORD)MAXSUBEXPRESSIONS);
 					Terminate(-1);
 				}
 				if ( subexpbuffers+insubexpbuffers >= topsubexpbuffers ) {
@@ -2142,8 +2177,8 @@ dopowerd:
 	e->vflags |= ISFACTORIZED;
 	subexp = CodeGenerator(outtokens);
 	if ( subexp < 0 ) error = -1;
-	if ( insubexpbuffers >= 0x3FFFFFL ) {
-		MesPrint("&More than 2^22 subexpressions inside one expression");
+	if ( insubexpbuffers >= MAXSUBEXPRESSIONS ) {
+		MesPrint("&More than %d subexpressions inside one expression",(WORD)MAXSUBEXPRESSIONS);
 		Terminate(-1);
 	}
 	if ( subexpbuffers+insubexpbuffers >= topsubexpbuffers ) {
@@ -2189,8 +2224,8 @@ WORD GenerateFactors(WORD n,WORD inc)
 	subexp = CodeGenerator(tokenbuffer);
 	if ( subexp < 0 ) error = -1;
 	M_free(tokenbuffer,"GenerateFactors");
-	if ( insubexpbuffers >= 0x3FFFFFL ) {
-		MesPrint("&More than 2^22 subexpressions inside one expression");
+	if ( insubexpbuffers >= MAXSUBEXPRESSIONS ) {
+		MesPrint("&More than %d subexpressions inside one expression",(WORD)MAXSUBEXPRESSIONS);
 		Terminate(-1);
 	}
 	if ( subexpbuffers+insubexpbuffers >= topsubexpbuffers ) {
