@@ -1,5 +1,4 @@
 #ifndef __FDECLARE__
- 
 #define __FDECLARE__
 
 /** @file declare.h
@@ -9,7 +8,7 @@
 
 /* #[ License : */
 /*
- *   Copyright (C) 1984-2013 J.A.M. Vermaseren
+ *   Copyright (C) 1984-2023 J.A.M. Vermaseren
  *   When using this file you are requested to refer to the publication
  *   J.A.M.Vermaseren "New features of FORM" math-ph/0010025
  *   This is considered a matter of courtesy as the development was paid
@@ -42,12 +41,15 @@
 #define MiN(x,y) ((x) < (y) ? (x): (y))
 #define ABS(x) ( (x) < 0 ? -(x): (x) )
 #define SGN(x) ( (x) > 0 ? 1 : (x) < 0 ? -1 : 0 )
-#define REDLENG(x) ((((x)<0)?((x)+1):((x)-1))>>1)
-#define INCLENG(x) (((x)<0)?(((x)<<1)-1):(((x)<<1)+1))
+#define REDLENG(x) ((((x)<0)?((x)+1):((x)-1))/2)
+#define INCLENG(x) (((x)<0)?(((x)*2)-1):(((x)*2)+1))
 #define GETCOEF(x,y) x += *x;y = x[-1];x -= ABS(y);y=REDLENG(y)
 #define GETSTOP(x,y) y=x+(*x)-1;y -= ABS(*y)-1
 #define StuffAdd(x,y)  (((x)<0?-1:1)*(y)+((y)<0?-1:1)*(x))
  
+#define EXCHN(t1,t2,n) { WORD a,i; for(i=0;i<n;i++){a=t1[i];t1[i]=t2[i];t2[i]=a;} }
+#define EXCH(x,y) { WORD a = (x); (x) = (y); (y) = a; }
+ 
 #define TOKENTOLINE(x,y) if ( AC.OutputSpaces == NOSPACEFORMAT ) { \
 		TokenToLine((UBYTE *)(y)); } else { TokenToLine((UBYTE *)(x)); }
 
@@ -105,7 +107,7 @@
 #define CYCLE1(t,a,i) {t iX=*a; WORD jX; for(jX=1;jX<i;jX++)a[jX-1]=a[jX]; a[i-1]=iX;}
 
 #define AddToCB(c,wx) if(c->Pointer>=c->Top) \
-		DoubleCbuffer(c-cbuf,c->Pointer); \
+		DoubleCbuffer(c-cbuf,c->Pointer,21); \
 		*(c->Pointer)++ = wx;
 
 #define EXCHINOUT { FILEHANDLE *ffFi = AR.outfile; \
@@ -162,6 +164,18 @@
 #define COPY1ARG(s1,t1) { int ica; if ( (ica=*t1) > 0 ) { NCOPY(s1,t1,ica) } \
 		else if(*t1<=-FUNCTION){*s1++=*t1++;} else{*s1++=*t1++;*s1++=*t1++;} }
 
+/**
+ * Fills a buffer by zero in the range [begin,end).
+ *
+ * @param  w      The buffer.
+ * @param  begin  The index for the beginning of the range.
+ * @param  end    The index for the end of the range (exclusive).
+ */
+#define ZeroFillRange(w,begin,end) do { \
+	int tmp_i; \
+	for ( tmp_i = begin; tmp_i < end; tmp_i++ ) { (w)[tmp_i] = 0; } \
+} while (0)
+
 #define TABLESIZE(a,b) (((WORD)sizeof(a))/((WORD)sizeof(b)))
 #define WORDDIF(x,y) (WORD)(x-y)
 #define wsizeof(a) ((WORD)sizeof(a))
@@ -222,6 +236,7 @@
 #define ISLESSPOS(pp1,pp2) ( (pp1).p1 < (pp2).p1 )
 #define ISGEPOS(pp1,pp2) ( (pp1).p1 >= (pp2).p1 )
 #define ISNOTZEROPOS(pp) ( (pp).p1 != 0 )
+#define ISZEROPOS(pp) ( (pp).p1 == 0 )
 #define ISPOSPOS(pp) ( (pp).p1 > 0 )
 #define ISNEGPOS(pp) ( (pp).p1 < 0 )
 extern VOID TELLFILE(int,POSITION *);
@@ -260,15 +275,19 @@ extern VOID TELLFILE(int,POSITION *);
 
 #define TermMalloc(x) ( (AT.TermMemTop <= 0 ) ? TermMallocAddMemory(BHEAD0), AT.TermMemHeap[--AT.TermMemTop]: AT.TermMemHeap[--AT.TermMemTop] )
 #define NumberMalloc(x) ( (AT.NumberMemTop <= 0 ) ? NumberMallocAddMemory(BHEAD0), AT.NumberMemHeap[--AT.NumberMemTop]: AT.NumberMemHeap[--AT.NumberMemTop] )
+#define CacheNumberMalloc(x) ( (AT.CacheNumberMemTop <= 0 ) ? CacheNumberMallocAddMemory(BHEAD0), AT.CacheNumberMemHeap[--AT.CacheNumberMemTop]: AT.CacheNumberMemHeap[--AT.CacheNumberMemTop] )
 #define TermFree(TermMem,x) AT.TermMemHeap[AT.TermMemTop++] = (WORD *)(TermMem)
 #define NumberFree(NumberMem,x) AT.NumberMemHeap[AT.NumberMemTop++] = (UWORD *)(NumberMem)
+#define CacheNumberFree(NumberMem,x) AT.CacheNumberMemHeap[AT.CacheNumberMemTop++] = (UWORD *)(NumberMem)
 
 #else
 
 #define TermMalloc(x) TermMalloc2(BHEAD (char *)(x))
 #define NumberMalloc(x) NumberMalloc2(BHEAD (char *)(x))
+#define CacheNumberMalloc(x) CacheNumberMalloc2(BHEAD (char *)(x))
 #define TermFree(x,y) TermFree2(BHEAD (WORD *)(x),(char *)(y))
 #define NumberFree(x,y) NumberFree2(BHEAD (UWORD *)(x),(char *)(y))
+#define CacheNumberFree(x,y) CacheNumberFree2(BHEAD (UWORD *)(x),(char *)(y))
 
 #endif
 
@@ -287,11 +306,108 @@ extern VOID TELLFILE(int,POSITION *);
  * using increment/decrement operators on it in the left-hand side may be
  * confusing.
  */
-#define NestingChecksum() (AC.IfLevel + AC.RepLevel + AC.arglevel + AC.insidelevel + AC.termlevel + AC.inexprlevel + AC.dolooplevel)
+#define NestingChecksum() (AC.IfLevel + AC.RepLevel + AC.arglevel + AC.insidelevel + AC.termlevel + AC.inexprlevel + AC.dolooplevel +AC.SwitchLevel)
 #define MesNesting() MesPrint("&Illegal nesting of if, repeat, argument, inside, term, inexpression and do")
 
+#define MarkPolyRatFunDirty(T) {if(*T&&AR.PolyFunType==2){WORD *TP,*TT;TT=T+*T;TT-=ABS(TT[-1]);\
+TP=T+1;while(TP<TT){if(*TP==AR.PolyFun){TP[2]|=(DIRTYFLAG|MUSTCLEANPRF);}TP+=TP[1];}}}
+
 /*
+	Macros for nesting input levels for #$name = ...; assign instructions.
+	Note that the level should never go below zero.
+*/
+#define PUSHPREASSIGNLEVEL AP.PreAssignLevel++; { GETIDENTITY \
+    if ( AP.PreAssignLevel >= AP.MaxPreAssignLevel ) { int i; \
+		LONG *ap = (LONG *)Malloc1(2*AP.MaxPreAssignLevel*sizeof(LONG *),"PreAssignStack"); \
+		for ( i = 0; i < AP.MaxPreAssignLevel; i++ ) ap[i] = AP.PreAssignStack[i]; \
+		M_free(AP.PreAssignStack,"PreAssignStack"); \
+		AP.MaxPreAssignLevel *= 2; AP.PreAssignStack = ap; \
+	} \
+	*AT.WorkPointer++ = AP.PreContinuation; AP.PreContinuation = 0; \
+	AP.PreAssignStack[AP.PreAssignLevel] = AC.iPointer - AC.iBuffer; }
+
+#define POPPREASSIGNLEVEL if ( AP.PreAssignLevel > 0 ) { GETIDENTITY \
+	AC.iPointer = AC.iBuffer + AP.PreAssignStack[AP.PreAssignLevel--]; \
+	AP.PreContinuation = *--AT.WorkPointer; \
+	*AC.iPointer = 0; }
+
+/*
+	MesPrint("P-level popped to %d with %d",AP.PreAssignLevel,(WORD)(AC.iPointer - AC.iBuffer));
+
   	#] Macro's : 
+  	#[ Inline functions :
+*/
+
+/*
+ * The following three functions give the unsigned absolute value of a signed
+ * integer even for the most negative integer. This is beyond the scope of
+ * the standard abs() function and its family, whose return-values are signed.
+ * In short, we should not use the unary minus operator with signed numbers
+ * unless we are sure that there are no integer overflows. Instead, we rely on
+ * two well-defined operations: (i) signed-to-unsigned conversion and
+ * (ii) unary minus of unsigned operands.
+ *
+ * See also:
+ *   https://stackoverflow.com/a/4536188   (Unary minus and signed-to-unsigned conversion)
+ *   https://stackoverflow.com/q/8026694   (C: unary minus operator behavior with unsigned operands)
+ *   https://stackoverflow.com/q/1610947   (Why does stdlib.h's abs() family of functions return a signed value?)
+ *   https://blog.regehr.org/archives/226  (A Guide to Undefined Behavior in C and C++, Part 2)
+ */
+static inline unsigned int IntAbs(int x)
+{
+	if ( x >= 0 ) return x;
+	return(-((unsigned int)x));
+}
+
+static inline UWORD WordAbs(WORD x)
+{
+	if ( x >= 0 ) return x;
+	return(-((UWORD)x));
+}
+
+static inline ULONG LongAbs(LONG x)
+{
+	if ( x >= 0 ) return x;
+	return(-((ULONG)x));
+}
+
+/*
+ * The following functions provide portable unsigned-to-signed conversions
+ * (to avoid the implementation-defined behaviour), which is expected to be
+ * optimized to a no-op.
+ *
+ * See also:
+ *   https://stackoverflow.com/a/13208789  (Efficient unsigned-to-signed cast avoiding implementation-defined behavior)
+ */
+static inline int UnsignedToInt(unsigned int x)
+{
+	extern void Terminate(int);
+	if ( x <= INT_MAX ) return(x);
+	if ( x >= (unsigned int)INT_MIN ) return((int)(x - INT_MIN) + INT_MIN);
+	Terminate(1);
+	return(0);
+}
+
+static inline WORD UWordToWord(UWORD x)
+{
+	extern void Terminate(int);
+	if ( x <= WORD_MAX_VALUE ) return(x);
+	if ( x >= (UWORD)WORD_MIN_VALUE ) return((WORD)(x - WORD_MIN_VALUE) + WORD_MIN_VALUE);
+	Terminate(1);
+	return(0);
+}
+
+static inline LONG ULongToLong(ULONG x)
+{
+	extern void Terminate(int);
+	if ( x <= LONG_MAX_VALUE ) return(x);
+	if ( x >= (ULONG)LONG_MIN_VALUE ) return((LONG)(x - LONG_MIN_VALUE) + LONG_MIN_VALUE);
+	Terminate(1);
+	return(0);
+}
+
+/*
+  	#] Inline functions : 
   	#[ Thread objects :
 */
 
@@ -304,7 +420,7 @@ extern VOID TELLFILE(int,POSITION *);
 #ifdef WITHPTHREADS
 
 #define EXTERNLOCK(x) extern pthread_mutex_t x;
-#define INILOCK(x)    pthread_mutex_t x = PTHREAD_MUTEX_INITIALIZER
+#define INILOCK(x)    pthread_mutex_t x = PTHREAD_MUTEX_INITIALIZER;
 #define EXTERNRWLOCK(x) extern pthread_rwlock_t x;
 #define INIRWLOCK(x)    pthread_rwlock_t x = PTHREAD_RWLOCK_INITIALIZER;
 #ifdef DEBUGGINGLOCKS
@@ -351,6 +467,10 @@ extern VOID TELLFILE(int,POSITION *);
   	#] Thread objects : 
   	#[ Declarations :
 */
+ 
+#ifdef TERMMALLOCDEBUG
+extern WORD **DebugHeap1, **DebugHeap2;
+#endif
 
 /**
  *	All functions (well, nearly all) are declared here.
@@ -381,13 +501,14 @@ extern WORD   Chisholm(PHEAD WORD *,WORD);
 extern WORD   CleanExpr(WORD);
 extern VOID   CleanUp(WORD);
 extern VOID   ClearWild(PHEAD0);
+extern WORD   CompareFunctions(WORD *,WORD *);
 extern WORD   Commute(WORD *,WORD *);
 extern WORD   DetCommu(WORD *);
 extern WORD   DoesCommu(WORD *);
 extern int    CompArg(WORD *,WORD *);
 extern WORD   CompCoef(WORD *, WORD *);
 extern WORD   CompGroup(PHEAD WORD,WORD **,WORD *,WORD *,WORD);
-extern WORD   Compare1(PHEAD WORD *,WORD *,WORD);
+extern WORD   Compare1(WORD *,WORD *,WORD);
 extern WORD   CountDo(WORD *,WORD *);
 extern WORD   CountFun(WORD *,WORD *);
 extern WORD   DimensionSubterm(WORD *);
@@ -403,14 +524,24 @@ extern WORD   DivRat(PHEAD UWORD *,WORD,UWORD *,WORD,UWORD *,WORD *);
 extern WORD   Divvy(PHEAD UWORD *,WORD *,UWORD *,WORD);
 extern WORD   DoDelta(WORD *);
 extern WORD   DoDelta3(PHEAD WORD *,WORD);
+extern WORD   TestPartitions(WORD *, PARTI *);
+extern WORD   DoPartitions(PHEAD WORD *,WORD);
+extern int    CoCanonicalize(UBYTE *);
+extern int    DoCanonicalize(PHEAD WORD *, WORD *);
+extern WORD   GenTopologies(PHEAD WORD *,WORD);
+extern WORD   GenDiagrams(PHEAD WORD *,WORD);
+extern int    DoTopologyCanonicalize(PHEAD WORD *,WORD,WORD,WORD *);
+extern int    DoShattering(PHEAD WORD *,WORD *,WORD *,WORD);
+extern WORD   GenerateTopologies(PHEAD WORD,WORD,WORD,WORD);
 extern WORD   DoTableExpansion(WORD *,WORD);
 extern WORD   DoDistrib(PHEAD WORD *,WORD);
-extern WORD   DoShuffle(PHEAD WORD *,WORD,WORD,WORD);
-extern int    Shuffle(PHEAD WORD *, WORD *, WORD *);
-extern int    FinishShuffle(PHEAD WORD *);
-extern WORD   DoStuffle(PHEAD WORD *,WORD,WORD,WORD);
-extern int    Stuffle(PHEAD WORD *, WORD *, WORD *);
-extern int    FinishStuffle(PHEAD WORD *);
+extern WORD   DoShuffle(WORD *,WORD,WORD,WORD);
+extern WORD   DoPermutations(PHEAD WORD *,WORD);
+extern int    Shuffle(WORD *, WORD *, WORD *);
+extern int    FinishShuffle(WORD *);
+extern WORD   DoStuffle(WORD *,WORD,WORD,WORD);
+extern int    Stuffle(WORD *, WORD *, WORD *);
+extern int    FinishStuffle(WORD *);
 extern WORD  *StuffRootAdd(WORD *, WORD *, WORD *);
 extern WORD   TestUse(WORD *,WORD);
 extern DBASE *FindTB(UBYTE *);
@@ -421,6 +552,7 @@ extern WORD   ApplyReset(WORD);
 extern WORD   TableReset(VOID);
 extern VOID   ReWorkT(WORD *,WORD *,WORD);
 extern WORD   GetIfDollarNum(WORD *, WORD *);
+extern int    FindVar(WORD *,WORD *);
 extern WORD   DoIfStatement(PHEAD WORD *,WORD *);
 extern WORD   DoOnePow(PHEAD WORD *,WORD,WORD,WORD *,WORD *,WORD,WORD *);
 extern void   DoRevert(WORD *,WORD *);
@@ -494,6 +626,7 @@ extern WORD   MultDo(PHEAD WORD *,WORD *);
 extern WORD   NewSort(PHEAD0);
 extern WORD   ExtraSymbol(WORD,WORD,WORD,WORD *,WORD *);
 extern WORD   Normalize(PHEAD WORD *);
+extern WORD   BracketNormalize(PHEAD WORD *);
 extern VOID   DropCoefficient(PHEAD WORD *);
 extern VOID   DropSymbols(PHEAD WORD *);
 extern int    PutInside(PHEAD WORD *, WORD *);
@@ -501,9 +634,10 @@ extern WORD   OpenTemp(VOID);
 extern VOID   Pack(UWORD *,WORD *,UWORD *,WORD );
 extern LONG   PasteFile(PHEAD WORD,WORD *,POSITION *,WORD **,RENUMBER,WORD *,WORD);
 extern WORD   Permute(PERM *,WORD);
+extern WORD   PermuteP(PERMP *,WORD);
 extern WORD   PolyFunMul(PHEAD WORD *);
 extern WORD   PopVariables(VOID);
-extern WORD   PrepPoly(PHEAD WORD *);
+extern WORD   PrepPoly(PHEAD WORD *,WORD);
 extern WORD   Processor(VOID);
 extern WORD   Product(UWORD *,WORD *,WORD);
 extern VOID   PrtLong(UWORD *,WORD,UBYTE *);
@@ -541,11 +675,7 @@ extern WORD   Sflush(FILEHANDLE *);
 extern WORD   Simplify(PHEAD UWORD *,WORD *,UWORD *,WORD *);
 extern WORD   SortWild(WORD *,WORD);
 extern FILE  *LocateBase(char **,char **);
-#ifdef NEWSPLITMERGE
 extern LONG   SplitMerge(PHEAD WORD **,LONG);
-#else
-extern VOID   SplitMerge(PHEAD WORD **,LONG);
-#endif
 extern WORD   StoreTerm(PHEAD WORD *);
 extern VOID   SubPLon(UWORD *,WORD,UWORD *,WORD,UWORD *,WORD *);
 extern VOID   Substitute(PHEAD WORD *,WORD *,WORD);
@@ -586,7 +716,7 @@ extern VOID   UnPack(UWORD *,WORD,WORD *,WORD *);
 extern WORD   VarStore(UBYTE *,WORD,WORD,WORD);
 extern WORD   WildFill(PHEAD WORD *,WORD *,WORD *);
 extern WORD   WriteAll(VOID);
-extern WORD   WriteOne(UBYTE *,int,int);
+extern WORD   WriteOne(UBYTE *,int,int,WORD);
 extern VOID   WriteArgument(WORD *);
 extern WORD   WriteExpression(WORD *,LONG);
 extern WORD   WriteInnerTerm(WORD *,WORD);
@@ -625,12 +755,15 @@ extern STREAM *OpenStream(UBYTE *,int,int,int);
 extern int    LocateFile(UBYTE **,int);
 extern STREAM *CloseStream(STREAM *);
 extern VOID   PositionStream(STREAM *,LONG);
+extern int    ReverseStatements(STREAM *);
 extern int    ProcessOption(UBYTE *,UBYTE *,int);
 extern int    DoSetups(VOID);
 extern VOID   Terminate(int);
 extern NAMENODE *GetNode(NAMETREE *,UBYTE *);
 extern int    AddName(NAMETREE *,UBYTE *,WORD,WORD,int *);
 extern int    GetName(NAMETREE *,UBYTE *,WORD *,int);
+extern UBYTE  *GetFunction(UBYTE *,WORD *);
+extern UBYTE  *GetNumber(UBYTE *,WORD *);
 extern int    GetLastExprName(UBYTE *,WORD *);
 extern int    GetAutoName(UBYTE *,WORD *);
 extern int    GetVar(UBYTE *,WORD *,WORD *,int,int);
@@ -653,6 +786,7 @@ extern int    AddDubious(UBYTE *);
 extern int    AddIndex(UBYTE *,int,int);
 extern UBYTE *DoDimension(UBYTE *,int *,int *);
 extern int    AddFunction(UBYTE *,int,int,int,int,int,int,int);
+extern int    CoCommuteInSet(UBYTE *);
 extern int    CoFunction(UBYTE *,int,int);
 extern int    TestName(UBYTE *);
 extern int    AddSet(UBYTE *,WORD);
@@ -677,7 +811,7 @@ extern int    GetPosFile(int,fpos_t *);
 extern int    SetPosFile(int,fpos_t *);
 extern VOID   SynchFile(int);
 extern VOID   TruncateFile(int);
-extern int    GetChannel(char *);
+extern int    GetChannel(char *,int);
 extern int    GetAppendChannel(char *);
 extern int    CloseChannel(char *);
 extern VOID   inictable(VOID);
@@ -702,10 +836,12 @@ extern int    StrCmp(UBYTE *,UBYTE *);
 extern int    StrICmp(UBYTE *,UBYTE *);
 extern int    StrHICmp(UBYTE *,UBYTE *);
 extern int    StrICont(UBYTE *,UBYTE *);
+extern int    CmpArray(WORD *,WORD *,WORD);
 extern int    ConWord(UBYTE *,UBYTE *);
 extern int    StrLen(UBYTE *);
 extern UBYTE *GetPreVar(UBYTE *,int);
 extern void   ToGeneral(WORD *,WORD *,WORD);
+extern WORD   ToPolyFunGeneral(PHEAD WORD *);
 extern int    ToFast(WORD *,WORD *);
 extern SETUPPARAMETERS *GetSetupPar(UBYTE *);
 extern int    RecalcSetups(VOID);
@@ -734,6 +870,8 @@ extern int    TheUndefine(UBYTE *);
 extern int    ClearMacro(UBYTE *);
 extern int    DoUndefine(UBYTE *);
 extern int    DoInclude(UBYTE *);
+extern int    DoReverseInclude(UBYTE *);
+extern int    Include(UBYTE *,int);
 /*[14apr2004 mt]:*/
 extern int    DoExternal(UBYTE *);
 extern int    DoToExternal(UBYTE *);
@@ -750,6 +888,8 @@ extern WORD   GetDollarNumber(UBYTE **,DOLLARS);
 extern int    DoSetRandom(UBYTE *);
 extern int    DoOptimize(UBYTE *);
 extern int    DoClearOptimize(UBYTE *);
+extern int    DoSkipExtraSymbols(UBYTE *);
+extern int    DoTimeOutAfter(UBYTE *);
 extern int    DoMessage(UBYTE *);
 extern int    DoPreOut(UBYTE *);
 extern int    DoPreAppend(UBYTE *);
@@ -766,6 +906,8 @@ extern int    DoSystem(UBYTE *);
 extern int    DoPipe(UBYTE *);
 extern VOID   StartPrepro(VOID);
 extern int    DoIfdef(UBYTE *,int);
+extern int    DoIfydef(UBYTE *);
+extern int    DoIfndef(UBYTE *);
 extern int    DoElse(UBYTE *);
 extern int    DoElseif(UBYTE *);
 extern int    DoEndif(UBYTE *);
@@ -786,8 +928,10 @@ extern int    DoPreClose(UBYTE *);
 extern int    DoPreRemove(UBYTE *);
 extern int    DoCommentChar(UBYTE *);
 extern int    DoPrcExtension(UBYTE *);
+extern int    DoPreReset(UBYTE *);
 extern VOID   WriteString(int,UBYTE *,int);
 extern VOID   WriteUnfinString(int,UBYTE *,int);
+extern UBYTE *AddToString(UBYTE *,UBYTE *,int);
 extern UBYTE *PreCalc(VOID);
 extern UBYTE *PreEval(UBYTE *,LONG *);
 extern VOID   NumToStr(UBYTE *,LONG);
@@ -822,6 +966,7 @@ extern VOID   PrintTerm(WORD *,char *);
 extern VOID   PrintTermC(WORD *,char *);
 extern VOID   PrintSubTerm(WORD *,char *);
 extern VOID   PrintWords(WORD *,LONG);
+extern void   PrintSeq(WORD *,char *);
 extern int    ExpandTripleDots(int);
 extern LONG   ComPress(WORD **,LONG *);
 extern VOID   StageSort(FILEHANDLE *);
@@ -837,11 +982,11 @@ extern void   ResetVariables(int);
 extern void   AddToPreTypes(int);
 extern void   MessPreNesting(int);
 extern LONG   GetStreamPosition(STREAM *);
-extern WORD  *DoubleCbuffer(int,WORD *);
+extern WORD  *DoubleCbuffer(int,WORD *,int);
 extern WORD  *AddLHS(int);
 extern WORD  *AddRHS(int,int);
 extern int    AddNtoL(int,WORD *);
-extern int    AddNtoC(int,int,WORD *);
+extern int    AddNtoC(int,int,WORD *,int);
 extern VOID   DoubleIfBuffers(VOID);
 extern STREAM *CreateStream(UBYTE *);
 
@@ -994,6 +1139,7 @@ extern int    CoTable(UBYTE *);
 extern int    CoTerm(UBYTE *);
 extern int    CoNTable(UBYTE *);
 extern int    CoCTable(UBYTE *);
+extern void   EmptyTable(TABLES);
 extern int    CoToTensor(UBYTE *);
 extern int    CoToVector(UBYTE *);
 extern int    CoTrace4(UBYTE *);
@@ -1009,6 +1155,11 @@ extern int    CoVector(UBYTE *);
 extern int    CoWhile(UBYTE *);
 extern int    CoWrite(UBYTE *);
 extern int    CoAuto(UBYTE *);
+extern int    CoSwitch(UBYTE *);
+extern int    CoCase(UBYTE *);
+extern int    CoBreak(UBYTE *);
+extern int    CoDefault(UBYTE *);
+extern int    CoEndSwitch(UBYTE *);
 extern int    CoTBaddto(UBYTE *);
 extern int    CoTBaudit(UBYTE *);
 extern int    CoTBcleanup(UBYTE *);
@@ -1058,8 +1209,9 @@ extern int    AssignDollar(PHEAD WORD *,WORD);
 extern UBYTE *WriteDollarToBuffer(WORD,WORD);
 extern UBYTE *WriteDollarFactorToBuffer(WORD,WORD,WORD);
 extern void   AddToDollarBuffer(UBYTE *);
+extern int    PutTermInDollar(WORD *,WORD);
 extern void   TermAssign(WORD *);
-extern void   WildDollars(PHEAD0);
+extern void   WildDollars(PHEAD WORD *);
 extern LONG   numcommute(WORD *,LONG *);
 extern int    FullRenumber(PHEAD WORD *,WORD);
 extern int    Lus(WORD *,WORD,WORD,WORD,WORD,WORD);
@@ -1119,14 +1271,21 @@ extern int    FindTableTree(TABLES,WORD *,int);
 extern void   finishcbuf(WORD);
 extern void   clearcbuf(WORD);
 extern void   CleanUpSort(int);
-extern FILEHANDLE *AllocFileHandle(VOID);
+extern FILEHANDLE *AllocFileHandle(WORD,char *);
 extern VOID   DeAllocFileHandle(FILEHANDLE *);
 extern VOID   LowerSortLevel(VOID);
+extern WORD  *PolyRatFunSpecial(PHEAD WORD *, WORD *);
+extern VOID   SimpleSplitMergeRec(WORD *,WORD,WORD *);
+extern VOID   SimpleSplitMerge(WORD *,WORD);
+extern WORD   BinarySearch(WORD *,WORD,WORD);
 extern int    InsideDollar(PHEAD WORD *,WORD);
 extern DOLLARS DolToTerms(PHEAD WORD);
 extern WORD   EvalDoLoopArg(PHEAD WORD *,WORD);
 extern int    SetExprCases(int,int,int);
 extern int    TestSelect(WORD *,WORD *);
+extern VOID   SubsInAll(PHEAD0);
+extern VOID   TransferBuffer(int,int,int);
+extern int    TakeIDfunction(PHEAD WORD *);
 extern int    MakeSetupAllocs(VOID);
 extern int    TryFileSetups(VOID);
 extern void   ExchangeExpressions(int,int);
@@ -1138,7 +1297,9 @@ extern int    CleanupTerm(WORD *);
 extern WORD   ContentMerge(PHEAD WORD *,WORD *);
 extern UBYTE *PreIfDollarEval(UBYTE *,int *);
 extern LONG   TermsInDollar(WORD);
+extern LONG   SizeOfDollar(WORD);
 extern LONG   TermsInExpression(WORD);
+extern LONG   SizeOfExpression(WORD);
 extern WORD   *TranslateExpression(UBYTE *);
 extern int    IsSetMember(WORD *,WORD);
 extern int    IsMultipleOf(WORD *,WORD *);
@@ -1147,7 +1308,7 @@ extern void   UpdatePositions(VOID);
 extern void   M_check(VOID);
 extern void   M_print(VOID);
 extern void   M_check1(VOID);
-extern void   PrintTime(VOID);
+extern void   PrintTime(UBYTE *);
 
 extern POSITION *FindBracket(WORD,WORD *);
 extern VOID   PutBracketInIndex(PHEAD WORD *,POSITION *);
@@ -1212,6 +1373,7 @@ extern int    PutOutputGZIP(FILEHANDLE *);
 extern int    FlushOutputGZIP(FILEHANDLE *);
 extern int    SetupAllInputGZIP(SORTING *);
 extern LONG   FillInputGZIP(FILEHANDLE *,POSITION *,UBYTE *,LONG,int);
+extern void   ClearSortGZIP(FILEHANDLE *f);
 #endif
 
 #ifdef WITHPTHREADS
@@ -1226,7 +1388,7 @@ extern int    BalanceRunThread(PHEAD int,WORD *,WORD);
 extern void   WakeupThread(int,int);
 extern int    MasterWait(VOID);
 extern int    InParallelProcessor(VOID);
-extern int    ThreadsProcessor(EXPRESSIONS,WORD);
+extern int    ThreadsProcessor(EXPRESSIONS,WORD,WORD);
 extern int    MasterMerge(VOID);
 extern int    PutToMaster(PHEAD WORD *);
 extern void   SetWorkerFiles(VOID);
@@ -1286,13 +1448,10 @@ typedef int (*GETCFROMEXTCHANNEL)(VOID);
 typedef int (*SETTERMINATORFOREXTERNALCHANNEL)(char *);
 typedef int (*SETKILLMODEFOREXTERNALCHANNEL)(int,int);
 typedef LONG (*WRITEFILE)(int,UBYTE *,LONG);
-typedef WORD (*COMPARE)(PHEAD WORD *,WORD *,WORD);
 typedef WORD (*GETTERM)(PHEAD WORD *);
-typedef WORD (*FINISHUFFLE)(PHEAD WORD *);
-typedef WORD (*DO_UFFLE)(PHEAD WORD *,WORD,WORD,WORD);
 
 #define CompareTerms ((COMPARE)AR.CompareRoutine)
-#define FiniShuffle ((FINISHUFFLE)AN.SHvar.finishuf)
+#define FiniShuffle AN.SHvar.finishuf
 #define DoShtuffle ((DO_UFFLE)AN.SHvar.do_uffle)
 
 extern UBYTE *defineChannel(UBYTE*, HANDLERS*);
@@ -1312,8 +1471,11 @@ extern int    writeBufToExtChannelFailure(char *,size_t);
 extern int    ReleaseTB(VOID);
 
 extern int    SymbolNormalize(WORD *);
-extern int    CompareSymbols(PHEAD WORD *,WORD *,WORD);
+extern int    TestFunFlag(PHEAD WORD *);
+extern WORD   CompareSymbols(WORD *,WORD *,WORD);
+extern WORD   CompareHSymbols(WORD *,WORD *,WORD);
 extern WORD   NextPrime(PHEAD WORD);
+extern WORD   Moebius(PHEAD WORD);
 extern UWORD  wranf(PHEAD0);
 extern UWORD  iranf(PHEAD UWORD);
 extern void   iniwranf(PHEAD0);
@@ -1321,6 +1483,7 @@ extern UBYTE *PreRandom(UBYTE *);
 
 extern WORD *PolyNormPoly (PHEAD WORD);
 extern WORD  *EvaluateGcd(PHEAD WORD *);
+extern int TreatPolyRatFun(PHEAD WORD *);
 
 extern WORD   ReadSaveHeader(VOID);
 extern WORD   ReadSaveIndex(FILEINDEX *);
@@ -1337,12 +1500,15 @@ extern int    DoRecovery(int *);
 extern void   DoCheckpoint(int);
 
 extern VOID NumberMallocAddMemory(PHEAD0);
+extern VOID CacheNumberMallocAddMemory(PHEAD0);
 extern VOID TermMallocAddMemory(PHEAD0);
 #ifndef MEMORYMACROS
 extern WORD *TermMalloc2(PHEAD char *text);
 extern VOID TermFree2(PHEAD WORD *term,char *text);
 extern UWORD *NumberMalloc2(PHEAD char *text);
+extern UWORD *CacheNumberMalloc2(PHEAD char *text);
 extern VOID NumberFree2(PHEAD UWORD *NumberMem,char *text);
+extern VOID CacheNumberFree2(PHEAD UWORD *NumberMem,char *text);
 #endif
 
 extern void ExprStatus(EXPRESSIONS);
@@ -1358,11 +1524,17 @@ extern WORD RunExplode(PHEAD WORD *fun, WORD *args);
 extern int TestArgNum(int n, int totarg, WORD *args);
 extern WORD PutArgInScratch(WORD *arg,UWORD *scrat);
 extern UBYTE *ReadRange(UBYTE *s, WORD *out, int par);
+extern int  FindRange(PHEAD WORD *,WORD *,WORD *,WORD);
 extern WORD RunPermute(PHEAD WORD *fun, WORD *args, WORD *info);
 extern WORD RunReverse(PHEAD WORD *fun, WORD *args);
 extern WORD RunCycle(PHEAD WORD *fun, WORD *args, WORD *info);
+extern WORD RunAddArg(PHEAD WORD *fun, WORD *args);
+extern WORD RunMulArg(PHEAD WORD *fun, WORD *args);
 extern WORD RunIsLyndon(PHEAD WORD *fun, WORD *args, int par);
 extern WORD RunToLyndon(PHEAD WORD *fun, WORD *args, int par);
+extern WORD RunDropArg(PHEAD WORD *fun, WORD *args);
+extern WORD RunSelectArg(PHEAD WORD *fun, WORD *args);
+extern WORD RunDedup(PHEAD WORD *fun, WORD *args);
 
 extern int NormPolyTerm(PHEAD WORD *);
 extern WORD ComparePoly(WORD *, WORD *, WORD);
@@ -1379,39 +1551,57 @@ extern void UpdateMaxSize(VOID);
 
 extern int CoToPolynomial(UBYTE *);
 extern int CoFromPolynomial(UBYTE *);
+extern int CoArgToExtraSymbol(UBYTE *);
 extern int CoExtraSymbols(UBYTE *);
 extern UBYTE *GetDoParam(UBYTE *, WORD **, int);
 extern WORD *GetIfDollarFactor(UBYTE **, WORD *);
 extern int CoDo(UBYTE *);
 extern int CoEndDo(UBYTE *);
 extern int ExtraSymFun(PHEAD WORD *,WORD);
+extern int PruneExtraSymbols(WORD);
 extern int IniFbuffer(WORD);
 extern void IniFbufs(VOID);
 extern int GCDfunction(PHEAD WORD *,WORD);
 extern WORD *GCDfunction3(PHEAD WORD *,WORD *);
+extern WORD *GCDfunction4(PHEAD WORD *,WORD *);
+extern int ReadPolyRatFun(PHEAD WORD *);
+extern int FromPolyRatFun(PHEAD WORD *, WORD **, WORD **);
+extern void PRFnormalize(PHEAD WORD *);
+extern WORD *PRFadd(PHEAD WORD *, WORD *);
+extern WORD *PolyDiv(PHEAD WORD *,WORD *,char *);
+extern WORD *PolyGCD(PHEAD WORD *,WORD *);
+extern WORD *PolyAdd(PHEAD WORD *,WORD *);
+extern void GCDclean(PHEAD WORD *, WORD *);
+extern int RatFunNormalize(PHEAD WORD *);
+extern WORD *TakeSymbolContent(PHEAD WORD *,WORD *);
 extern int GCDterms(PHEAD WORD *,WORD *,WORD *);
 extern WORD *PutExtraSymbols(PHEAD WORD *,WORD,int *);
 extern WORD *TakeExtraSymbols(PHEAD WORD *,WORD);
-extern WORD *MultiplyWithTerm(PHEAD WORD *, WORD *);
+extern WORD *MultiplyWithTerm(PHEAD WORD *, WORD *,WORD);
 extern WORD *TakeContent(PHEAD WORD *, WORD *);
 extern int MergeSymbolLists(PHEAD WORD *, WORD *, int);
 extern int MergeDotproductLists(PHEAD WORD *, WORD *, int);
 extern WORD *CreateExpression(PHEAD WORD);
 extern int DIVfunction(PHEAD WORD *,WORD,int);
+extern WORD *MULfunc(PHEAD WORD *, WORD *);
 extern WORD *ConvertArgument(PHEAD WORD *,int *);
+extern int ExpandRat(PHEAD WORD *);
+extern int InvPoly(PHEAD WORD *,WORD,WORD);
 extern WORD TestDoLoop(PHEAD WORD *,WORD);
 extern WORD TestEndDoLoop(PHEAD WORD *,WORD);
 
-extern WORD  *poly_gcd(PHEAD WORD *, WORD *);
-extern WORD  *poly_div(PHEAD WORD *, WORD *);
-extern WORD  *poly_rem(PHEAD WORD *, WORD *);
+extern WORD  *poly_gcd(PHEAD WORD *, WORD *, WORD);
+extern WORD  *poly_div(PHEAD WORD *, WORD *, WORD);
+extern WORD  *poly_rem(PHEAD WORD *, WORD *, WORD);
 extern WORD  *poly_inverse(PHEAD WORD *, WORD *);
+extern WORD  *poly_mul(PHEAD WORD *, WORD *);
 extern WORD *poly_ratfun_add(PHEAD WORD *, WORD *);
 extern int   poly_ratfun_normalize(PHEAD WORD *);
 extern int   poly_factorize_argument(PHEAD WORD *, WORD *);
 extern WORD *poly_factorize_dollar(PHEAD WORD *);
 extern int   poly_factorize_expression(EXPRESSIONS);
 extern int   poly_unfactorize_expression(EXPRESSIONS);
+extern void  poly_free_poly_vars(PHEAD const char *);
 
 extern VOID optimize_print_code (int);
 
@@ -1421,6 +1611,62 @@ extern void find_Horner_MCTS_expand_tree_threaded();
 extern void optimize_expression_given_Horner();
 extern void optimize_expression_given_Horner_threaded();
 #endif
+ 
+extern int DoPreAdd(UBYTE *s);
+extern int DoPreUseDictionary(UBYTE *s);
+extern int DoPreCloseDictionary(UBYTE *s);
+extern int DoPreOpenDictionary(UBYTE *s);
+extern void RemoveDictionary(DICTIONARY *dict);
+extern void UnSetDictionary(VOID);
+extern int SetDictionaryOptions(UBYTE *options);
+extern int SelectDictionary(UBYTE *name,UBYTE *options);
+extern int AddToDictionary(DICTIONARY *dict,UBYTE *left,UBYTE *right);
+extern int AddDictionary(UBYTE *name);
+extern int FindDictionary(UBYTE *name);
+extern UBYTE *IsExponentSign(VOID);
+extern UBYTE *IsMultiplySign(VOID);
+extern VOID TransformRational(UWORD *a, WORD na);
+extern void WriteDictionary(DICTIONARY *);
+extern void ShrinkDictionary(DICTIONARY *);
+extern void MultiplyToLine(VOID);
+extern UBYTE *FindSymbol(WORD num);
+extern UBYTE *FindVector(WORD num);
+extern UBYTE *FindIndex(WORD num);
+extern UBYTE *FindFunction(WORD num);
+extern UBYTE *FindFunWithArgs(WORD *t);
+extern UBYTE *FindExtraSymbol(WORD num);
+extern LONG DictToBytes(DICTIONARY *dict,UBYTE *buf);
+extern DICTIONARY *DictFromBytes(UBYTE *buf);
+extern int CoCreateSpectator(UBYTE *inp);
+extern int CoToSpectator(UBYTE *inp);
+extern int CoRemoveSpectator(UBYTE *inp);
+extern int CoEmptySpectator(UBYTE *inp);
+extern int CoCopySpectator(UBYTE *inp);
+extern int PutInSpectator(WORD *,WORD);
+extern void ClearSpectators(WORD);
+extern WORD GetFromSpectator(WORD *,WORD);
+extern void FlushSpectators(VOID);
+
+extern WORD *PreGCD(PHEAD WORD *, WORD *,int);
+extern WORD *FindCommonVariables(PHEAD int,int);
+extern VOID AddToSymbolList(PHEAD WORD);
+extern int AddToListPoly(PHEAD0);
+extern int InvPoly(PHEAD WORD *,WORD,WORD);
+
+extern int ReadFromScratch(FILEHANDLE *,POSITION *,UBYTE *,POSITION *);
+extern int AddToScratch(FILEHANDLE *,POSITION *,UBYTE *,POSITION *,int);
+
+extern int DoPreAppendPath(UBYTE *);
+extern int DoPrePrependPath(UBYTE *);
+
+extern int DoSwitch(PHEAD WORD *, WORD *);
+extern int DoEndSwitch(PHEAD WORD *, WORD *);
+extern SWITCHTABLE *FindCase(WORD, WORD);
+extern VOID SwitchSplitMergeRec(SWITCHTABLE *, WORD, SWITCHTABLE *);
+extern VOID SwitchSplitMerge(SWITCHTABLE *, WORD);
+extern int DoubleSwitchBuffers(VOID);
+
+extern int DistrN(int, int *, int, int *);
 
 /*
   	#] Declarations : 
