@@ -5,7 +5,7 @@
  */
 /* #[ License : */
 /*
- *   Copyright (C) 1984-2013 J.A.M. Vermaseren
+ *   Copyright (C) 1984-2023 J.A.M. Vermaseren
  *   When using this file you are requested to refer to the publication
  *   J.A.M.Vermaseren "New features of FORM" math-ph/0010025
  *   This is considered a matter of courtesy as the development was paid
@@ -70,6 +70,7 @@ WORD CleanExpr(WORD par)
 		if ( e_in->status == HIDDENLEXPRESSION
 		|| e_in->status == HIDDENGEXPRESSION ) numhid++;
 		switch ( e_in->status ) {
+			case SPECTATOREXPRESSION:
 			case LOCALEXPRESSION:
 			case HIDDENLEXPRESSION:
 				if ( par ) {
@@ -79,6 +80,7 @@ WORD CleanExpr(WORD par)
 						ClearBracketIndex(e_in-Expressions);
 					break;
 				}
+				/* fall through */
 			case GLOBALEXPRESSION:
 			case HIDDENGEXPRESSION:
 				if ( par ) {
@@ -134,18 +136,20 @@ WORD CleanExpr(WORD par)
 					if ( e_in->status != HIDDENGEXPRESSION )
 						ClearBracketIndex(e_in-Expressions);
 				}
-				/* Fall through is intentional */
+				/* fall through */
 			case SKIPLEXPRESSION:
 			case DROPLEXPRESSION:
 			case DROPHLEXPRESSION:
 			case DROPGEXPRESSION:
 			case DROPHGEXPRESSION:
 			case STOREDEXPRESSION:
+			case DROPSPECTATOREXPRESSION:
 				if ( e_out != e_in ) {
 					node = AC.exprnames->namenode + e_in->node;
 					node->number = e_out - Expressions;
 
 					e_out->onfile = e_in->onfile;
+					e_out->size = e_in->size;
 					e_out->printflag = 0;
 					if ( par ) e_out->status = STOREDEXPRESSION;
 					else e_out->status = e_in->status;
@@ -161,6 +165,7 @@ WORD CleanExpr(WORD par)
 					e_out->numdummies = e_in->numdummies;
 					e_out->numfactors = e_in->numfactors;
 					e_out->vflags = e_in->vflags;
+					e_out->sizeprototype = e_in->sizeprototype;
 				}
 #ifdef PARALLELCODE
 				e_out->partodo = 0;
@@ -189,6 +194,7 @@ WORD CleanExpr(WORD par)
 		AR.hidefile->POfill = AR.hidefile->PObuffer;
 		PUTZERO(AR.hidefile->POposition);
 	}
+	FlushSpectators();
 	return(0);
 }
 
@@ -266,19 +272,28 @@ WORD PopVariables()
 			MakeInverses();
 	AC.funpowers = AM.gfunpowers;
 	AC.lPolyFun = AM.gPolyFun;
+	AC.lPolyFunInv = AM.gPolyFunInv;
 	AC.lPolyFunType = AM.gPolyFunType;
+	AC.lPolyFunExp = AM.gPolyFunExp;
+	AR.PolyFunVar = AC.lPolyFunVar = AM.gPolyFunVar;
+	AC.lPolyFunPow = AM.gPolyFunPow;
 	AC.parallelflag = AM.gparallelflag;
 	AC.ProcessBucketSize = AC.mProcessBucketSize = AM.gProcessBucketSize;
 	AC.properorderflag = AM.gproperorderflag;
     AC.ThreadBucketSize = AM.gThreadBucketSize;
 	AC.ThreadStats = AM.gThreadStats;
 	AC.FinalStats = AM.gFinalStats;
+	AC.OldGCDflag = AM.gOldGCDflag;
+	AC.WTimeStatsFlag = AM.gWTimeStatsFlag;
 	AC.ThreadsFlag = AM.gThreadsFlag;
 	AC.ThreadBalancing = AM.gThreadBalancing;
 	AC.ThreadSortFileSynch = AM.gThreadSortFileSynch;
 	AC.ProcessStats = AM.gProcessStats;
 	AC.OldParallelStats = AM.gOldParallelStats;
 	AC.IsFortran90 = AM.gIsFortran90;
+	AC.SizeCommuteInSet = AM.gSizeCommuteInSet;
+	PruneExtraSymbols(AM.gnumextrasym);
+
 	if ( AC.Fortran90Kind ) {
 		M_free(AC.Fortran90Kind,"Fortran90 Kind");
 		AC.Fortran90Kind = 0;
@@ -317,6 +332,40 @@ WORD PopVariables()
 	AC.OutNumberType = AM.gOutNumberType;
 	AR.SortType = AC.SortType = AM.gSortType;
 	AC.ShortStatsMax = AM.gShortStatsMax;
+/*
+	Now we have to clean up the commutation properties
+*/
+	for ( i = 0; i < NumFunctions; i++ ) functions[i].flags &= ~COULDCOMMUTE;
+	if ( AC.CommuteInSet ) {
+		WORD *g, *gg;
+		g = AC.CommuteInSet;
+		while ( *g ) {
+			gg = g+1; g += *g;
+			while ( gg < g ) {
+				if ( *gg <= GAMMASEVEN && *gg >= GAMMA ) {
+					functions[GAMMA-FUNCTION].flags |= COULDCOMMUTE;
+					functions[GAMMAI-FUNCTION].flags |= COULDCOMMUTE;
+					functions[GAMMAFIVE-FUNCTION].flags |= COULDCOMMUTE;
+					functions[GAMMASIX-FUNCTION].flags |= COULDCOMMUTE;
+					functions[GAMMASEVEN-FUNCTION].flags |= COULDCOMMUTE;
+				}
+				else {
+					functions[*gg-FUNCTION].flags |= COULDCOMMUTE;
+				}
+			}
+		}
+	}
+/*
+	Clean up the dictionaries.
+*/
+	for ( i = AO.NumDictionaries-1; i >= AO.gNumDictionaries; i-- ) {
+		RemoveDictionary(AO.Dictionaries[i]);
+		M_free(AO.Dictionaries[i],"Dictionary");
+	}
+	for( ; i >= 0; i-- ) {
+		ShrinkDictionary(AO.Dictionaries[i]);
+	}
+	AO.NumDictionaries = AO.gNumDictionaries;
 	return(retval);
 }
 
@@ -356,19 +405,27 @@ VOID MakeGlobal()
 	AM.gOutNumberType = AC.OutNumberType;
 	AM.gfunpowers = AC.funpowers;
 	AM.gPolyFun = AC.lPolyFun;
+	AM.gPolyFunInv = AC.lPolyFunInv;
 	AM.gPolyFunType = AC.lPolyFunType;
+	AM.gPolyFunExp = AC.lPolyFunExp;
+	AM.gPolyFunVar = AC.lPolyFunVar;
+	AM.gPolyFunPow = AC.lPolyFunPow;
 	AM.gparallelflag = AC.parallelflag;
 	AM.gProcessBucketSize = AC.ProcessBucketSize;
 	AM.gproperorderflag = AC.properorderflag;
     AM.gThreadBucketSize = AC.ThreadBucketSize;
 	AM.gThreadStats = AC.ThreadStats;
 	AM.gFinalStats = AC.FinalStats;
+	AM.gOldGCDflag = AC.OldGCDflag;
+	AM.gWTimeStatsFlag = AC.WTimeStatsFlag;
 	AM.gThreadsFlag = AC.ThreadsFlag;
 	AM.gThreadBalancing = AC.ThreadBalancing;
 	AM.gThreadSortFileSynch = AC.ThreadSortFileSynch;
 	AM.gProcessStats = AC.ProcessStats;
 	AM.gOldParallelStats = AC.OldParallelStats;
 	AM.gIsFortran90 = AC.IsFortran90;
+	AM.gSizeCommuteInSet = AC.SizeCommuteInSet;
+	AM.gnumextrasym = (cbuf+AM.sbufnum)->numrhs;
 	if ( AM.gFortran90Kind ) {
 		M_free(AM.gFortran90Kind,"Fortran 90 Kind");
 		AM.gFortran90Kind = 0;
@@ -390,6 +447,23 @@ VOID MakeGlobal()
 	NCOPY(pp,mm,i);
 	AM.gSortType = AC.SortType;
 	AM.gShortStatsMax = AC.ShortStatsMax;
+
+	if ( AO.CurrentDictionary > 0 || AP.OpenDictionary > 0 ) {
+		Warning("You cannot have an open or selected dictionary at a .global. Dictionary closed.");
+		AP.OpenDictionary = 0;
+		AO.CurrentDictionary = 0;
+	}
+
+	AO.gNumDictionaries = AO.NumDictionaries;
+	for ( i = 0; i < AO.NumDictionaries; i++ ) {
+		AO.Dictionaries[i]->gnumelements = AO.Dictionaries[i]->numelements;
+	}
+	if ( AM.NumSpectatorFiles > 0 ) {
+		for ( i = 0; i < AM.SizeForSpectatorFiles; i++ ) {
+			if ( AM.SpectatorFiles[i].name != 0 )
+					AM.SpectatorFiles[i].flags |= GLOBALSPECTATORFLAG;
+		}
+	}
 }
 
 /*
@@ -429,6 +503,7 @@ VOID TestDrop()
 			case DROPGEXPRESSION:
 			case DROPHLEXPRESSION:
 			case DROPHGEXPRESSION:
+			case DROPSPECTATOREXPRESSION:
 				e->status = DROPPEDEXPRESSION;
 				ClearBracketIndex(j);
 				e->bracketinfo = e->newbracketinfo; e->newbracketinfo = 0;
@@ -451,13 +526,18 @@ VOID TestDrop()
 			case HIDDENGEXPRESSION:
 				break;
 			case INTOHIDELEXPRESSION:
+				ClearBracketIndex(j);
+				e->bracketinfo = e->newbracketinfo; e->newbracketinfo = 0;
 				e->status = HIDDENLEXPRESSION;
 				break;
 			case INTOHIDEGEXPRESSION:
+				ClearBracketIndex(j);
+				e->bracketinfo = e->newbracketinfo; e->newbracketinfo = 0;
 				e->status = HIDDENGEXPRESSION;
 				break;
 			default:
 				ClearBracketIndex(j);
+				e->bracketinfo = 0;
 				break;
 		}
 		if ( e->replace == NEWLYDEFINEDEXPRESSION ) e->replace = REGULAREXPRESSION;
@@ -485,7 +565,7 @@ restart:;
 		AS.OldOnFile = (POSITION *)Malloc1(2*AS.NumOldOnFile*sizeof(POSITION),"file pointers");
 		for ( i = 0; i < AS.NumOldOnFile; i++ ) AS.OldOnFile[i] = old[i];
 		AS.NumOldOnFile = 2*AS.NumOldOnFile;
-		M_free(old,"proces file pointers");
+		M_free(old,"process file pointers");
 	}
 	if ( AS.OldNumFactors == 0 ) {
 		AS.NumOldNumFactors = 20;
@@ -560,49 +640,25 @@ WORD DoExecute(WORD par, WORD skip)
 			}
 		}
 	}
+	if ( AC.SwitchLevel > 0 ) {
+		MesPrint(" %d endswitch statement(s) missing",AC.SwitchLevel);
+		RetCode = 1;
+	}
 	if ( AC.dolooplevel > 0 ) {
 		MesPrint(" %d enddo statement(s) missing",AC.dolooplevel);
 		RetCode = 1;
 	}
+	if ( AP.OpenDictionary > 0 ) {
+		MesPrint(" Dictionary %s has not been closed.",
+			AO.Dictionaries[AP.OpenDictionary-1]->name);
+		AP.OpenDictionary = 0;
+		RetCode = 1;
+	}
 	if ( RetCode ) return(RetCode);
 	AR.Cnumlhs = cbuf[AM.rbufnum].numlhs;
 
 	if ( ( AS.ExecMode = par ) == GLOBALMODULE ) AS.ExecMode = 0;
-/*
-	Now we compare whether all elements of PotModdollars are contained in
-	ModOptdollars. If not, we may not run parallel.
-*/
-#ifdef WITHMPI
-	if ( NumPotModdollars > 0 && AC.mparallelflag == PARALLELFLAG ) {
-	  if ( NumPotModdollars > NumModOptdollars ) 
-		AC.mparallelflag |= NOPARALLEL_DOLLAR;
-	  else 
-		for ( i = 0; i < NumPotModdollars; i++ ) {
-		  for ( j = 0; j < NumModOptdollars; j++ ) 
-			if ( PotModdollars[i] == ModOptdollars[j].number ) break;
-		  if ( j >= NumModOptdollars ) {
-			AC.parallelflag |= NOPARALLEL_DOLLAR;
-			break;
-		  }
-		}
-	}
-	/*
-	 * Set $-variables with MODSUM to zero on the slaves.
-	 */
-	if ( AC.mparallelflag == PARALLELFLAG && PF.me != MASTER ) {
-		for ( i = 0; i < NumModOptdollars; i++ ) {
-			if ( ModOptdollars[i].type == MODSUM ) {
-				DOLLARS d = Dollars + ModOptdollars[i].number;
-				d->type = DOLZERO;
-				if ( d->where && d->where != &AM.dollarzero ) M_free(d->where, "old content of dollar");
-				d->where = &AM.dollarzero;
-				d->size = 0;
-				CleanDollarFactors(d);
-			}
-		}
-	}
-#endif
-#ifdef WITHPTHREADS
+#ifdef PARALLELCODE
 /*
 		Now check whether we have either the regular parallel flag or the
 		mparallel flag set.
@@ -615,10 +671,17 @@ WORD DoExecute(WORD par, WORD skip)
 			if ( Expressions[i].partodo ) { AC.partodoflag = 1; break; }
 		}
 	}
+#ifdef WITHMPI
+	if ( AC.partodoflag > 0 && PF.numtasks < 3 ) {
+		AC.partodoflag = 0;
+	}
+#endif
 	if ( AC.partodoflag > 0 || ( NumPotModdollars > 0 && AC.mparallelflag == PARALLELFLAG ) ) {
 	  if ( NumPotModdollars > NumModOptdollars ) {
 		AC.mparallelflag |= NOPARALLEL_DOLLAR;
+#ifdef WITHPTHREADS
 		AS.MultiThreaded = 0;
+#endif
 		AC.partodoflag = 0;
 	  }
 	  else {
@@ -627,7 +690,9 @@ WORD DoExecute(WORD par, WORD skip)
 			if ( PotModdollars[i] == ModOptdollars[j].number ) break;
 		  if ( j >= NumModOptdollars ) {
 			AC.mparallelflag |= NOPARALLEL_DOLLAR;
+#ifdef WITHPTHREADS
 			AS.MultiThreaded = 0;
+#endif
 			AC.partodoflag = 0;
 			break;
 		  }
@@ -647,7 +712,9 @@ WORD DoExecute(WORD par, WORD skip)
 	  }
 	}
 	else if ( ( AC.mparallelflag & NOPARALLEL_USER ) != 0 ) {
+#ifdef WITHPTHREADS
 		AS.MultiThreaded = 0;
+#endif
 		AC.partodoflag = 0;
 	}
 	if ( AC.partodoflag == 0 ) {
@@ -660,8 +727,10 @@ WORD DoExecute(WORD par, WORD skip)
 	}
 #endif
 #ifdef WITHMPI
-/*[20oct2009 mt]:*/
-	if ( AC.RhsExprInModuleFlag && AC.mparallelflag == PARALLELFLAG ) {
+	/*
+	 * Check RHS expressions.
+	 */
+	if ( AC.RhsExprInModuleFlag && (AC.mparallelflag == PARALLELFLAG || AC.partodoflag) ) {
 		if (PF.rhsInParallel) {
 			PF.mkSlaveInfile=1;
 			if(PF.me != MASTER){
@@ -674,9 +743,27 @@ WORD DoExecute(WORD par, WORD skip)
 		} 
 		else {
 			AC.mparallelflag |= NOPARALLEL_RHS;
+			AC.partodoflag = 0;
+			for ( i = 0; i < NumExpressions; i++ ) {
+				Expressions[i].partodo = 0;
+			}
+		}
+	}
+	/*
+	 * Set $-variables with MODSUM to zero on the slaves.
+	 */
+	if ( (AC.mparallelflag == PARALLELFLAG || AC.partodoflag) && PF.me != MASTER ) {
+		for ( i = 0; i < NumModOptdollars; i++ ) {
+			if ( ModOptdollars[i].type == MODSUM ) {
+				DOLLARS d = Dollars + ModOptdollars[i].number;
+				d->type = DOLZERO;
+				if ( d->where && d->where != &AM.dollarzero ) M_free(d->where, "old content of dollar");
+				d->where = &AM.dollarzero;
+				d->size = 0;
+				CleanDollarFactors(d);
+			}
 		}
 	}
-/*:[20oct2009 mt]*/
 #endif
 	AR.SortType = AC.SortType;
 #ifdef WITHMPI
@@ -688,20 +775,6 @@ WORD DoExecute(WORD par, WORD skip)
 	}
 	if ( par == GLOBALMODULE ) MakeGlobal();
 	if ( RevertScratch() ) return(-1);
-/*[20oct2009 mt]:*/
-#ifdef WITHMPI
-	AC.partodoflag = 0;
-	if ( PF.numtasks >= 3 ) {
-		for ( i = 0; i < NumExpressions; i++ ) {
-			if ( Expressions[i].partodo > 0 ) { AC.partodoflag = 1; break; }
-		}
-	}
-	else {
-		for ( i = 0; i < NumExpressions; i++ ) {
-			Expressions[i].partodo = 0;
-		}
-	}
-#endif
 	if ( AC.ncmod ) SetMods();
 /*
 	Warn if the module has to run in sequential mode due to some problems.
@@ -714,13 +787,42 @@ WORD DoExecute(WORD par, WORD skip)
 			/* The user switched off the parallel execution explicitly. */
 		}
 		else if ( AC.mparallelflag & NOPARALLEL_DOLLAR ) {
-			HighWarning("This module is forced to run in sequential mode due to $-variables");
+			if ( AC.WarnFlag >= 2 ) {  /* HighWarning */
+				int i, j, k, n;
+				UBYTE *s, *s1;
+				s = strDup1((UBYTE *)"","NOPARALLEL_DOLLAR s");
+				n = 0;
+				j = NumPotModdollars;
+				for ( i = 0; i < j; i++ ) {
+					for ( k = 0; k < NumModOptdollars; k++ )
+						if ( ModOptdollars[k].number == PotModdollars[i] ) break;
+					if ( k >= NumModOptdollars ) {
+						/* global $-variable */
+						if ( n > 0 )
+							s = AddToString(s,(UBYTE *)", ",0);
+						s = AddToString(s,(UBYTE *)"$",0);
+						s = AddToString(s,DOLLARNAME(Dollars,PotModdollars[i]),0);
+						n++;
+					}
+				}
+				s1 = strDup1((UBYTE *)"This module is forced to run in sequential mode due to $-variable","NOPARALLEL_DOLLAR s1");
+				if ( n != 1 )
+					s1 = AddToString(s1,(UBYTE *)"s",0);
+				s1 = AddToString(s1,(UBYTE *)": ",0);
+				s1 = AddToString(s1,s,0);
+				HighWarning((char *)s1);
+				M_free(s,"NOPARALLEL_DOLLAR s");
+				M_free(s1,"NOPARALLEL_DOLLAR s1");
+			}
 		}
 		else if ( AC.mparallelflag & NOPARALLEL_RHS ) {
 			HighWarning("This module is forced to run in sequential mode due to RHS expression names");
 		}
 		else if ( AC.mparallelflag & NOPARALLEL_CONVPOLY ) {
-			HighWarning("This module is forced to run in sequential mode due to topolynomial/frompolynomial");
+			HighWarning("This module is forced to run in sequential mode due to conversion to extra symbols");
+		}
+		else if ( AC.mparallelflag & NOPARALLEL_SPECTATOR ) {
+			HighWarning("This module is forced to run in sequential mode due to tospectator/copyspectator");
 		}
 		else if ( AC.mparallelflag & NOPARALLEL_TBLDOLLAR ) {
 			HighWarning("This module is forced to run in sequential mode due to $-variable assignments in tables");
@@ -837,6 +939,34 @@ WORD DoExecute(WORD par, WORD skip)
 	NumModOptdollars = 0;
 
 skipexec:
+/*
+	Clean up the switch information.
+	We keep the switch array and heap.
+*/
+if ( AC.SwitchInArray > 0 ) {
+	for ( i = 0; i < AC.SwitchInArray; i++ ) {
+		SWITCH *sw = AC.SwitchArray + i;
+		if ( sw->table ) M_free(sw->table,"Switch table");
+		sw->table = 0;
+		sw->defaultcase.ncase = 0;
+		sw->defaultcase.value = 0;
+		sw->defaultcase.compbuffer = 0;
+		sw->endswitch.ncase = 0;
+		sw->endswitch.value = 0;
+		sw->endswitch.compbuffer = 0;
+		sw->typetable = 0;
+		sw->maxcase = 0;
+		sw->mincase = 0;
+		sw->numcases = 0;
+		sw->tablesize = 0;
+		sw->caseoffset = 0;
+		sw->iflevel = 0;
+		sw->whilelevel = 0;
+		sw->nestingsum = 0;
+	}
+	AC.SwitchInArray = 0;
+	AC.SwitchLevel = 0;
+}
 #ifdef PARALLELCODE
 	AC.numpfirstnum = 0;
 #endif
@@ -881,8 +1011,11 @@ skipexec:
 			ResetVariables(2);
 			AM.gProcessBucketSize = AM.hProcessBucketSize;
 			AM.gparallelflag = PARALLELFLAG;
+			AM.gnumextrasym = AM.ggnumextrasym;
+			PruneExtraSymbols(AM.ggnumextrasym);
 			IniVars();
 		}
+		ClearSpectators(par);
 	}
 	else {
 		if ( CleanExpr(0) ) RetCode = -1;
@@ -934,7 +1067,7 @@ WORD PutBracket(PHEAD WORD *termin)
 	WORD *t2, *s1, *s2;
 	WORD *bStop, *bb, *bf, *tStop;
 	WORD *term1,*term2, *m1, *m2, *tStopa;
-	WORD *bbb = 0, *bind, *binst = 0, bwild = 0;
+	WORD *bbb = 0, *bind, *binst = 0, bwild = 0, *bss = 0, *bns = 0, bset = 0;
 	term1 = AT.WorkPointer+1;
 	term2 = (WORD *)(((UBYTE *)(term1)) + AM.MaxTer);
 	if ( ( (WORD *)(((UBYTE *)(term2)) + AM.MaxTer) ) > AT.WorkTop ) return(MesWork());
@@ -946,7 +1079,8 @@ WORD PutBracket(PHEAD WORD *termin)
 	}
 	b = AT.BrackBuf; bStop = b+*b; b++;
 	while ( b < bStop ) {
-		if ( *b == INDEX ) { bwild = 1; bbb = b+2; binst = b + b[1]; break; }
+		if ( *b == INDEX ) { bwild = 1; bbb = b+2; binst = b + b[1]; }
+		if ( *b == SETSET ) { bset = 1; bss = b+2; bns = b + b[1]; }
 		b += b[1];
 	}
 
@@ -968,6 +1102,23 @@ WORD PutBracket(PHEAD WORD *termin)
 				}
 				b += b[1];
 			}
+			if ( bset ) {
+				b = bss;
+				while ( b < bns ) {
+					if ( b[1] == CFUNCTION ) { /* Set of functions */
+						SETS set = Sets+b[0]; WORD i;
+						for ( i = set->first; i < set->last; i++ ) {
+							if ( SetElements[i] == *t ) {
+								lastfun = t;
+								while ( t < tStop && *t >= FUNCTION
+									&& functions[*t-FUNCTION].commute ) t += t[1];
+								goto NextNcom1;
+							}
+						}
+					}
+					b += 2;
+				}
+			}
 			if ( bwild && *t >= FUNCTION && functions[*t-FUNCTION].spec ) {
 				s1 = t + t[1];
 				s2 = t + FUNHEAD;
@@ -1007,6 +1158,21 @@ NextNcom1:
 				if ( *b == *t ) { lastfun = t + t[1]; goto NextNcom; }
 				b += b[1];
 			}
+			if ( bset ) {
+				b = bss;
+				while ( b < bns ) {
+					if ( b[1] == CFUNCTION ) { /* Set of functions */
+						SETS set = Sets+b[0]; WORD i;
+						for ( i = set->first; i < set->last; i++ ) {
+							if ( SetElements[i] == *t ) {
+								lastfun = t + t[1];
+								goto NextNcom;
+							}
+						}
+					}
+					b += 2;
+				}
+			}
 			if ( bwild && *t >= FUNCTION && functions[*t-FUNCTION].spec ) {
 				s1 = t + t[1];
 				s2 = t + FUNHEAD;
@@ -1036,12 +1202,24 @@ NextNcom:
 	}
 	bf = b;
 
-	while ( t < tStop && ( bf < bStop || bwild ) ) {
+	while ( t < tStop && ( bf < bStop || bwild || bset ) ) {
 		b = bf;
 		while ( b < bStop && *b != *t ) { b += b[1]; }
 		i = t[1];
 		if ( *t >= FUNCTION ) { /* We are in function territory */
 			if ( b < bStop && *b == *t ) goto FunBrac;
+			if ( bset ) {
+				b = bss;
+				while ( b < bns ) {
+					if ( b[1] == CFUNCTION ) { /* Set of functions */
+						SETS set = Sets+b[0]; WORD i;
+						for ( i = set->first; i < set->last; i++ ) {
+							if ( SetElements[i] == *t ) goto FunBrac;
+						}
+					}
+					b += 2;
+				}
+			}
 			if ( bwild && *t >= FUNCTION && functions[*t-FUNCTION].spec ) {
 				s1 = t + t[1];
 				s2 = t + FUNHEAD;
@@ -1081,7 +1259,51 @@ FunBrac:	NCOPY(t1,t,i);
 					if ( *bind == *t && bind < binst ) {
 						*t1++ = *t++;
 					}
+					else if ( bset ) {
+						WORD *b3 = bss;
+						while ( b3 < bns ) {
+							if ( b3[1] == CVECTOR ) {
+								SETS set = Sets+b3[0]; WORD i;
+								for ( i = set->first; i < set->last; i++ ) {
+									if ( SetElements[i] == *t ) {
+										*t1++ = *t++;
+										goto nextind;
+									}
+								}
+							}
+							b3 += 2;
+						}
+						*t2++ = *t++;
+					}
 					else *t2++ = *t++;
+nextind:;
+				}
+				m1[1] = WORDDIF(t1,m1);
+				if ( m1[1] == 2 ) t1 = m1;
+				m2[1] = WORDDIF(t2,m2);
+				if ( m2[1] == 2 ) t2 = m2;
+			}
+			else if ( bset ) {
+				m1 = t1; m2 = t2;
+				*t1++ = *t; t1++; *t2++ = *t; t2++;
+				j = t[1] -2;
+				t += 2;
+				while ( --j >= 0 ) {
+					WORD *b3 = bss;
+					while ( b3 < bns ) {
+						if ( b3[1] == CVECTOR ) {
+							SETS set = Sets+b3[0]; WORD i;
+							for ( i = set->first; i < set->last; i++ ) {
+								if ( SetElements[i] == *t ) {
+									*t1++ = *t++;
+									goto nextind2;
+								}
+							}
+						}
+						b3 += 2;
+					}
+					*t2++ = *t++;
+nextind2:;
 				}
 				m1[1] = WORDDIF(t1,m1);
 				if ( m1[1] == 2 ) t1 = m1;
@@ -1117,6 +1339,21 @@ FunBrac:	NCOPY(t1,t,i);
 							bind++;
 						}
 					}
+					if ( bset ) {
+						WORD *b3 = bss;
+						while ( b3 < bns ) {
+							if ( b3[1] == CVECTOR ) {
+								SETS set = Sets+b3[0]; WORD i;
+								for ( i = set->first; i < set->last; i++ ) {
+									if ( SetElements[i] == *t ) {
+										*t1++ = *t++; *t1++ = *t++;
+										goto nextvec;
+									}
+								}
+							}
+							b3 += 2;
+						}
+					}
 					*t2++ = *t++; *t2++ = *t++;
 nextvec:;
 				}
@@ -1125,6 +1362,32 @@ nextvec:;
 				m2[1] = WORDDIF(t2,m2);
 				if ( m2[1] == 2 ) t2 = m2;
 			}
+			else if ( bset ) {
+				m1 = t1; *t1++ = *t; t1++;
+				m2 = t2; *t2++ = *t; t2++;
+				s2 = t + i; t += 2;
+				while ( t < s2 ) {
+					WORD *b3 = bss;
+					while ( b3 < bns ) {
+						if ( b3[1] == CVECTOR ) {
+							SETS set = Sets+b3[0]; WORD i;
+							for ( i = set->first; i < set->last; i++ ) {
+								if ( SetElements[i] == *t ) {
+									*t1++ = *t++; *t1++ = *t++;
+									goto nextvec2;
+								}
+							}
+						}
+						b3 += 2;
+					}
+					*t2++ = *t++; *t2++ = *t++;
+nextvec2:;
+				}
+				m1[1] = WORDDIF(t1,m1);
+				if ( m1[1] == 2 ) t1 = m1;
+				m2[1] = WORDDIF(t2,m2);
+				if ( m2[1] == 2 ) t2 = m2;
+			}
 			else {
 				NCOPY(t2,t,i);
 			}
@@ -1138,7 +1401,7 @@ nextvec:;
 					s1 = b + b[1]; bb = b + 2;
 				}
 				s2 = t + i; t += 2;
-				while ( t < s2 && ( bb < s1 || bwild ) ) {
+				while ( t < s2 && ( bb < s1 || bwild || bset ) ) {
 					while ( bb < s1 && ( *bb < *t ||
 					( *bb == *t && bb[1] < t[1] ) ) ) bb += 3;
 					if ( bb < s1 && *bb == *t && bb[1] == t[1] ) {
@@ -1155,6 +1418,21 @@ nextvec:;
 							bind++;
 						}
 					}
+					if ( bset ) {
+						WORD *b3 = bss;
+						while ( b3 < bns ) {
+							if ( b3[1] == CVECTOR ) {
+								SETS set = Sets+b3[0]; WORD i;
+								for ( i = set->first; i < set->last; i++ ) {
+									if ( SetElements[i] == *t || SetElements[i] == t[1] ) {
+										*t1++ = *t++; *t1++ = *t++; *t1++ = *t++;
+										goto nextdot;
+									}
+								}
+							}
+							b3 += 2;
+						}
+					}
 					*t2++ = *t++; *t2++ = *t++; *t2++ = *t++;
 nextdot:;
 				}
@@ -1164,6 +1442,32 @@ nextdot:;
 				m2[1] = WORDDIF(t2,m2);
 				if ( m2[1] == 2 ) t2 = m2;
 			}
+			else if ( bset ) {
+				m1 = t1; *t1++ = *t; t1++;
+				m2 = t2; *t2++ = *t; t2++;
+				s2 = t + i; t += 2;
+				while ( t < s2 ) {
+					WORD *b3 = bss;
+					while ( b3 < bns ) {
+						if ( b3[1] == CVECTOR ) {
+							SETS set = Sets+b3[0]; WORD i;
+							for ( i = set->first; i < set->last; i++ ) {
+								if ( SetElements[i] == *t || SetElements[i] == t[1] ) {
+									*t1++ = *t++; *t1++ = *t++; *t1++ = *t++;
+									goto nextdot2;
+								}
+							}
+						}
+						b3 += 2;
+					}
+					*t2++ = *t++; *t2++ = *t++; *t2++ = *t++;
+nextdot2:;
+				}
+				m1[1] = WORDDIF(t1,m1);
+				if ( m1[1] == 2 ) t1 = m1;
+				m2[1] = WORDDIF(t2,m2);
+				if ( m2[1] == 2 ) t2 = m2;
+			}
 			else { NCOPY(t2,t,i); }
 		}
 		else if ( *t == SYMBOL ) {
@@ -1174,9 +1478,31 @@ nextdot:;
 				s2 = t + i; t += 2;
 				while ( bb < s1 && t < s2 ) {
 					while ( bb < s1 && *bb < *t ) bb += 2;
-					if ( bb >= s1 ) break;
+					if ( bb >= s1 ) {
+						if ( bset ) goto TrySymbolSet;
+						break;
+					}
 					if ( *bb == *t ) { *t1++ = *t++; *t1++ = *t++; }
+					else if ( bset ) {
+						WORD *bbb;
+TrySymbolSet:
+						bbb = bss;
+						while ( bbb < bns ) {
+							if ( bbb[1] == CSYMBOL ) { /* Set of symbols */
+								SETS set = Sets+bbb[0]; WORD i;
+								for ( i = set->first; i < set->last; i++ ) {
+									if ( SetElements[i] == *t ) {
+										*t1++ = *t++; *t1++ = *t++;
+										goto NextSymbol;
+									}
+								}
+							}
+							bbb += 2;
+						}
+						*t2++ = *t++; *t2++ = *t++;
+					}
 					else { *t2++ = *t++; *t2++ = *t++; } 
+NextSymbol:;
 				}
 				while ( t < s2 ) *t2++ = *t++;
 				m1[1] = WORDDIF(t1,m1);
@@ -1184,6 +1510,33 @@ nextdot:;
 				m2[1] = WORDDIF(t2,m2);
 				if ( m2[1] == 2 ) t2 = m2;
 			}
+			else if ( bset ) {
+				WORD *bbb;
+				m1 = t1; *t1++ = *t; t1++;
+				m2 = t2; *t2++ = *t; t2++;
+				s2 = t + i; t += 2;
+				while ( t < s2 ) {
+					bbb = bss;
+					while ( bbb < bns ) {
+						if ( bbb[1] == CSYMBOL ) { /* Set of symbols */
+							SETS set = Sets+bbb[0]; WORD i;
+							for ( i = set->first; i < set->last; i++ ) {
+								if ( SetElements[i] == *t ) {
+									*t1++ = *t++; *t1++ = *t++;
+									goto NextSymbol2;
+								}
+							}
+						}
+						bbb += 2;
+					}
+					*t2++ = *t++; *t2++ = *t++;
+NextSymbol2:;
+				}
+				m1[1] = WORDDIF(t1,m1);
+				if ( m1[1] == 2 ) t1 = m1;
+				m2[1] = WORDDIF(t2,m2);
+				if ( m2[1] == 2 ) t2 = m2;
+			}
 			else { NCOPY(t2,t,i); }
 		}
 		else {
@@ -1691,6 +2044,7 @@ WORD ContentMerge(PHEAD WORD *content, WORD *term)
 	WORD *outfill, *outb = TermMalloc("ContentMerge"), *ct;
 	WORD *t, *tstop, tsize, trsize, *told;
 	WORD *t1, *t2, *c1, *c2, i1, i2, *out1;
+	WORD didsymbol = 0, diddotp = 0, tfirst;
 	cstop = content + *content;
 	csize = cstop[-1];
 	if ( csize < 0 ) { sign = -sign; csize = -csize; }
@@ -1722,6 +2076,7 @@ WORD ContentMerge(PHEAD WORD *content, WORD *term)
 	while ( ct < cstop ) {
 		switch ( *ct ) {
 			case SYMBOL:
+				didsymbol = 1;
 				t = term+1;
 				while ( t < tstop && *t != *ct ) t += t[1];
 				if ( t >= tstop ) break;
@@ -1764,6 +2119,7 @@ WORD ContentMerge(PHEAD WORD *content, WORD *term)
 				if ( out1[1] == 2 ) outfill = out1;
 				break;
 			case DOTPRODUCT:
+				diddotp = 1;
 				t = term+1;
 				while ( t < tstop && *t != *ct ) t += t[1];
 				if ( t >= tstop ) break;
@@ -1849,26 +2205,57 @@ WORD ContentMerge(PHEAD WORD *content, WORD *term)
 			case GAMMA:
 			default:			/* Functions */
 				told = t;
-				while ( *t < *ct && t < tstop ) t += t[1];
-				if ( t >= tstop ) { t = told; }
-				else {
+				t = term+1;
+				while ( t < tstop ) {
+					if ( *t != *ct ) { t += t[1]; continue; }
+					if ( ct[1] != t[1] ) { t += t[1]; continue; }
+					if ( ct[2] != t[2] ) { t += t[1]; continue; }
 					t1 = t; t2 = ct; i1 = t1[1]; i2 = t2[1];
-					if ( i1 != i2 ) { t = told; }
-					else {
-						while ( i1 > 0 ) {
-							if ( *t1 != *t2 ) break;
-							t1++; t2++; i1--;
-						}
-						if ( i1 == 0 ) {
-							for ( i = 0; i < i2; i++ ) { *outfill++ = *t++; }
-						}
-						else { t = told; }
+					while ( i1 > 0 ) {
+						if ( *t1 != *t2 ) break;
+						t1++; t2++; i1--;
 					}
+					if ( i1 != 0 ) { t += t[1]; continue; }
+					t1 = t;
+					for ( i = 0; i < i2; i++ ) { *outfill++ = *t++; }
+/*
+					Mark as 'used'. The flags must be different!
+*/
+					t1[2] |= SUBTERMUSED1;
+					ct[2] |= SUBTERMUSED2;
+					t = told;
+					break;
 				}
 				break;
 		}
 		ct += ct[1];
 	}
+	if ( diddotp == 0 ) {
+		t = term+1; while ( t < tstop && *t != DOTPRODUCT ) t += t[1];
+		if ( t < tstop ) { /* now we need the negative powers */
+			tfirst = 1; told = outfill;
+			for ( i = 2; i < t[1]; i += 3 ) {
+				if ( t[i+2] < 0 ) {
+					if ( tfirst ) { *outfill++ = DOTPRODUCT; *outfill++ = 0; tfirst = 0; }
+					*outfill++ = t[i]; *outfill++ = t[i+1]; *outfill++ = t[i+2];
+				}
+			}
+			if ( outfill > told ) told[1] = outfill-told;
+		}
+	}
+	if ( didsymbol == 0 ) {
+		t = term+1; while ( t < tstop && *t != SYMBOL ) t += t[1];
+		if ( t < tstop ) { /* now we need the negative powers */
+			tfirst = 1; told = outfill;
+			for ( i = 2; i < t[1]; i += 2 ) {
+				if ( t[i+1] < 0 ) {
+					if ( tfirst ) { *outfill++ = SYMBOL; *outfill++ = 0; tfirst = 0; }
+					*outfill++ = t[i]; *outfill++ = t[i+1];
+				}
+			}
+			if ( outfill > told ) told[1] = outfill-told;
+		}
+	}
 /*
 	Now put the coefficient back.
 */
@@ -1890,6 +2277,16 @@ WORD ContentMerge(PHEAD WORD *content, WORD *term)
 	NumberFree(num,"ContentMerge");
 	for ( i = 0; i < *outb; i++ ) content[i] = outb[i];
 	TermFree(outb,"ContentMerge");
+/*
+	Now we have to 'restore' the term to its original.
+	We do not restore the content, because if anything was used the
+	new content overwrites the old. 6-mar-2018 JV
+*/
+	t = term + 1;
+	while ( t < tstop ) {
+		if ( *t >= FUNCTION ) t[2] &= ~SUBTERMUSED1;
+		t += t[1];
+	}
 	return(*content);
 CalledFrom:
 	MLOCK(ErrorMessageLock);
@@ -1912,6 +2309,18 @@ LONG TermsInExpression(WORD num)
 
 /*
  		#] TermsInExpression : 
+ 		#[ SizeOfExpression :
+*/
+
+LONG SizeOfExpression(WORD num)
+{
+	LONG x = (LONG)(DIVPOS(Expressions[num].size,sizeof(WORD)));
+	if ( x >= 0 ) return(x);
+	return(-1);
+}
+
+/*
+ 		#] SizeOfExpression : 
  		#[ UpdatePositions :
 */
 
@@ -1928,7 +2337,7 @@ void UpdatePositions()
 			AS.OldOnFile = (POSITION *)Malloc1(NumExpressions*sizeof(POSITION),"file pointers");
 			for ( i = 0; i < AS.NumOldOnFile; i++ ) AS.OldOnFile[i] = old[i];
 			AS.NumOldOnFile = NumExpressions;
-			M_free(old,"proces file pointers");
+			M_free(old,"process file pointers");
 		}
 		else {
 			AS.OldOnFile = (POSITION *)Malloc1(NumExpressions*sizeof(POSITION),"file pointers");
@@ -2031,9 +2440,8 @@ LONG CountTerms1(PHEAD0)
 	for(;;) {
 		numterms++;
 		retval = GetOneTerm(BHEAD AT.WorkPointer,AR.infile,&startposition,0);
+		if ( retval >= 0 ) AR.CompressPointer = oldipointer;
 		if ( retval <= 0 ) break;
-
-		AR.CompressPointer = oldipointer;
 		t = AR.CompressPointer;
 		if ( *t < (1 + decr + ABS(*(t+*t-1))) ) break;
 		t++;
