@@ -8,7 +8,7 @@
 
 /* #[ License : */
 /*
- *   Copyright (C) 1984-2017 J.A.M. Vermaseren
+ *   Copyright (C) 1984-2023 J.A.M. Vermaseren
  *   When using this file you are requested to refer to the publication
  *   J.A.M.Vermaseren "New features of FORM" math-ph/0010025
  *   This is considered a matter of courtesy as the development was paid
@@ -65,6 +65,7 @@
 #define EXTERNALCHANNELSTREAM 7
 #define PREREADSTREAM3 8
 #define REVERSEFILESTREAM 9
+#define INPUTSTREAM 10
  
 #define ENDOFSTREAM 0xFF
 #define ENDOFINPUT 0xFF
@@ -302,6 +303,7 @@ typedef int (*TFUN1)();
 #define NODOUBLEMASK 31
 #define QUADRUPLEFORTRANMODE 65
 #define QUADRUPLEPRECISIONFLAG 64
+#define ALLINTEGERDOUBLE 128
 #define NOQUADMASK 63
 #define NORMALFORMAT 0
 #define NOSPACEFORMAT 1
@@ -447,14 +449,15 @@ typedef int (*TFUN1)();
 #define PERMUTATIONS 103
 #define PARTITIONS 104
 #define MULFUNCTION 105
-#define TOPOLOGIES 106
-#define DIAGRAMS 107
-#define VERTEX 108
-#define EDGE 109
+#define MOEBIUS 106
+#define TOPOLOGIES 107
+#define DIAGRAMS 108
+#define VERTEX 109
+#define EDGE 110
 /*#define ALLWILD 109 ???? */
-#define SIZEOFFUNCTION 110
+#define SIZEOFFUNCTION 111
 
-#define MAXBUILTINFUNCTION 110
+#define MAXBUILTINFUNCTION 111
 #define FIRSTUSERFUNCTION 150
 
 /*
