@@ -7,7 +7,7 @@
  */
 /* #[ License : */
 /*
- *   Copyright (C) 1984-2013 J.A.M. Vermaseren
+ *   Copyright (C) 1984-2023 J.A.M. Vermaseren
  *   When using this file you are requested to refer to the publication
  *   J.A.M.Vermaseren "New features of FORM" math-ph/0010025
  *   This is considered a matter of courtesy as the development was paid
@@ -145,6 +145,7 @@ void PolyFunDirty(PHEAD WORD *term)
 	t = term+1;
 	while ( t < tstop ) {
 		if ( *t == AR.PolyFun ) {
+			if ( AR.PolyFunType == 2 ) t[2] |= MUSTCLEANPRF;
 			endarg = t + t[1];
 			t[2] |= DIRTYFLAG;
 			t += FUNHEAD;
@@ -179,7 +180,7 @@ void PolyFunClean(PHEAD WORD *term)
 	t = term+1;
 	while ( t < tstop ) {
 		if ( *t == AR.PolyFun ) {
-			t[2] |= CLEANPRF;
+			t[2] &= ~MUSTCLEANPRF;
 		}
 		t += t[1];
 	}
@@ -790,7 +791,7 @@ int ChainOut(PHEAD WORD *term, WORD funnum)
 
 /*
  		#] ChainOut : 
-  	#] Utilities :
+  	#] Utilities : 
 	#[ Patterns :
  		#[ MatchFunction :			WORD MatchFunction(pattern,interm,wilds)
 
@@ -809,11 +810,8 @@ int ChainOut(PHEAD WORD *term, WORD funnum)
 
 		The ones with a minus sign in front have been implemented.
 
-		There are still a few considerations:
-		1:  the dummy indices should be reset in multiple ?? matches.
-		2:	currently we cannot have a match with multiple ?? if
-			first there is a match and later the assignment isn't right.
-			we cannot go back at the moment to continue searching.
+		Note: the argument wilds allows backtracking when multiple
+		?a,?b give a match that later turns out to be useless.
 */
 
 WORD MatchFunction(PHEAD WORD *pattern, WORD *interm, WORD *wilds)
@@ -1241,7 +1239,8 @@ IndAll:				i = m[1] - WILDOFFSET;
 			else if ( *m == -ARGWILD ) goto ArgAll;
 			else if ( *m == -INDEX && m[1] >= AM.OffsetIndex+WILDOFFSET
 			&& m[1] < AM.OffsetIndex+(WILDOFFSET<<1) ) {
-				if ( *t == -VECTOR || *t == -SNUMBER ) goto IndAll;
+				if ( *t == -VECTOR ) goto IndAll;
+				if ( *t == -SNUMBER && t[1] >= 0 && t[1] < AM.OffsetIndex ) goto IndAll;
 				if ( *t == -MINVECTOR ) {
 					i = m[1] - WILDOFFSET;
 					AN.argaddress = AT.MinVecArg;
@@ -1257,10 +1256,30 @@ IndAll:				i = m[1] - WILDOFFSET;
 			}
 			else if ( *m == -VECTOR && *t == -MINVECTOR &&
 			( i = m[1] - WILDOFFSET ) >= AM.OffsetVector ) {
+/*
+================================
 				AN.argaddress = AT.MinVecArg;
 				AT.MinVecArg[ARGHEAD+3] = t[1];
 				if ( CheckWild(BHEAD i,VECTOSUB,1,AN.argaddress) ) goto endofloop;
 				AddWild(BHEAD i,VECTOSUB,(WORD)0);
+================================
+*/
+				if ( CheckWild(BHEAD i,VECTOMIN,t[1],&newvalue) ) goto endofloop;
+				AddWild(BHEAD i,VECTOMIN,newvalue);
+
+			}
+			else if ( *m == -MINVECTOR && *t == -VECTOR &&
+			( i = m[1] - WILDOFFSET ) >= AM.OffsetVector ) {
+/*
+================================
+				AN.argaddress = AT.MinVecArg;
+				AT.MinVecArg[ARGHEAD+3] = t[1];
+				if ( CheckWild(BHEAD i,VECTOSUB,1,AN.argaddress) ) goto endofloop;
+				AddWild(BHEAD i,VECTOSUB,(WORD)0);
+================================
+*/
+				if ( CheckWild(BHEAD i,VECTOMIN,t[1],&newvalue) ) goto endofloop;
+				AddWild(BHEAD i,VECTOMIN,newvalue);
 			}
 			else goto endofloop;
 		}
@@ -1318,9 +1337,13 @@ IndAll:				i = m[1] - WILDOFFSET;
 			else goto endofloop;
 		}
 		else if ( *m > 0 && *t > 0 ) {
+			WORD ii = *t-*m;
 			i = *m;
 			do { if ( *m++ != *t++ ) break; } while ( --i > 0 );
-			if ( i > 0 ) {
+			if ( i == 1 && ii == 0 ) {	/* sign difference */
+				goto endofloop;
+			}
+			else if ( i > 0 ) {
 				WORD *cto, *cfrom, *csav, ci;
 				WORD oRepFunNum;
 				WORD *oRepFunList;
@@ -1364,6 +1387,12 @@ IndAll:				i = m[1] - WILDOFFSET;
 				AN.RepFunNum = 0;
 				AN.RepFunList = AT.WorkPointer;
 		        AT.WorkPointer = (WORD *)(((UBYTE *)(AT.WorkPointer)) + AM.MaxTer);
+				if ( AT.WorkPointer+*t+5 > AT.WorkTop ) {
+					MLOCK(ErrorMessageLock);
+					MesWork();
+					MUNLOCK(ErrorMessageLock);
+					return(-1);
+				}
 				csav = cto = AT.WorkPointer;
 				cfrom = t;
 				ci = *t;
@@ -1422,6 +1451,21 @@ nomatch:
 				if ( *m == 1 || m[1] < FUNCTION ) {
 					if ( AN.ExpectedSign ) goto nomatch;
 				}
+				else {
+					if ( m[1] > FUNCTION + WILDOFFSET ) {
+						if ( functions[m[1]-FUNCTION-WILDOFFSET].spec >= TENSORFUNCTION ) {
+							if ( AN.ExpectedSign != AN.RepFunList[AN.RepFunNum-1] ) goto nomatch;
+						}
+					}
+             		else {
+						if ( AN.ExpectedSign != AN.RepFunList[AN.RepFunNum-1] ) goto nomatch;
+/*
+						if ( functions[m[1]-FUNCTION].spec >= TENSORFUNCTION ) {
+							if ( AN.ExpectedSign != AN.RepFunList[AN.RepFunNum-1] ) goto nomatch;
+						}
+*/
+					}
+				}
 				AN.nogroundlevel--;
 				AN.ExpectedSign = oExpectedSign;
 				AN.WildArgs = wildargs;
@@ -1568,6 +1612,7 @@ WORD ScanFunctions(PHEAD WORD *inpat, WORD *inter, WORD par)
 	CBUF *C = cbuf+AT.ebufnum;
 	int ntwa = AN.NumTotWildArgs;
 	LONG oldcpointer = C->Pointer - C->Buffer;
+	WORD oldSignCheck = AN.SignCheck;
 	instart = inter;
 /*
 	Only active for the last function in the pattern.
@@ -1639,7 +1684,7 @@ WORD ScanFunctions(PHEAD WORD *inpat, WORD *inter, WORD par)
 				if ( i >= AN.RepFunNum ) break;
 				inter += inter[1];
 			}
-			if ( inter >= AN.terstop ) { AT.WorkPointer = OldWork; return(0); }
+			if ( inter >= AN.terstop ) goto Failure;
 trythis:;
 		}
 		else {
@@ -1658,7 +1703,7 @@ trythis:;
 			if ( i >= AN.RepFunNum ) break;
 			inter += inter[1];
 		} while ( inter < AN.terstop );
-		if ( inter >= AN.terstop ) { AT.WorkPointer = OldWork; return(0); }
+		if ( inter >= AN.terstop ) goto Failure;
 #endif
 		wilds = 0;
 		/* We found one */
@@ -1727,6 +1772,10 @@ rewild:
 		if ( *inter != SUBEXPRESSION && MatchFunction(BHEAD inpat,inter,&wilds) ) {
 			AN.terfirstcomm = Oterfirstcomm;
 			if ( wilds ) {
+/*
+				Store wildcards to continue in MatchFunction if the current
+				wildcards do not work out.
+*/
 				wildargs = AN.WildArgs;
 				wildeat = AN.WildEat;
 				for ( i = 0; i < wildargs; i++ ) wildargtaken[i] = AT.WildArgTaken[i];
@@ -1757,7 +1806,7 @@ rewild:
 			}
 			if ( *inter < FUNCTION || functions[*inter-FUNCTION].commute ) {
 				newter = inter + inter[1];
-				if ( newter >= AN.terstop ) { AT.WorkPointer = OldWork; return(0); }
+				if ( newter >= AN.terstop ) goto Failure;
 				if ( *inter == GAMMA && inpat[1] <
 				inter[1] - AN.RepFunList[AN.RepFunNum-1] ) {
 					if ( ScanFunctions(BHEAD newpat,newter,2) ) goto OnSuccess;
@@ -1834,10 +1883,24 @@ maybenext:
 		}}
 		inter += inter[1];
 	} while ( inter < AN.terstop );
+Failure:
+    AN.SignCheck = oldSignCheck;
 	AT.WorkPointer = OldWork;
 	return(0);
 OnSuccess:
+	if ( AT.idallflag && AN.nogroundlevel <= 0 ) {
+		if ( AT.idallmaxnum > 0 && AT.idallnum >= AT.idallmaxnum ) {
+			AN.terfirstcomm = Oterfirstcomm;
+    		AN.SignCheck = oldSignCheck;
+			AT.WorkPointer = OldWork;
+			return(0);
+		}
+		SubsInAll(BHEAD0);
+		AT.idallnum++;
+		if ( AT.idallmaxnum == 0 || AT.idallnum < AT.idallmaxnum ) goto NoMat;
+	}
 	AN.terfirstcomm = Oterfirstcomm;
+    AN.SignCheck = oldSignCheck;
 /*
 	Now the disorder test
 */
@@ -1851,8 +1914,7 @@ OnSuccess:
 			t = AN.terstart + AN.RepFunList[i];
 			if ( *m != *t ) {
 				if ( *m > *t ) continue;
-jexch:			AT.WorkPointer = OldWork;
-				return(1);
+				goto doesmatch;
 			}
 			if ( *m >= FUNCTION && functions[*m-FUNCTION].spec >=
 				TENSORFUNCTION ) {
@@ -1869,15 +1931,16 @@ jexch:			AT.WorkPointer = OldWork;
 			}
 			while ( k > 0 && kk > 0 ) {
 				if ( *m < *t ) goto NextFor;
-				else if ( *m++ > *t++ ) goto jexch;
+				else if ( *m++ > *t++ ) goto doesmatch;
 				k--; kk--;
 			}
-			if ( k > 0 ) goto jexch;
+			if ( k > 0 ) goto doesmatch;
 NextFor:;
 		}
 		SetStop = 1;
 		goto NoMat;
 	}
+doesmatch:
 	AT.WorkPointer = OldWork;
 	return(1);
 }
@@ -1886,5 +1949,3 @@ NextFor:;
  		#] ScanFunctions : 
 	#] Patterns :
 */
-
-/* temporary commentary for forcing cvs merge */
