@@ -8,7 +8,7 @@
  */
 /* #[ License : */
 /*
- *   Copyright (C) 1984-2013 J.A.M. Vermaseren
+ *   Copyright (C) 1984-2023 J.A.M. Vermaseren
  *   When using this file you are requested to refer to the publication
  *   J.A.M.Vermaseren "New features of FORM" math-ph/0010025
  *   This is considered a matter of courtesy as the development was paid
@@ -450,7 +450,8 @@ va_dcl
 					*t = 0;
 					AddToLine((UBYTE *)Out);
 					if ( d->nfactors >= 1 && AN.listinprint[2] == DOLLAREXPR2 ) {
-						if ( d->type == 0 || d->factors == 0 ) goto dollarzero;
+						if ( d->type == 0 ||
+						 ( d->factors == 0 && d->nfactors != 1 ) ) goto dollarzero;
 						num = EvalDoLoopArg(BHEAD AN.listinprint+2,-1);
 						if ( num == 0 ) {
 							value[0] = 4; value[1] = d->nfactors; value[2] = 1; value[3] = 3; value[4] = 0;
@@ -954,5 +955,19 @@ VOID PrintWords(WORD *buffer, LONG number)
 
 /*
  		#] PrintWords : 
+ 		#[ PrintSeq :
+*/
+ 
+void PrintSeq(WORD *a,char *text)
+{
+	MesPrint(" %s:",text);
+	while ( *a ) {
+		MesPrint("     %a",a[0],a);
+		a += *a;
+	}
+}
+
+/*
+ 		#] PrintSeq : 
 	#] exit :
 */
