@@ -6,7 +6,7 @@
  */
 /* #[ License : */
 /*
- *   Copyright (C) 1984-2013 J.A.M. Vermaseren
+ *   Copyright (C) 1984-2023 J.A.M. Vermaseren
  *   When using this file you are requested to refer to the publication
  *   J.A.M.Vermaseren "New features of FORM" math-ph/0010025
  *   This is considered a matter of courtesy as the development was paid
@@ -322,6 +322,17 @@ VOID FullCleanUp()
 
 	CompactifyTree(AC.exprnames,EXPRNAMES);
 
+	for ( j = AO.NumDictionaries-1; j >= 0; j-- ) {
+		RemoveDictionary(AO.Dictionaries[j]);
+		M_free(AO.Dictionaries[j],"Dictionary");
+	}
+	AO.NumDictionaries = AO.gNumDictionaries = 0;
+	M_free(AO.Dictionaries,"Dictionaries");
+	AO.Dictionaries = 0;
+	AO.SizeDictionaries = 0;
+	AP.OpenDictionary = 0;
+	AO.CurrentDictionary = 0;
+
 	AP.ComChar = AP.cComChar;
 	if ( AP.procedureExtension ) M_free(AP.procedureExtension,"procedureextension");
 	AP.procedureExtension = strDup1(AP.cprocedureExtension,"procedureextension");
@@ -335,12 +346,14 @@ VOID FullCleanUp()
 	AC.ThreadStats = AM.gThreadStats = AM.ggThreadStats;
 	AC.OldFactArgFlag = AM.gOldFactArgFlag = AM.ggOldFactArgFlag;
 	AC.FinalStats = AM.gFinalStats = AM.ggFinalStats;
+	AC.OldGCDflag = AM.gOldGCDflag = AM.ggOldGCDflag;
 	AC.ThreadsFlag = AM.gThreadsFlag = AM.ggThreadsFlag;
 	if ( AC.ThreadsFlag && AM.totalnumberofthreads > 1 ) AS.MultiThreaded = 1;
 	AC.ThreadBucketSize = AM.gThreadBucketSize = AM.ggThreadBucketSize;
 	AC.ThreadBalancing = AM.gThreadBalancing = AM.ggThreadBalancing;
 	AC.ThreadSortFileSynch = AM.gThreadSortFileSynch = AM.ggThreadSortFileSynch;
 	AC.ShortStatsMax = AM.gShortStatsMax = AM.ggShortStatsMax;
+	AC.SizeCommuteInSet = AM.gSizeCommuteInSet = 0;
 
 	NumExpressions = 0;
 	if ( DeleteStore(0) < 0 ) {
@@ -428,21 +441,21 @@ int DoPolyratfun(UBYTE *s)
 	WORD funnum;
 	if ( AC.origin == FROMPOINTINSTRUCTION ) {
 		if ( *s == 0 || *s == ',' || *s == ')' ) {
-			AR.PolyFun = 0; AR.PolyFunType = 0;
+			AR.PolyFun = 0; AR.PolyFunType = 0; AR.PolyFunInv = 0; AR.PolyFunExp = 0;
 			return(0);
 		}
 		if ( *s != '=' ) {
-			MesPrint("@Proper use in point instructions is: PolyRatFun[=functionname]");
+			MesPrint("@Proper use in point instructions is: PolyRatFun[=functionname[+functionname]]");
 			return(-1);
 		}
 	}
 	else {
 		if ( *s == 0 ) {
-			AR.PolyFun = 0; AR.PolyFunType = 0;
+			AR.PolyFun = 0; AR.PolyFunType = 0; AR.PolyFunInv = 0; AR.PolyFunExp = 0;
 			return(0);
 		}
 		if ( *s != '=' && *s != ',' ) {
-			MesPrint("@Proper use is: PolyRatFun[{ ,=}functionname]");
+			MesPrint("@Proper use is: PolyRatFun[{ ,=}functionname[+functionname]]");
 			return(-1);
 		}
 	}
@@ -452,18 +465,31 @@ int DoPolyratfun(UBYTE *s)
 	c = *t; *t = 0;
 
 	if ( GetName(AC.varnames,s,&funnum,WITHAUTO) != CFUNCTION ) {
+Error1:;
 		MesPrint("@ %s is not a properly declared function",s);
 		*t = c;
 		return(-1);
 	}
 	if ( functions[funnum].spec != 0 || functions[funnum].commute != 0 ) {
+Error2:;
 		MesPrint("@The PolyRatFun must be a regular commuting function!");
 		*t = c;
 		return(-1);
 	}
 	AR.PolyFun = funnum+FUNCTION; AR.PolyFunType = 2;
+	AR.PolyFunInv = 0;
+	AR.PolyFunExp = 0;
 	AC.PolyRatFunChanged = 1;
 	*t = c;
+	if ( *t == '+' ) {
+		t++; s = t;
+		t = EndOfToken(s);
+		c = *t; *t = 0;
+		if ( GetName(AC.varnames,s,&funnum,WITHAUTO) != CFUNCTION ) goto Error1;
+		if ( functions[funnum].spec != 0 || functions[funnum].commute != 0 ) goto Error2;
+		AR.PolyFunInv = funnum+FUNCTION;
+		*t = c;
+	}
 	SKIPBLANKS(t)
 	if ( *t && *t != ',' && *t != ')' ) {
 		t++; c = *t; *t = 0;
@@ -475,7 +501,7 @@ int DoPolyratfun(UBYTE *s)
 }
 
 /*
- 		#] DoPolyratfun :
+ 		#] DoPolyratfun : 
  		#[ DoNoParallel :
 */
 
