@@ -8,7 +8,7 @@
  */
 /* #[ License : */
 /*
- *   Copyright (C) 1984-2017 J.A.M. Vermaseren
+ *   Copyright (C) 1984-2023 J.A.M. Vermaseren
  *   When using this file you are requested to refer to the publication
  *   J.A.M.Vermaseren "New features of FORM" math-ph/0010025
  *   This is considered a matter of courtesy as the development was paid
@@ -1161,6 +1161,7 @@ UBYTE *DoDimension(UBYTE *s, int *dim, int *dim4)
 	int type, error = 0;
 	WORD numsymbol;
 	NAMETREE **oldtree = AC.activenames;
+	LIST* oldsymbols = AC.Symbols;
 	*dim4 = -NMIN4SHIFT;
 	if ( FG.cTable[*s] == 1 ) {
 retry:
@@ -1174,6 +1175,7 @@ retry:
 	else if ( ( (FG.cTable[*s] == 0 ) || ( *s == '[' ) )
 		&& ( s = SkipAName(s) ) != 0 ) {
 		AC.activenames = &(AC.varnames);
+		AC.Symbols = &(AC.SymbolList);
 		c = *s; *s = 0;
 		if ( ( ( type = GetName(AC.exprnames,t,&numsymbol,NOAUTO) ) != NAMENOTFOUND )
 		|| ( ( type = GetName(AC.varnames,t,&numsymbol,WITHAUTO) ) != NAMENOTFOUND ) ) {
@@ -1181,7 +1183,7 @@ retry:
 		}
 		else {
 			numsymbol = AddSymbol(t,-MAXPOWER,MAXPOWER,0,0);
-			if ( *oldtree != AC.autonames && AC.WarnFlag )
+			if ( AC.WarnFlag )
 			MesPrint("&Warning: Implicit declaration of %s as a symbol",t);
 		}
 		*dim = -numsymbol;
@@ -1195,7 +1197,7 @@ retry:
 			}
 			else {
 				numsymbol = AddSymbol(t,-MAXPOWER,MAXPOWER,0,0);
-				if ( *oldtree != AC.autonames && AC.WarnFlag )
+				if ( AC.WarnFlag )
 				MesPrint("&Warning: Implicit declaration of %s as a symbol",t);
 			}
 			*dim4 = -numsymbol-NMIN4SHIFT;
@@ -1208,6 +1210,7 @@ retry:
 illeg:	MesPrint("&Illegal dimension specification. Should be number >= 0, symbol or symbol:symbol");
 		return(0);
 	}
+	AC.Symbols = oldsymbols;
 	AC.activenames = oldtree;
 	if ( error ) return(0);
 	return(s);
