@@ -11,7 +11,7 @@
  */
 /* #[ License : */
 /*
- *   Copyright (C) 1984-2013 J.A.M. Vermaseren
+ *   Copyright (C) 1984-2023 J.A.M. Vermaseren
  *   When using this file you are requested to refer to the publication
  *   J.A.M.Vermaseren "New features of FORM" math-ph/0010025
  *   This is considered a matter of courtesy as the development was paid
@@ -97,10 +97,11 @@
 WORD TestMatch(PHEAD WORD *term, WORD *level)
 {
 	GETBIDENTITY
-	WORD *ll, *m, *w, *llf, *OldWork, *StartWork, *ww, *mm;
-	WORD power = 0, match = 0, /* *rep, */ i, msign = 0;
-	int numdollars = 0, protosize;
-	CBUF *C = cbuf+AM.rbufnum;
+	WORD *ll, *m, *w, *llf, *OldWork, *StartWork, *ww, *mm, *t, *OldTermBuffer = 0;
+	WORD power = 0, match = 0, i, msign = 0, ll2;
+	int numdollars = 0, protosize, oldallnumrhs;
+	CBUF *C = cbuf+AM.rbufnum, *CC;
+	AT.idallflag = 0;
 	do {
 /*
  		#[ Preliminaries :
@@ -146,7 +147,6 @@ WORD TestMatch(PHEAD WORD *term, WORD *level)
 	OldWork = AT.WorkPointer;
 	if ( AT.WorkPointer < term + *term ) AT.WorkPointer = term + *term;
 	ww = AT.WorkPointer;
-#ifdef WITHPTHREADS
 /*
 		Here we need to make a copy of the subexpression object because we
 		will be writing the values of the wildcards in it. 
@@ -173,34 +173,18 @@ WORD TestMatch(PHEAD WORD *term, WORD *level)
 		m = ma + IDHEAD;
 		NCOPY(ma,ta,ja);
 		*ma = 0;
-/*
-		Old code
-
-		WORD *ma = AddRHS(AT.ebufnum,1);
-		CBUF *CC = cbuf+AT.ebufnum;
-		if ( ( ma+ja+2 ) > CC->Top ) {
-			ma = DoubleCbuffer(AT.ebufnum,ma);
-		}
-		m = ma + IDHEAD;
-		NCOPY(ma,ta,ja);
-		*ma++ = 0;
-		CC->rhs[CC->numrhs+1] = ma;
-		CC->Pointer = ma;
-*/
 	}
-#else
-	m = ll + IDHEAD;
-#endif
 	AN.FullProto = m;
 	AN.WildValue = w = m + SUBEXPSIZE;
 	protosize = IDHEAD + m[1];
 	m += m[1];
 	AN.WildStop = m;
 	StartWork = ww;
+	ll2 = ll[2];
 /*
  		#[ Expand dollars :
 */
-	if ( ( ll[4] & 1 ) != 0 ) {	/* We have at least one dollar in the pattern */
+	if ( ( ll[4] & DOLLARFLAG ) != 0 ) {	/* We have at least one dollar in the pattern */
 		WORD oldRepPoint = *AN.RepPoint, olddefer = AR.DeferFlag;
 		AR.Eside = LHSIDEX;
 /*
@@ -261,14 +245,74 @@ WORD TestMatch(PHEAD WORD *term, WORD *level)
 		m = AN.patternbuffer + protosize;
 		AR.Eside = RHSIDE;
 		*mm = 0;
-
+/*
+		Test the pattern. If only wildcard powers -> SUBONCE
+*/
+		{
+			WORD *mmm = m + *m, *m1 = m+1, jm, noveto = 0;
+			while ( m1 < mmm ) {
+				if ( *m1 == SYMBOL ) {
+					for ( jm = 2; jm < m1[1]; jm+=2 ) {
+						if ( m1[jm+1] < MAXPOWER && m1[jm+1] > -MAXPOWER ) break;
+					}
+					if ( jm < m1[1] ) { noveto = 1; break; }
+				}
+				else if ( *m1 == DOTPRODUCT ) {
+					for ( jm = 2; jm < m1[1]; jm+=3 ) {
+						if ( m1[jm+2] < MAXPOWER && m1[jm+2] > -MAXPOWER ) break;
+					}
+					if ( jm < m1[1] ) { noveto = 1; break; }
+				}
+				else { noveto = 1; break; }
+				m1 += m1[1];
+			}
+			if ( noveto == 0 ) {
+				ll2 = ll2 & ~SUBMASK;
+				ll2 |= SUBONCE;
+			}
+		}
 		AT.WorkPointer = ww = StartWork;
 		*AN.RepPoint = oldRepPoint;
 	}
 /*
  		#] Expand dollars : 
 
-	AT.WorkPointer = ww = term + *term;
+	In case of id,all we have to check at this point that there are only
+	functions in the pattern.
+*/
+	if ( ( ll2 & SUBMASK ) == SUBALL ) {
+		WORD *t = AN.patternbuffer+IDHEAD, *tt;
+		WORD *tstop, *ttstop, ii;
+		t += t[1]; tstop = t + *t; t++;
+		while ( t < tstop ) {
+			if ( *t < FUNCTION ) break;
+			t += t[1];
+		}
+		if ( t < tstop ) {
+			MLOCK(ErrorMessageLock);
+			MesPrint("Error: id,all can only be used with (products of) functions and/or tensors.");
+			MUNLOCK(ErrorMessageLock);
+			return(-1);
+		}
+		OldTermBuffer = AN.termbuffer;
+		AN.termbuffer = TermMalloc("id,all");
+/*
+		Now make sure that only regular functions and tensors can take part.
+*/
+		tt = term; ttstop = tt+*tt; ttstop -= ABS(ttstop[-1]); tt++;
+		t = AN.termbuffer+1; 
+		while ( tt < ttstop ) {
+			if ( *tt >= FUNCTION && *tt != AR.PolyFun && *tt != AR.PolyFunInv ) {
+				ii = tt[1]; NCOPY(t,tt,ii);
+			}
+			else tt += tt[1];
+		}
+		*t++ = 1; *t++ = 1; *t++ = 3; AN.termbuffer[0] = t-AN.termbuffer;
+	}
+/*
+	To be puristic, we need to check that all wildcards in the prototype
+	are actually present. If the LHS contained a replace_ this may not be
+	the case.
 */
 	ClearWild(BHEAD0);
 	while ( w < AN.WildStop ) {
@@ -284,9 +328,9 @@ WORD TestMatch(PHEAD WORD *term, WORD *level)
 		MUNLOCK(ErrorMessageLock);
 		return(-1);
 	}
-	AN.DisOrderFlag = ll[2] & SUBDISORDER;
+	AN.DisOrderFlag = ll2 & SUBDISORDER;
 	AN.nogroundlevel = 0;
-	switch ( ll[2] & SUBMASK ) {
+	switch ( ll2 & SUBMASK ) {
 		case SUBONLY :
 			/* Must be an exact match */
 			AN.UseFindOnly = 1; AN.ForFindOnly = 0;
@@ -306,7 +350,7 @@ WORD TestMatch(PHEAD WORD *term, WORD *level)
 						if ( msign ) term[term[0]-1] = -term[term[0]-1];
 						Substitute(BHEAD term,m,1);
 						if ( numdollars ) {
-							WildDollars(BHEAD0);
+							WildDollars(BHEAD (WORD *)0);
 							numdollars = 0;
 						}
 						if ( ww < term+term[0] ) ww = term+term[0];
@@ -339,7 +383,7 @@ WORD TestMatch(PHEAD WORD *term, WORD *level)
 						if ( msign ) term[term[0]-1] = -term[term[0]-1];
 						Substitute(BHEAD term,m,1);
 						if ( numdollars ) {
-							WildDollars(BHEAD0);
+							WildDollars(BHEAD (WORD *)0);
 							numdollars = 0;
 						}
 						if ( ww < term+term[0] ) ww = term+term[0];
@@ -372,7 +416,7 @@ WORD TestMatch(PHEAD WORD *term, WORD *level)
 						if ( msign ) term[term[0]-1] = -term[term[0]-1];
 						Substitute(BHEAD term,m,1);
 						if ( numdollars ) {
-							WildDollars(BHEAD0);
+							WildDollars(BHEAD (WORD *)0);
 							numdollars = 0;
 						}
 						if ( ww < term+term[0] ) ww = term+term[0];
@@ -399,16 +443,12 @@ WORD TestMatch(PHEAD WORD *term, WORD *level)
 					match = 1;
 				}
 			}
-#if IDHEAD > 3
 			if ( match ) {
-				if ( ( ll[2] & SUBAFTER ) != 0 ) *level = AC.Labels[ll[3]];
+				if ( ( ll2 & SUBAFTER ) != 0 ) *level = AC.Labels[ll[3]];
 			}
 			else {
-				if ( ( ll[2] & SUBAFTERNOT ) != 0 ) *level = AC.Labels[ll[3]];
+				if ( ( ll2 & SUBAFTERNOT ) != 0 ) *level = AC.Labels[ll[3]];
 			}
-#endif
-/*			AT.WorkPointer = AN.RepFunList;
-			return(match); */
 			goto nextlevel;
 		case SUBONCE :
 			AN.UseFindOnly = 0;
@@ -422,7 +462,7 @@ WORD TestMatch(PHEAD WORD *term, WORD *level)
 			power = FindMulti(BHEAD term,m);
 			if ( ( power & 1 ) != 0 && msign ) term[term[0]-1] = -term[term[0]-1];
 			break;
-		case SUBALL :
+		case SUBVECTOR :
 			while ( ( power = FindAll(BHEAD term,m,*level,(WORD *)0) ) != 0 ) {
 				if ( ( power & 1 ) != 0 && msign ) term[term[0]-1] = -term[term[0]-1];
 				match = 1;
@@ -458,7 +498,7 @@ WORD TestMatch(PHEAD WORD *term, WORD *level)
 						t1 = term; t2 = AN.selecttermundo; i = *t2;
 						NCOPY(t1,t2,i);
 #if IDHEAD > 3
-						if ( ( ll[2] & SUBAFTERNOT ) != 0 ) {
+						if ( ( ll2 & SUBAFTERNOT ) != 0 ) {
 							*level = AC.Labels[ll[3]];
 						}
 #endif
@@ -466,48 +506,124 @@ WORD TestMatch(PHEAD WORD *term, WORD *level)
 					}
 				}
 				if ( numdollars ) {
-					WildDollars(BHEAD0);
+					WildDollars(BHEAD (WORD *)0);
 					numdollars = 0;
 				}
 				match = 1;
-#if IDHEAD > 3
-				if ( ( ll[2] & SUBAFTER ) != 0 ) {
+				if ( ( ll2 & SUBAFTER ) != 0 ) {
 					*level = AC.Labels[ll[3]];
 				}
-#endif
 			}
 			else {
-#if IDHEAD > 3
-				if ( ( ll[2] & SUBAFTERNOT ) != 0 ) {
+				if ( ( ll2 & SUBAFTERNOT ) != 0 ) {
 					*level = AC.Labels[ll[3]];
 				}
-#endif
 				power = 0;
 			}
 			goto nextlevel;
+		case SUBALL:
+			AN.UseFindOnly = 0;
+			CC = cbuf+AT.allbufnum;
+			oldallnumrhs = CC->numrhs;
+			t = AddRHS(AT.allbufnum,1);
+			*t = 0;
+			AT.idallflag = 1;
+			AT.idallmaxnum = ll[5];
+			AT.idallnum = 0;
+			if ( FindRest(BHEAD AN.termbuffer,m) || AT.idallflag > 1 ) {
+				WORD *t, *tstop, *tt, first = 1, ii;
+				power = 1;
+				*CC->Pointer++ = 0;
+				if ( msign ) term[term[0]-1] = -term[term[0]-1];
+/*
+				If we come here the matches are all already in the
+				compiler buffer. All we need to do is take out all
+				functions and replace them by a SUBEXPRESSION that
+				points to this buffer.
+				Note: the PolyFun/PolyRatFun should be excluded from this.
+				This works because each match writes incrementally to
+				the buffer using the routine SubsInAll.
+
+				The call to WildDollars should be made in Generator.....
+*/
+				t = term; tstop = t + *t; ii = ABS(tstop[-1]); tstop -= ii;
+				tt = AT.WorkPointer+1;
+				t++;
+				while ( t < tstop ) {
+					if ( *t >= FUNCTION && *t != AR.PolyFun && *t != AR.PolyFunInv ) {
+						if ( first ) { /* SUBEXPRESSION */
+							*tt++ = SUBEXPRESSION;
+							*tt++ = SUBEXPSIZE;
+							*tt++ = CC->numrhs;
+							*tt++ = 1;
+							*tt++ = AT.allbufnum;
+							FILLSUB(tt)
+							first = 0;
+						}
+						t += t[1];
+					}
+					else {
+						i = t[1]; NCOPY(tt,t,i);
+					}
+				}
+				if ( ( ll[4] & NORMALIZEFLAG ) != 0 ) {
+/*
+					In case of the normalization option, we have to divide
+					by AT.idallnum;
+*/
+					WORD na = t[ii-1];
+					na = REDLENG(na);
+					for ( i = 0; i < ii; i++ ) tt[i] = t[i];
+					Divvy(BHEAD (UWORD *)tt,&na,(UWORD *)(&(AT.idallnum)),1);
+					na = INCLENG(na);
+					ii = ABS(na);
+					tt[ii-1] = na;
+					tt += ii;
+				}
+				else {
+					NCOPY(tt,t,ii);
+				}
+				ii = tt-AT.WorkPointer;
+				*(AT.WorkPointer) = ii;
+				tt = AT.WorkPointer; t = term;
+				NCOPY(t,tt,ii);
+
+				if ( ( ll2 & SUBAFTER ) != 0 ) { /* ifmatch -> */
+					*level = AC.Labels[ll[3]];
+				}
+				TermFree(AN.termbuffer,"id,all");
+				AN.termbuffer = OldTermBuffer;
+				AT.WorkPointer = AN.RepFunList;
+				AT.idallflag = 0;
+				CC->Pointer[0] = 0;
+				TransferBuffer(AT.aebufnum,AT.ebufnum,AT.allbufnum);
+				return(1);
+			}
+			AT.idallflag = 0;
+			power = 0;
+			CC->numrhs = oldallnumrhs;
+			TermFree(AN.termbuffer,"id,all");
+			AN.termbuffer = OldTermBuffer;
+			break;
 		default :
 			break;
 	}
 	if ( power ) {
 		Substitute(BHEAD term,m,power);
 		if ( numdollars ) {
-			WildDollars(BHEAD0);
+			WildDollars(BHEAD (WORD *)0);
 			numdollars = 0;
 		}
 		match = 1;
-#if IDHEAD > 3
-		if ( ( ll[2] & SUBAFTER ) != 0 ) {
+		if ( ( ll2 & SUBAFTER ) != 0 ) { /* ifmatch -> */
 			*level = AC.Labels[ll[3]];
 		}
-#endif
 	}
 	else {
 		AT.WorkPointer = AN.RepFunList;
-#if IDHEAD > 3
-		if ( ( ll[2] & SUBAFTERNOT ) != 0 ) {
+		if ( ( ll2 & SUBAFTERNOT ) != 0 ) { /* ifnomatch -> */
 			*level = AC.Labels[ll[3]];
 		}
-#endif
 	}
 nextlevel:;
 	} while ( (*level)++ < AR.Cnumlhs && C->lhs[*level][0] == TYPEIDOLD );
@@ -1263,7 +1379,7 @@ CopRest:				t = tstop;
 						&& C->lhs[level][0] == TYPEIDOLD ) {
 							m = C->lhs[level];
 							m += IDHEAD;
-							if ( m[-IDHEAD+2] == SUBALL ) {
+							if ( m[-IDHEAD+2] == SUBVECTOR ) {
 							if ( ( vv = m[m[1]+3] ) == r[1] ) {
 OnePV:							TwoProto = AN.FullProto;
 TwoPV:							m = AT.WorkPointer;
@@ -1384,7 +1500,7 @@ Hitlevel1:							level2 = level;
 										if ( !par ) m = C->lhs[level2];
 										else m = par;
 										m += IDHEAD;
-										if ( m[-IDHEAD+2] == SUBALL ) {
+										if ( m[-IDHEAD+2] == SUBVECTOR ) {
 										if ( ( vv = m[m[1]+3] ) == r[1] )
 											goto OnePV;
 										else if ( vv >= OffNum ) {
@@ -1440,7 +1556,7 @@ Hitlevel2:							level2 = level;
 										if ( !par ) m = C->lhs[level2];
 										else m = par;
 										m += IDHEAD;
-										if ( m[-IDHEAD+2] == SUBALL ) {
+										if ( m[-IDHEAD+2] == SUBVECTOR ) {
 										if ( ( vv = m[6] ) == *r )
 											goto OnePV;
 										else if ( vv >= OffNum ) {
@@ -1849,6 +1965,423 @@ dotensor:
 
 /*
  		#] TestSelect : 
-  	#] Patterns :
+ 		#[ SubsInAll :			VOID SubsInAll()
+
+		This routine takes a match in id,all and stores it away in
+		the AT.allbufnum 'compiler' buffer, after taking out the pattern.
+		The main problem here is that id,all usually has (lots of) wildcards
+		and their assignments are on stack and the difficult ones are in
+		AT.ebufnum. Popping the stack while looking for more matches would
+		loose those. Hence we have to copy them into yet another compiler
+		buffer: AT.aebufnum. Because this may involve many matches and
+		because the original term has only a limited number of arguments,
+		it will pay to look for already existing ones in this buffer.
+		(to be done later).
+*/
+
+VOID SubsInAll(PHEAD0)
+{
+	GETBIDENTITY
+	WORD *TemTerm;
+	WORD *t, *m, *term;
+	WORD *tstop, *mstop, *xstop;
+	WORD nt, *fill, nq, mt;
+	WORD *tcoef, i = 0;
+	WORD PutExpr = 0, sign = 0;
+/*
+	We start with building the term in the WorkSpace.
+	Afterwards we will transfer it to AT.allbufnum.
+	We have to make sure there is room in the WorkSpace.
+*/
+	AT.idallflag = 2;
+	TemTerm = AT.WorkPointer;
+	if ( ( (WORD *)(((UBYTE *)(AT.WorkPointer)) + AM.MaxTer*2) ) > AT.WorkTop ) {
+		MLOCK(ErrorMessageLock);
+		MesWork();
+		MUNLOCK(ErrorMessageLock);
+		Terminate(-1);
+	}
+	m = AN.patternbuffer + IDHEAD; m += m[1];
+	mstop = m + *m;
+	m++;
+	term = AN.termbuffer;
+	tstop = term + *term; tcoef = tstop-1; tstop -= ABS(tstop[-1]);
+ 	t = term;
+	t++;
+	fill = TemTerm;
+	fill++;
+	while ( m < mstop ) {
+		while ( t < tstop ) {
+			nt = WORDDIF(t,term);
+			for ( mt = 0; mt < AN.RepFunNum; mt += 2 ) {
+                if ( nt == AN.RepFunList[mt] ) break;
+          	}
+			if ( mt >= AN.RepFunNum ) {
+				nq = t[1];
+				NCOPY(fill,t,nq);
+			}
+			else {
+				WORD *oldt = 0;
+				if ( *m == GAMMA && m[1] != FUNHEAD+1 ) {
+					oldt = t;
+					if ( ( i = AN.RepFunList[mt+1] ) > 0 ) {
+						*fill++ = GAMMA;
+						*fill++ = i + FUNHEAD+1;
+						FILLFUN(fill)
+						nq = i + 1;
+						t += FUNHEAD;
+						NCOPY(fill,t,nq);
+					}
+					t = oldt;
+				}
+				else if ( ( *t == LEVICIVITA ) || ( *t >= FUNCTION
+				&& (functions[*t-FUNCTION].symmetric & ~REVERSEORDER) == ANTISYMMETRIC )
+								 ) sign += AN.RepFunList[mt+1];
+				else if ( *m >= FUNCTION+WILDOFFSET
+				&& (functions[*m-FUNCTION-WILDOFFSET].symmetric & ~REVERSEORDER) == ANTISYMMETRIC
+								 ) sign += AN.RepFunList[mt+1];
+				if ( !PutExpr ) {
+					WORD *pstart = fill, *p, *w, *ww;
+					xstop = t + t[1];
+					t = AN.FullProto;
+					nq = t[1];
+					t[3] = 1;
+					NCOPY(fill,t,nq);
+					t = xstop;
+					PutExpr = 1;
+/*
+					Here we need provisions for keeping wildcard matches
+					that reside in AT.ebufnum. We will move them to
+					AT.aebufnum.
+					Problem: the SUBEXPRESSION assumes automatically
+					that the compiler buffer is AT.ebufnum. We have to
+					correct that in TranferBuffer.
+*/
+					p = pstart + SUBEXPSIZE;
+					while ( p < fill ) {
+						switch ( *p ) {
+							case SYMTOSUB:
+							case VECTOSUB:
+							case INDTOSUB:
+							case ARGTOARG:
+							case ARLTOARL:
+								w = cbuf[AT.ebufnum].rhs[p[3]];
+								ww = cbuf[AT.ebufnum].rhs[p[3]+1];
+/*
+								Here we could search for whether this
+								object sits in the buffer already.
+								To be done later.
+								By the way: ww-w fits inside a WORD.
+*/
+								AddRHS(AT.aebufnum,1);
+								AddNtoC(AT.aebufnum,ww-w,w,11);
+								p[3] = cbuf[AT.aebufnum].numrhs;
+								cbuf[AT.aebufnum].rhs[p[3]+1] = cbuf[AT.aebufnum].Pointer;
+								p += p[1];
+								break;
+							case FROMSET:
+							case SETTONUM:
+							case LOADDOLLAR:
+								p += p[1];
+								break;
+							default:
+								p += p[1];
+								break;
+						}
+						
+					}
+				}
+				else t += t[1];
+				if ( *m == GAMMA && m[1] != FUNHEAD+1 ) {
+					i = oldt[1] - m[1] - i;
+					if ( i > 0 ) {
+						*fill++ = GAMMA;
+						*fill++ = i + FUNHEAD+1;
+						FILLFUN(fill)
+						*fill++ = oldt[FUNHEAD];
+						t = t - i;
+						NCOPY(fill,t,i);
+					}
+				}
+				break;
+			}
+		}
+		m += m[1];
+	}
+	while ( t < tstop ) *fill++ = *t++;
+	if ( !PutExpr ) {
+		t = AN.FullProto;
+		nq = t[1];
+		t[3] = 1;
+		NCOPY(fill,t,nq);
+	}
+	t = tcoef;
+	nq = ABS(*t);
+	t = tstop;
+	NCOPY(fill,t,nq);
+	if ( sign ) {
+		if ( ( sign & 1 ) != 0 ) fill[-1] = -fill[-1];
+	}
+	*TemTerm = fill-TemTerm;
+/*
+	And now we copy this to AT.allbufnum
+*/
+	AddNtoC(AT.allbufnum,TemTerm[0],TemTerm,12);
+	cbuf[AT.allbufnum].Pointer[0] = 0;
+	AN.RepFunNum = 0;
+}
+
+/*
+ 		#] SubsInAll : 
+ 		#[ TransferBuffer :
+
+		Adds the whole content of a (compiler)buffer to another buffer.
+		In spectator we have an expression in the RHS that needs the 
+		wildcard resolutions adapted by an offset.
+*/
+
+VOID TransferBuffer(int from,int to,int spectator)
+{
+	CBUF *C  = cbuf + spectator;
+	CBUF *Cf = cbuf + from;
+	CBUF *Ct = cbuf + to;
+	int offset = Ct->numrhs;
+	LONG i;
+	WORD *t, *tt, *ttt, *tstop, size;
+	for ( i = 1; i <= Cf->numrhs; i++ ) {
+		size = Cf->rhs[i+1]-Cf->rhs[i];
+		AddRHS(to,1);
+		AddNtoC(to,size,Cf->rhs[i],13);
+	}
+	Ct->rhs[Ct->numrhs+1] = Ct->Pointer;
+	Cf->numrhs = 0;
+/*
+	Now we have to update the 'pointers' in the spectator.
+*/
+	t = C->rhs[C->numrhs];
+	while ( *t ) {
+		tt = t+1; t += *t;
+		tstop = t-ABS(t[-1]);
+		while ( tt < tstop ) {
+			if ( *tt == SUBEXPRESSION ) {
+				ttt = tt+SUBEXPSIZE; tt += tt[1];
+				while ( ttt < tt ) {
+					switch ( *ttt ) {
+						case SYMTOSUB:
+						case VECTOSUB:
+						case INDTOSUB:
+						case ARGTOARG:
+						case ARLTOARL:
+							ttt[3] += offset;
+							break;
+						default:
+							break;
+					}
+					ttt += 4;
+				}
+			}
+			else tt += tt[1];
+		}
+	}
+}
+
+/*
+ 		#] TransferBuffer : 
+ 		#[ TakeIDfunction :
+*/
+
+#define PutInBuffers(pow) \
+	AddRHS(AT.ebufnum,1); \
+	*out++ = SUBEXPRESSION; \
+	*out++ = SUBEXPSIZE; \
+	*out++ = C->numrhs; \
+	*out++ = pow; \
+	*out++ = AT.ebufnum; \
+	FILLSUB(out) \
+	r = AT.pWorkSpace[rhs+i]; \
+	if ( *r > 0 ) { \
+		oldinr = r[*r]; r[*r] = 0; \
+		AddNtoC(AT.ebufnum,(*r+1-ARGHEAD),(r+ARGHEAD),14); \
+		r[*r] = oldinr; \
+	} \
+	else { \
+		ToGeneral(r,buffer,1); \
+		buffer[buffer[0]] = 0; \
+		AddNtoC(AT.ebufnum,buffer[0]+1,buffer,15); \
+	}
+
+int TakeIDfunction(PHEAD WORD *term)
+{
+	WORD *tstop, *t, *r, *m, *f, *nextf, *funstop, *left, *l, *newterm;
+	WORD *out, oldinr, pow;
+	WORD buffer[20];
+	int i, ii, j, numsub, numfound = 0, first;
+	LONG lhs,rhs;
+	CBUF *C;
+	GETSTOP(term,tstop);
+	for ( t = term+1; t < tstop; t += t[1] ) { if ( *t == IDFUNCTION ) break; }
+	if ( t >= tstop ) return(0);
+/*
+	Step 1: test validity
+*/
+	funstop = t + t[1]; f = t + FUNHEAD;
+	left = term + *term;
+	l = left+1; numsub = 0;
+	while ( f < funstop ) {
+		nextf = f; NEXTARG(nextf)
+		if ( nextf >= funstop ) { return(0); } /* odd number of arguments */
+		if ( *f == -SYMBOL ) { *l++ = SYMBOL; *l++ = 4; *l++ = f[1]; *l++ = 1; }
+		else if ( *f < -FUNCTION ) { *l++ = *f; *l++ = FUNHEAD; FILLFUN(l) }
+		else if ( *f > 0 ) {
+			if ( *f != f[ARGHEAD]+ARGHEAD ) goto noaction;
+			if ( nextf[-1] != 3 || nextf[-2] != 1 || nextf[-3] != 1 ) goto noaction;
+			if ( f[ARGHEAD] <= 4 ) goto noaction;
+			if ( f[ARGHEAD] != f[ARGHEAD+2]+4 ) goto noaction;
+			if ( f[ARGHEAD] == 8 && f[ARGHEAD+1] == SYMBOL ) {
+				for ( i = 0; i < 4; i++ ) *l++ = f[ARGHEAD+1+i];
+			}
+			else if ( f[ARGHEAD] == 9 && f[ARGHEAD+1] == DOTPRODUCT ) {
+				for ( i = 0; i < 5; i++ ) *l++ = f[ARGHEAD+1+i];
+			}
+			else if ( f[ARGHEAD+1] >= FUNCTION ) {
+				for ( i = 0; i < f[ARGHEAD+1]-4; i++ ) *l++ = f[ARGHEAD+1+i];
+			}
+			else goto noaction;
+		}
+		else goto noaction;
+		numsub++;
+		f = nextf;
+		NEXTARG(f)
+	}
+	C = cbuf+AT.ebufnum;
+	AT.WorkPointer = l;
+	*left = l-left;
+/*
+	Put the pointers to the lhs and the rhs in the pointer workspace
+*/
+	WantAddPointers(2*numsub);
+	lhs = AT.pWorkPointer;
+	rhs = lhs+numsub;
+	AT.pWorkPointer = rhs+numsub;
+	f = t + FUNHEAD; l = left+1;
+	for ( i = 0; i < numsub; i++ ) {
+		AT.pWorkSpace[lhs+i] = l; l += l[1];
+		NEXTARG(f);
+		AT.pWorkSpace[rhs+i] = f;
+		NEXTARG(f);
+	}
+/*
+	Take out the patterns and replace them by SUBEXPRESSIONs pointing at
+	the e buffer. We put the resulting term above the left sides.
+	Note that we take out only the first id_ if there is more than one!
+*/
+	first = 1;
+	t = term+1; newterm = AT.WorkPointer; out = newterm+1;
+	while ( t < tstop ) {
+		if ( *t == IDFUNCTION && first ) { first = 0; t += t[1]; continue; }
+		if ( *t >= FUNCTION ) {
+			for ( i = 0; i < numsub; i++ ) {
+				m = AT.pWorkSpace[lhs+i];
+				if ( *m != *t ) continue;
+				for ( j = 1; j < t[1]; j++ ) {
+					if ( m[j] != t[j] ) break;
+				}
+				if ( j != t[1] ) continue;
+				numfound++;
+/*
+				We have a match! Set up a SUBEXPRESSION subterm and put the
+				corresponding rhs in the eBuffer.
+*/
+				PutInBuffers(1)
+				t += t[1];
+			}
+			if ( i == numsub ) {	/* no match. Just copy to output. */
+				j = t[1]; NCOPY(out,t,j)
+			}
+		}
+		else if ( *t == SYMBOL ) {
+			for ( i = 0; i < numsub; i++ ) {
+				m = AT.pWorkSpace[lhs+i];
+				if ( *m != SYMBOL ) continue;
+				for ( ii = 2; ii < t[1]; ii += 2 ) {
+					if ( m[2] != t[ii] ) continue;
+					pow = t[ii+1]/m[3];
+					if ( pow <= 0 ) continue;
+					t[ii+1] = t[ii+1]%m[3];
+					numfound++;
+/*
+					Create the proper rhs in the eBuffer and set up a
+					SUBEXPRESSION subterm.
+*/
+					PutInBuffers(pow)
+				}
+			}
+/*
+			Now we copy whatever remains of the SYMBOL subterm to the output
+*/
+			m = out; *out++ = t[0]; *out++ = t[1];
+			for ( ii = 2; ii < t[1]; ii += 2 ) {
+				if ( t[ii+1] ) { *out++ = t[ii]; *out++ = t[ii+1]; }
+			}
+			m[1] = out-m;
+			if ( m[1] == 2 ) out = m;
+			t += t[1];
+		}
+		else if ( *t == DOTPRODUCT ) {
+			for ( i = 0; i < numsub; i++ ) {
+				m = AT.pWorkSpace[lhs+i];
+				if ( *m != DOTPRODUCT ) continue;
+				for ( ii = 2; ii < t[1]; ii += 3 ) {
+					if ( m[2] != t[ii] || m[3] != t[ii+1] ) continue;
+					pow = t[ii+2]/m[4];
+					if ( pow <= 0 ) continue;
+					t[ii+2] = t[ii+2]%m[4];
+					numfound++;
+/*
+					Create the proper rhs in the eBuffer and set up a
+					SUBEXPRESSION subterm.
+*/
+					PutInBuffers(pow)
+				}
+			}
+/*
+			Now we copy whatever remains of the DOTPRODUCT subterm to the output
+*/
+			m = out; *out++ = t[0]; *out++ = t[1];
+			for ( ii = 2; ii < t[1]; ii += 3 ) {
+				if ( t[ii+2] ) { *out++ = t[ii]; *out++ = t[ii+1]; *out++ = t[ii+2]; }
+			}
+			m[1] = out-m;
+			if ( m[1] == 2 ) out = m;
+			t += t[1];
+		}
+		else {
+			j = t[1]; NCOPY(out,t,j)
+		}
+	}
+/*
+	Copy the coefficient and set the size.
+*/
+	t = tstop; r = term+*term; while ( t < r ) *out++ = *t++;
+	*newterm = out-newterm;
+/*
+	Finally we move the new term over the original term.
+*/
+	i = *newterm;
+	t = term; r = newterm; NCOPY(t,r,i)
+/*
+	At this point we can return and if the calling Generator jumps back to
+	its start, TestSub can take care of the expansions of SUBEXPRESSIONs.
+*/
+	AT.pWorkPointer = lhs;
+	AT.WorkPointer = t;
+	return(numfound);
+noaction:
+	return(0);
+}
+
+/*
+ 		#] TakeIDfunction : 
+  	#] Patterns : 
 */
 
