@@ -4,7 +4,7 @@
  */
 /* #[ License : */
 /*
- *   Copyright (C) 1984-2017 J.A.M. Vermaseren
+ *   Copyright (C) 1984-2023 J.A.M. Vermaseren
  *   When using this file you are requested to refer to the publication
  *   J.A.M.Vermaseren "New features of FORM" math-ph/0010025
  *   This is considered a matter of courtesy as the development was paid
@@ -73,7 +73,7 @@ const string factorized_poly::tostring () const {
 		if (power[i]>1) {
 			res += "^";
 			char tmp[100];
-			sprintf (tmp,"%i",power[i]);
+			snprintf (tmp,100,"%i",power[i]);
 			res += tmp;
 		}
 	}
@@ -82,10 +82,10 @@ const string factorized_poly::tostring () const {
 	if (factor[0].modp>0) {
 		res += " (mod ";
 		char tmp[12];
-		sprintf (tmp,"%i",factor[0].modp);
+		snprintf (tmp,12,"%i",factor[0].modp);
 		res += tmp;
 		if (factor[0].modn > 1) {
-			sprintf (tmp,"%i",factor[0].modn);
+			snprintf (tmp,12,"%i",factor[0].modn);
 			res += "^";
 			res += tmp;
 		}
@@ -176,7 +176,7 @@ const vector<poly> polyfact::extended_gcd_Euclidean_lifted (const poly &a, const
 	sa /= s.integer_lcoeff();
 	sb /= s.integer_lcoeff();
 
-	// Lift the result to modolu p^n with p-adic Newton's iteration.
+	// Lift the result to modulo p^n with p-adic Newton's iteration.
 	poly samodp(sa);
 	poly sbmodp(sb);
 	poly term(BHEAD 1);
