@@ -7,7 +7,7 @@
 
 /* #[ License : */
 /*
- *   Copyright (C) 1984-2013 J.A.M. Vermaseren
+ *   Copyright (C) 1984-2023 J.A.M. Vermaseren
  *   When using this file you are requested to refer to the publication
  *   J.A.M.Vermaseren "New features of FORM" math-ph/0010025
  *   This is considered a matter of courtesy as the development was paid
@@ -40,6 +40,7 @@
 #include <vector>
 #include <map>
 #include <climits>
+#include <cassert>
 
 //#define DEBUG
 
@@ -47,6 +48,9 @@
 #include "mytime.h"
 #endif
 
+// denompower is increased by this factor when divmod_heap fails
+const int POLYWRAP_DENOMPOWER_INCREASE_FACTOR = 2;
+
 using namespace std;
 
 /*
@@ -117,11 +121,28 @@ WORD poly_determine_modulus (PHEAD bool multi_error, bool is_fun_arg, string mes
  *   - Called from ratio.c
  *   - Calls polygcd::gcd
  */
-WORD *poly_gcd(PHEAD WORD *a, WORD *b) {
+WORD *poly_gcd(PHEAD WORD *a, WORD *b, WORD fit) {
 
 #ifdef DEBUG
 	cout << "*** [" << thetime() << "]  CALL : poly_gcd" << endl;
 #endif
+
+//
+//MesPrint("Calling poly_gcd with:");
+//{
+//	WORD *at = a;
+//	MesPrint("      a:");
+//	while ( *at ) {
+//		MesPrint("   %a",*at,at);
+//		at += *at;
+//	}
+//	MesPrint("      b:");
+//	at = b;
+//	while ( *at ) {
+//		MesPrint("   %a",*at,at);
+//		at += *at;
+//	}
+//}
 	
 	// Extract variables
 	vector<WORD *> e;
@@ -140,11 +161,23 @@ WORD *poly_gcd(PHEAD WORD *a, WORD *b) {
 	poly gcd(polygcd::gcd(pa,pb));
 	
 	// Allocate new memory and convert to Form notation
-	WORD *res = (WORD *)Malloc1((gcd.size_of_form_notation()+1)*sizeof(WORD), "poly_gcd");
+	int newsize = (gcd.size_of_form_notation()+1);
+	WORD *res;
+	if ( fit ) {
+		if ( newsize*sizeof(WORD) >= (size_t)(AM.MaxTer) ) {
+			MLOCK(ErrorMessageLock);
+			MesPrint("poly_gcd: Term too complex. Maybe increasing MaxTermSize can help");
+			MUNLOCK(ErrorMessageLock);
+			Terminate(-1);
+		}
+		res = TermMalloc("poly_gcd");
+	}
+	else {
+		res = (WORD *)Malloc1(newsize*sizeof(WORD), "poly_gcd");
+	}
 	poly::poly_to_argument(gcd, res, false);
 
-	if (AN.poly_num_vars > 0)
-		M_free(AN.poly_vars, "AN.poly_vars");
+	poly_free_poly_vars(BHEAD "AN.poly_vars_qcd");
 
 	// reset modulo calculation
 	AN.ncmod = AC.ncmod;
@@ -153,10 +186,12 @@ WORD *poly_gcd(PHEAD WORD *a, WORD *b) {
 
 /*
   	#] poly_gcd : 
-  	#[ poly_divmod : 
+  	#[ poly_divmod :
+
+	if fit == 1 the answer must fit inside a term.
 */
 
-WORD *poly_divmod(PHEAD WORD *a, WORD *b, int divmod) {
+WORD *poly_divmod(PHEAD WORD *a, WORD *b, int divmod, WORD fit) {
 
 #ifdef DEBUG
 	cout << "*** [" << thetime() << "]  CALL : poly_divmod" << endl;
@@ -173,15 +208,15 @@ WORD *poly_divmod(PHEAD WORD *a, WORD *b, int divmod) {
 
 	// add extra variables to keep track of denominators
 	const int DENOMSYMBOL = MAXPOSITIVE;
-	const int DENOMPOWER  = MAXPOSITIVE;
 	
-	WORD *new_poly_vars = (WORD *)Malloc1((AN.poly_num_vars+1)*sizeof(WORD), "AN.poly_vars");
-	WCOPY(new_poly_vars, AN.poly_vars, AN.poly_num_vars);
-	new_poly_vars[AN.poly_num_vars] = DENOMSYMBOL;
-	if (AN.poly_num_vars > 0)
-		M_free(AN.poly_vars, "AN.poly_vars");
-	AN.poly_num_vars++;
-	AN.poly_vars = new_poly_vars;
+//	WORD *new_poly_vars = (WORD *)Malloc1((AN.poly_num_vars+1)*sizeof(WORD), "AN.poly_vars");
+//	WCOPY(new_poly_vars, AN.poly_vars, AN.poly_num_vars);
+//	new_poly_vars[AN.poly_num_vars] = DENOMSYMBOL;
+//	if (AN.poly_num_vars > 0)
+//		M_free(AN.poly_vars, "AN.poly_vars");
+//	AN.poly_num_vars++;
+//	AN.poly_vars = new_poly_vars;
+	AN.poly_vars[AN.poly_num_vars++] = DENOMSYMBOL;
 	
 	// convert to polynomials
 	poly dena(BHEAD 0);
@@ -209,7 +244,8 @@ WORD *poly_divmod(PHEAD WORD *a, WORD *b, int divmod) {
 
 	// determine lcoeff(b)
 	poly lcoeffb(pb.integer_lcoeff());
-	int denompower = 0;
+
+	poly pres(BHEAD 0);
 
 	if (!lcoeffb.is_one()) {
 
@@ -217,52 +253,89 @@ WORD *poly_divmod(PHEAD WORD *a, WORD *b, int divmod) {
 			// the original polynomial is multivariate (one dummy variable has
 			// been added), so it is not trivial to determine which power of
 			// lcoeff(b) can be in the answer
-			
-			// multiply a by DENOMSYMBOL^DENOMPOWER
-			poly modifya(poly::simple_poly(BHEAD AN.poly_num_vars-1,0,DENOMPOWER));
-			pa *= modifya;
-			
-			// replace lcoeff(b) by DENOMSYMBOL		
-			poly modifyb(BHEAD 1);
-			for (int i=0; i<AN.poly_num_vars; i++)
-				modifyb[2+i] = pb[2+i];
-			modifyb *= poly::simple_poly(BHEAD AN.poly_num_vars-1,0,1) - lcoeffb;
-			pb += modifyb;
-			
-			// divide and determine the power
-			poly ppow(pa/pb);
-			for (int i=1; i<ppow[0]; i+=ppow[i])
-				denompower = max(denompower, DENOMPOWER - ppow[i+AN.poly_num_vars]);
 
-			// restore a,b
-			pb -= modifyb;	
-			pa /= poly::simple_poly(BHEAD AN.poly_num_vars-1,0,DENOMPOWER);			
+			int denompower = 0, prevdenompower = 0;
+			poly pq(BHEAD 0);
+			poly pr(BHEAD 0);
+
+			// try denompower = 0, if this fails increase denompower until division succeeds
+			bool div_fail = true;
+			do
+			{
+				if(denompower < prevdenompower)
+				{
+					// denompower increased beyond INT_MAX
+					MLOCK(ErrorMessageLock);
+					MesPrint ((char*)"ERROR: pseudo-division failed in poly_divmod (denompower > INT_MAX)");
+					MUNLOCK(ErrorMessageLock);
+					Terminate(1);
+				}
+
+				if(denompower != 0)
+				{
+					// multiply a by lcoeffb^(denompower-prevdenompower)
+					WORD n = lcoeffb[lcoeffb[1]];
+					RaisPow(BHEAD (UWORD *)&lcoeffb[2+AN.poly_num_vars], &n, denompower-prevdenompower);
+					lcoeffb[1] = 2 + AN.poly_num_vars + ABS(n);
+					lcoeffb[0] = 1 + lcoeffb[1];
+					lcoeffb[lcoeffb[1]] = n;
+
+					pa *= lcoeffb;
+					denres *= lcoeffb;
+				}
+
+				// try division
+				poly ppow(BHEAD 0);
+				poly::divmod_heap(pa,pb,pq,pr,false,true,div_fail); // sets div_fail
+
+				// increase denompower for next iteration
+				prevdenompower = denompower;
+				denompower = (denompower==0 ? 1 : denompower*POLYWRAP_DENOMPOWER_INCREASE_FACTOR+1 ); // generates 2^n-1 when POLYWRAP_DENOMPOWER_INCREASE_FACTOR = 2
+			}
+			while(div_fail);
+
+			pres = (divmod==0 ? pq : pr);
+
 		}
 		else {
 			// one variable, so the power is the difference of the degrees
-			
-			denompower = MaX(0, pa.degree(0) - pb.degree(0) + 1);
+
+			int denompower = MaX(0, pa.degree(0) - pb.degree(0) + 1);
+
+			// multiply a by that power
+			WORD n = lcoeffb[lcoeffb[1]];
+			RaisPow(BHEAD (UWORD *)&lcoeffb[2+AN.poly_num_vars], &n, denompower);
+			lcoeffb[1] = 2 + AN.poly_num_vars + ABS(n);
+			lcoeffb[0] = 1 + lcoeffb[1];
+			lcoeffb[lcoeffb[1]] = n;
+
+			pa *= lcoeffb;
+			denres *= lcoeffb;
+
+			pres = (divmod==0 ? pa/pb : pa%pb);
+
 		}
 
-		// multiply a by that power
-		WORD n = lcoeffb[lcoeffb[1]];
-		RaisPow(BHEAD (UWORD *)&lcoeffb[2+AN.poly_num_vars], &n, denompower);
-		lcoeffb[1] = 2 + AN.poly_num_vars + ABS(n);
-		lcoeffb[0] = 1 + lcoeffb[1];
-		lcoeffb[lcoeffb[1]] = n;
+	}
+	else {
+
+		pres = (divmod==0 ? pa/pb : pa%pb);
 
-		pa *= lcoeffb;
-		denres *= lcoeffb;
 	}
 
-	poly pres(divmod==0 ? pa/pb : pa%pb);
-
-	// convert to Form notation	
+	// convert to Form notation
+	// NOTE: this part can be rewritten with poly::size_of_form_notation_with_den()
+	// and poly::poly_to_argument_with_den().
 	WORD *res;
 
 	// special case: a=0
 	if (pres[0]==1) {
-		res = (WORD *)Malloc1(sizeof(WORD), "poly_divmod");
+		if ( fit ) {
+			res = TermMalloc("poly_divmod");
+		}
+		else {
+			res = (WORD *)Malloc1(sizeof(WORD), "poly_divmod");
+		}
 		res[0] = 0;
 	}
 	else {
@@ -274,8 +347,19 @@ WORD *poly_divmod(PHEAD WORD *a, WORD *b, int divmod) {
 		// allocate the memory; note that this overestimates the size,
 		// since the estimated denominators are too large
 		int ressize = pres.size_of_form_notation() + pres.number_of_terms()*2*ABS(denres[denres[1]]) + 1;
-		res = (WORD *)Malloc1(ressize*sizeof(WORD), "poly_divmod");
 
+		if ( fit ) {
+			if ( ressize*sizeof(WORD) > (size_t)(AM.MaxTer) ) {
+				MLOCK(ErrorMessageLock);
+				MesPrint("poly_divmod: Term too complex. Maybe increasing MaxTermSize can help");
+				MUNLOCK(ErrorMessageLock);
+				Terminate(-1);
+			}
+			res = TermMalloc("poly_divmod");
+		}
+		else {
+			res = (WORD *)Malloc1(ressize*sizeof(WORD), "poly_divmod");
+		}
 		int L=0;
 
 		for (int i=1; i!=pres[0]; i+=pres[i]) {
@@ -318,8 +402,7 @@ WORD *poly_divmod(PHEAD WORD *a, WORD *b, int divmod) {
 	}
 
 	// clean up
-	if (AN.poly_num_vars > 0)
-		M_free(AN.poly_vars, "AN.poly_vars");
+	poly_free_poly_vars(BHEAD "AN.poly_vars_divmod");
 
 	// reset modulo calculation
 	AN.ncmod = AC.ncmod;
@@ -338,13 +421,13 @@ WORD *poly_divmod(PHEAD WORD *a, WORD *b, int divmod) {
 	The answer should be a buffer (allocated by Malloc1) with a zero
 	terminated sequence of terms (or just zero).
 */
-WORD *poly_div(PHEAD WORD *a, WORD *b) {
+WORD *poly_div(PHEAD WORD *a, WORD *b, WORD fit) {
 
 #ifdef DEBUG
 	cout << "*** [" << thetime() << "]  CALL : poly_div" << endl;
 #endif
 	
-	return poly_divmod(BHEAD a, b, 0);
+	return poly_divmod(BHEAD a, b, 0, fit);
 }
 
 /*
@@ -359,13 +442,13 @@ WORD *poly_div(PHEAD WORD *a, WORD *b) {
 	The answer should be a buffer (allocated by Malloc1) with a zero
 	terminated sequence of terms (or just zero).
 */
-WORD *poly_rem(PHEAD WORD *a, WORD *b) {
+WORD *poly_rem(PHEAD WORD *a, WORD *b, WORD fit) {
 
 #ifdef DEBUG
 	cout << "*** [" << thetime() << "]  CALL : poly_rem" << endl;
 #endif
 
-	return poly_divmod(BHEAD a, b, 1);
+	return poly_divmod(BHEAD a, b, 1, fit);
 }
 
 /*
@@ -379,7 +462,7 @@ WORD *poly_rem(PHEAD WORD *a, WORD *b) {
  *   ===========
  *   This method reads a polyratfun starting at the pointer a. The
  *   resulting numerator and denominator are written in num and
- *   den. If not CLEANPRF, the result is normalized.
+ *   den. If MUSTCLEANPRF, the result is normalized.
  *
  *   Notes
  *   =====
@@ -397,7 +480,7 @@ void poly_ratfun_read (WORD *a, poly &num, poly &den) {
 	
 	WORD *astop = a+a[1];
 
-	bool clean = (a[2] & CLEANPRF) != 0;
+	bool clean = (a[2] & MUSTCLEANPRF) == 0;
 		
 	a += FUNHEAD;
 	if (a >= astop) {
@@ -448,6 +531,7 @@ void poly_ratfun_read (WORD *a, poly &num, poly &den) {
 
 		num *= den_den;
 		den *= den_num;
+
 		poly gcd = polygcd::gcd(num,den);
 		num /= gcd;
 		den /= gcd;
@@ -475,24 +559,23 @@ void poly_sort(PHEAD WORD *a) {
 #ifdef DEBUG
 	cout << "*** [" << thetime() << "]  CALL : poly_sort" << endl;
 #endif
-	
 	if (NewSort(BHEAD0)) { Terminate(-1); }
-	AR.CompareRoutine = (void *)&CompareSymbols;
+	AR.CompareRoutine = &CompareSymbols;
 	
 	for (int i=ARGHEAD; i<a[0]; i+=a[i]) {
 		if (SymbolNormalize(a+i)<0 || StoreTerm(BHEAD a+i)) {
-			AR.CompareRoutine = (void *)&Compare1;
+			AR.CompareRoutine = &Compare1;
 			LowerSortLevel();
 			Terminate(-1);
 		}
 	}
 	
 	if (EndSort(BHEAD a+ARGHEAD,1) < 0) {
-		AR.CompareRoutine = (void *)&Compare1;
+		AR.CompareRoutine = &Compare1;
 		Terminate(-1);
 	}
 	
-	AR.CompareRoutine = (void *)&Compare1;
+	AR.CompareRoutine = &Compare1;
 	a[1] = 0; // set dirty flag to zero
 }
 
@@ -515,11 +598,13 @@ void poly_sort(PHEAD WORD *a) {
  *   - Calls poly::operators and polygcd::gcd
  */
 WORD *poly_ratfun_add (PHEAD WORD *t1, WORD *t2) {
+ 
+	if ( AR.PolyFunExp == 1 ) return PolyRatFunSpecial(BHEAD t1, t2);
 
 #ifdef DEBUG
 	cout << "*** [" << thetime() << "]  CALL : poly_ratfun_add" << endl;
 #endif
-	
+
 	WORD *oldworkpointer = AT.WorkPointer;
 	
 	// Extract variables
@@ -550,14 +635,22 @@ WORD *poly_ratfun_add (PHEAD WORD *t1, WORD *t2) {
 	// Calculate result
 	if (den1 != den2) {
 		gcd = polygcd::gcd(den1,den2);
+#ifdef OLDADDITION
 		num = num1*(den2/gcd) + num2*(den1/gcd);
 		den = (den1/gcd)*den2;
+		gcd = polygcd::gcd(num,den);
+#else
+		den = den1/gcd;
+		num = num1*(den2/gcd) + num2*den;
+		den = den*den2;
+		gcd = polygcd::gcd(num,gcd);
+#endif
 	}
 	else {
 		num = num1 + num2;
 		den = den1;
+		gcd = polygcd::gcd(num,den);
 	}
-	gcd = polygcd::gcd(num,den);
 
 	num /= gcd;
 	den /= gcd;
@@ -569,6 +662,8 @@ WORD *poly_ratfun_add (PHEAD WORD *t1, WORD *t2) {
 	if (num.size_of_form_notation() + den.size_of_form_notation() + 3 >= AM.MaxTer/(int)sizeof(WORD)) {
 		MLOCK(ErrorMessageLock);
 		MesPrint ("ERROR: PolyRatFun doesn't fit in a term");
+		MesPrint ("(1) num size = %d, den size = %d,  MaxTer = %d",num.size_of_form_notation(),
+				den.size_of_form_notation(),AM.MaxTer);
 		MUNLOCK(ErrorMessageLock);
 		Terminate(-1);
 	}
@@ -578,7 +673,8 @@ WORD *poly_ratfun_add (PHEAD WORD *t1, WORD *t2) {
 
 	*t++ = AR.PolyFun;                   // function 
 	*t++ = 0;                            // length (to be determined)
-	*t++ = CLEANPRF;                     // clean polyratfun
+//	*t++ &= ~MUSTCLEANPRF;               // clean polyratfun
+	*t++ = 0;
 	FILLFUN3(t);                         // header
 	poly::poly_to_argument(num,t, true); // argument 1 (numerator)
 	if (*t>0 && t[1]==DIRTYFLAG)          // to Form order
@@ -592,8 +688,7 @@ WORD *poly_ratfun_add (PHEAD WORD *t1, WORD *t2) {
 	oldworkpointer[1] = t - oldworkpointer; // length
 	AT.WorkPointer = t;
 
-	if (AN.poly_num_vars > 0) 
-		M_free(AN.poly_vars, "AN.poly_vars");
+	poly_free_poly_vars(BHEAD "AN.poly_vars_ratfun_add");
 
 	// reset modulo calculation
 	AN.ncmod = AC.ncmod;
@@ -610,7 +705,7 @@ WORD *poly_ratfun_add (PHEAD WORD *t1, WORD *t2) {
  *
  *   Description
  *   ===========
- *   This method seaches a term for multiple polyratfuns and
+ *   This method searches a term for multiple polyratfuns and
  *   multiplies their contents. The result is properly
  *   normalized. Normalization also works for terms with a single
  *   polyratfun.
@@ -638,29 +733,43 @@ int poly_ratfun_normalize (PHEAD WORD *term) {
 	for (WORD *t=term+1; t<tstop; t+=t[1]) 
 		if (*t == AR.PolyFun) {
 			num_polyratfun++;
-			if ((t[2] & CLEANPRF) == 0)
+			if ((t[2] & MUSTCLEANPRF) != 0)
 				num_polyratfun = INT_MAX;
 			if (num_polyratfun > 1) break;
 		}
 		
 	if (num_polyratfun <= 1) return 0;
+
+	WORD oldsorttype = AR.SortType;
+	AR.SortType = SORTHIGHFIRST;
+
+/*
+	When there are polyratfun's with only one variable: rename them
+	temporarily to TMPPOLYFUN.
+*/
+	for (WORD *t=term+1; t<tstop; t+=t[1]) {
+		if (*t == AR.PolyFun && (t[1] == FUNHEAD+t[FUNHEAD]
+			|| t[1] == FUNHEAD+2 ) ) { *t = TMPPOLYFUN; }
+	}
+
 	
 	// Extract all variables in the polyfuns
 	vector<WORD *> e;
 	
-	for (WORD *t=term+1; t<tstop; t+=t[1])
+	for (WORD *t=term+1; t<tstop; t+=t[1]) {
 		if (*t == AR.PolyFun) 
 			for (WORD *t2 = t+FUNHEAD; t2<t+t[1];) {
 				e.push_back(t2);
 				NEXTARG(t2);
 			}		
-
+	}
 	poly::get_variables(BHEAD e, true, true);
 
 	// Check for modulus calculus
 	WORD modp=poly_determine_modulus(BHEAD true, true, "PolyRatFun");
 	
 	// Accumulate total denominator/numerator and copy the remaining terms
+	// We start with 'trivial' polynomials
 	poly num1(BHEAD (UWORD *)tstop, ncoeff/2, modp, 1);
 	poly den1(BHEAD (UWORD *)tstop+ABS(ncoeff/2), ABS(ncoeff)/2, modp, 1);
 
@@ -672,8 +781,12 @@ int poly_ratfun_normalize (PHEAD WORD *term) {
 			poly num2(BHEAD 0,modp,1);
 			poly den2(BHEAD 0,modp,1);
 			poly_ratfun_read(t,num2,den2);
+			if ((t[2] & MUSTCLEANPRF) != 0) { // first normalize
+				poly gcd1(polygcd::gcd(num2,den2));
+				num2 = num2/gcd1;
+				den2 = den2/gcd1;
+			}
 			t += t[1];
-
 			poly gcd1(polygcd::gcd(num1,den2));
 			poly gcd2(polygcd::gcd(num2,den1));
 
@@ -685,7 +798,7 @@ int poly_ratfun_normalize (PHEAD WORD *term) {
 			if ( s != t ) {	NCOPY(s,t,i) }
 			else { t += i; s += i; }
 		}			
-	
+
 	// Fix sign
 	if (den1.sign() == -1) { num1*=poly(BHEAD -1); den1*=poly(BHEAD -1); }
 
@@ -693,6 +806,8 @@ int poly_ratfun_normalize (PHEAD WORD *term) {
 	if (num1.size_of_form_notation() + den1.size_of_form_notation() + 3 >= AM.MaxTer/(int)sizeof(WORD)) {
 		MLOCK(ErrorMessageLock);
 		MesPrint ("ERROR: PolyRatFun doesn't fit in a term");
+		MesPrint ("(2) num size = %d, den size = %d,  MaxTer = %d",num1.size_of_form_notation(),
+				den1.size_of_form_notation(),AM.MaxTer);
 		MUNLOCK(ErrorMessageLock);
 		Terminate(-1);
 	}
@@ -701,7 +816,7 @@ int poly_ratfun_normalize (PHEAD WORD *term) {
 	WORD *t = s;
 	*t++ = AR.PolyFun;                   // function
 	*t++ = 0;                            // size (to be determined)
-	*t++ = CLEANPRF;                     // clean polyratfun
+	*t++ &= ~MUSTCLEANPRF;                   // clean polyratfun
 	FILLFUN3(t);                         // header
 	poly::poly_to_argument(num1,t,true); // argument 1 (numerator)
 	if (*t>0 && t[1]==DIRTYFLAG)         // to Form order
@@ -720,12 +835,17 @@ int poly_ratfun_normalize (PHEAD WORD *term) {
 	
 	term[0] = t-term;                    // term length
 
-	if (AN.poly_num_vars > 0) 
-		M_free(AN.poly_vars, "AN.poly_vars");
+	poly_free_poly_vars(BHEAD "AN.poly_vars_ratfun_normalize");
 
 	// reset modulo calculation
 	AN.ncmod = AC.ncmod;
-	
+
+	tstop = term + *term; tstop -= ABS(tstop[-1]);
+	for (WORD *t=term+1; t<tstop; t+=t[1]) {
+		if (*t == TMPPOLYFUN ) *t = AR.PolyFun;
+	}
+
+	AR.SortType = oldsorttype;
 	return 0;
 }
 
@@ -736,6 +856,8 @@ int poly_ratfun_normalize (PHEAD WORD *term) {
 
 void poly_fix_minus_signs (factorized_poly &a) {
 
+	if ( a.factor.empty() ) return;
+
 	POLY_GETIDENTITY(a.factor[0]);
 	
 	int overall_sign = +1;
@@ -897,8 +1019,7 @@ WORD *poly_factorize (PHEAD WORD *argin, WORD *argout, bool with_arghead, bool i
 	
 	*argout=0;
 
-	if (AN.poly_num_vars > 0)
-		M_free(AN.poly_vars, "AN.poly_vars");
+	poly_free_poly_vars(BHEAD "AN.poly_vars_factorize");
 
 	// reset modulo calculation
 	AN.ncmod = AC.ncmod;
@@ -1301,8 +1422,7 @@ int poly_factorize_expression(EXPRESSIONS expr) {
 	AT.bracketindexflag = oldbracketindexflag;
 	strcpy((char*)AC.Commercial, oldCommercial);
 	
-	if (AN.poly_num_vars > 0)
-		M_free(AN.poly_vars, "AN.poly_vars");
+	poly_free_poly_vars(BHEAD "AN.poly_vars_factorize_expression");
 
 	return 0;
 }
@@ -1660,8 +1780,7 @@ WORD *poly_inverse(PHEAD WORD *arga, WORD *argb) {
 	}
 
 	// clean up and reset modulo calculation
-	if (AN.poly_num_vars > 0)
-		M_free(AN.poly_vars, "AN.poly_vars");
+	poly_free_poly_vars(BHEAD "AN.poly_vars_inverse");
 	
 	AN.ncmod = AC.ncmod;
 	return res;
@@ -1669,4 +1788,80 @@ WORD *poly_inverse(PHEAD WORD *arga, WORD *argb) {
 
 /*
   	#] poly_inverse : 
+  	#[ poly_mul :
+*/
+
+WORD *poly_mul(PHEAD WORD *a, WORD *b) {
+
+#ifdef DEBUG
+	cout << "*** [" << thetime() << "]  CALL : poly_mul" << endl;
+#endif
+
+	// Extract variables
+	vector<WORD *> e;
+	e.push_back(a);
+	e.push_back(b);
+	poly::get_variables(BHEAD e, false, false);  // TODO: any performance effect by sort_vars=true?
+
+	// Convert to polynomials
+	poly dena(BHEAD 0);
+	poly denb(BHEAD 0);
+	poly pa(poly::argument_to_poly(BHEAD a, false, true, &dena));
+	poly pb(poly::argument_to_poly(BHEAD b, false, true, &denb));
+
+	// Check for modulus calculus
+	WORD modp = poly_determine_modulus(BHEAD true, true, "polynomial multiplication");
+	pa.setmod(modp, 1);
+
+	// NOTE: mul_ is currently implemented by translating negative powers of
+	// symbols to extra symbols. For future improvement, it may be better to
+	// compute
+	//   (content(a) * content(b)) * (a/content(a)) * (b/content(b))
+	// to avoid introducing extra symbols for "mixed" cases, e.g.,
+	// (1+x) * (1/x) -> (1+x) * (1+Z1_).
+	assert(dena.is_integer());
+	assert(denb.is_integer());
+	assert(modp == 0 || dena.is_one());
+	assert(modp == 0 || denb.is_one());
+
+	// multiplication
+	pa *= pb;
+
+	// convert to Form notation
+	WORD *res;
+	if (dena.is_one() && denb.is_one()) {
+		res = (WORD *)Malloc1((pa.size_of_form_notation() + 1) * sizeof(WORD), "poly_mul");
+		poly::poly_to_argument(pa, res, false);
+	} else {
+		dena *= denb;
+		res = (WORD *)Malloc1((pa.size_of_form_notation_with_den(dena[dena[1]]) + 1) * sizeof(WORD), "poly_mul");
+		poly::poly_to_argument_with_den(pa, dena[dena[1]], (const UWORD *)&dena[2+AN.poly_num_vars], res, false);
+	}
+
+	// clean up and reset modulo calculation
+	poly_free_poly_vars(BHEAD "AN.poly_vars_mul");
+	AN.ncmod = AC.ncmod;
+
+	return res;
+}
+
+/*
+  	#] poly_mul : 
+  	#[ poly_free_poly_vars :
+*/
+
+void poly_free_poly_vars(PHEAD const char *text)
+{
+	if ( AN.poly_vars_type == 0 ) {
+		TermFree(AN.poly_vars, text);
+	}
+	else {
+		M_free(AN.poly_vars, text);
+	}
+	AN.poly_num_vars = 0;
+	AN.poly_vars = 0;
+}
+
+/*
+  	#] poly_free_poly_vars : 
 */
