@@ -5,7 +5,7 @@
  */
 /* #[ License : */
 /*
- *   Copyright (C) 1984-2017 J.A.M. Vermaseren
+ *   Copyright (C) 1984-2023 J.A.M. Vermaseren
  *   When using this file you are requested to refer to the publication
  *   J.A.M.Vermaseren "New features of FORM" math-ph/0010025
  *   This is considered a matter of courtesy as the development was paid
@@ -357,7 +357,7 @@ VOID RatToLine(UWORD *a, WORD na)
 		  if ( adenom == 1 && a[na] == 1 ) {
 			LongToLine(a,anumer);
 			if ( anumer > 1 ) {
-				if ( AO.DoubleFlag == 2 ) { AddToLine((UBYTE *)".Q0"); }
+				if ( ( AO.DoubleFlag & 2 ) == 2 ) { AddToLine((UBYTE *)".Q0"); }
 				else { AddToLine((UBYTE *)".D0"); }
 			}
 		  }
@@ -366,7 +366,7 @@ VOID RatToLine(UWORD *a, WORD na)
 			AddToLine((UBYTE *)"(one/");
 			LongToLine(a,adenom);
 			if ( adenom > 1 ) {
-				if ( AO.DoubleFlag == 2 ) { AddToLine((UBYTE *)".Q0"); }
+				if ( ( AO.DoubleFlag & 2 ) == 2 ) { AddToLine((UBYTE *)".Q0"); }
 				else { AddToLine((UBYTE *)".D0"); }
 			}
 			AddToLine((UBYTE *)")");
@@ -375,14 +375,14 @@ VOID RatToLine(UWORD *a, WORD na)
 			if ( anumer > 1 || adenom > 1 ) {
 				LongToLine(a,anumer);
 				if ( anumer > 1 ) {
-					if ( AO.DoubleFlag == 2 ) { AddToLine((UBYTE *)".Q0"); }
+					if (  ( AO.DoubleFlag & 2 ) == 2 ) { AddToLine((UBYTE *)".Q0"); }
 					else { AddToLine((UBYTE *)".D0"); }
 				}
 				a += na;
 				AddToLine((UBYTE *)"/");
 				LongToLine(a,adenom);
 				if ( adenom > 1 ) {
-					if ( AO.DoubleFlag == 2 ) { AddToLine((UBYTE *)".Q0"); }
+					if (  ( AO.DoubleFlag & 2 ) == 2 ) { AddToLine((UBYTE *)".Q0"); }
 					else { AddToLine((UBYTE *)".D0"); }
 				}
 			}
@@ -411,8 +411,8 @@ VOID RatToLine(UWORD *a, WORD na)
 				}
 			}
 			else if ( AC.OutputMode == FORTRANMODE || AC.OutputMode == CMODE ) {
-				if ( AO.DoubleFlag == 2 ) { AddToLine((UBYTE *)".Q0/"); }
-				else if ( AO.DoubleFlag == 1 ) { AddToLine((UBYTE *)".D0/"); }
+				if ( ( AO.DoubleFlag & 2 ) == 2 ) { AddToLine((UBYTE *)".Q0/"); }
+				else if ( ( AO.DoubleFlag & 1 ) == 1 ) { AddToLine((UBYTE *)".D0/"); }
 				else { AddToLine((UBYTE *)"./"); }
 			}
 			else AddToLine((UBYTE *)"/");
@@ -426,12 +426,12 @@ VOID RatToLine(UWORD *a, WORD na)
 				}
 			}
 			else if ( AC.OutputMode == FORTRANMODE || AC.OutputMode == CMODE ) {
-				if ( AO.DoubleFlag == 2 ) { AddToLine((UBYTE *)".Q0"); }
-				else if ( AO.DoubleFlag == 1 ) { AddToLine((UBYTE *)".D0"); }
+				if ( ( AO.DoubleFlag & 2 ) == 2 ) { AddToLine((UBYTE *)".Q0"); }
+				else if ( ( AO.DoubleFlag & 1 ) == 1 ) { AddToLine((UBYTE *)".D0"); }
 				else { AddToLine((UBYTE *)"."); }
 			}
 		  }
-		  else if ( anumer > 1 && ( AC.OutputMode == FORTRANMODE
+		  else if ( ( anumer > 1 || ( AO.DoubleFlag & 4 ) == 4 ) && ( AC.OutputMode == FORTRANMODE
 		  || AC.OutputMode == CMODE ) ) {
 			if ( AC.OutputMode == FORTRANMODE && AC.IsFortran90 == ISFORTRAN90 ) {
 				if ( AC.Fortran90Kind ) {
@@ -441,8 +441,8 @@ VOID RatToLine(UWORD *a, WORD na)
 					AddToLine((UBYTE *)".");
 				}
 			}
-			else if ( AO.DoubleFlag == 2 ) { AddToLine((UBYTE *)".Q0"); }
-			else if ( AO.DoubleFlag == 1 ) { AddToLine((UBYTE *)".D0"); }
+			else if ( ( AO.DoubleFlag & 2 ) == 2 ) { AddToLine((UBYTE *)".Q0"); }
+			else if ( ( AO.DoubleFlag & 1 ) == 1 ) { AddToLine((UBYTE *)".D0"); }
 			else { AddToLine((UBYTE *)"."); }
 		  }
 		  else if ( AC.OutputMode == FORTRANMODE && AC.IsFortran90 == ISFORTRAN90 ) {
@@ -455,9 +455,10 @@ VOID RatToLine(UWORD *a, WORD na)
 		  }
 		  else if ( ( AC.OutputMode == FORTRANMODE || AC.OutputMode == CMODE )
 		  && AO.DoubleFlag ) {
-			if ( anumer == 1 && adenom == 1 && a[0] == 1 ) {}
-			else if ( AO.DoubleFlag == 2 ) { AddToLine((UBYTE *)".Q0"); }
-			else if ( AO.DoubleFlag == 1 ) { AddToLine((UBYTE *)".D0"); }
+			if ( anumer == 1 && adenom == 1 && a[0] == 1 &&
+				 ( AO.DoubleFlag & 4 ) == 0 ) {}
+			else if ( ( AO.DoubleFlag & 2 ) == 2 ) { AddToLine((UBYTE *)".Q0"); }
+			else if ( ( AO.DoubleFlag & 1 ) == 1 ) { AddToLine((UBYTE *)".D0"); }
 		  }
 		}
 	}
@@ -753,7 +754,14 @@ UBYTE *WrtPower(UBYTE *Out, WORD Power)
 		*Out++ = '*'; *Out++ = '*';
 	}
 	else if ( AC.OutputMode == CMODE ) *Out++ = ',';
-	else *Out++ = '^';
+	else {
+		UBYTE *Out1 = IsExponentSign();
+		if ( Out1 == 0 ) *Out++ = '^';
+		else {
+			while ( *Out1 ) *Out++ = *Out1++;
+			*Out = 0;
+		}
+	}
 	if ( Power >= 0 ) {
 		if ( Power < 2*MAXPOWER )
 			Out = NumCopy(Power,Out);
@@ -764,8 +772,8 @@ UBYTE *WrtPower(UBYTE *Out, WORD Power)
 		*Out = 0;
 	}
 	else {
-		if ( ( AC.OutputMode >= FORTRANMODE || AC.OutputMode >= PFORTRANMODE )
-		 && AC.OutputMode != CMODE )
+		if ( ( AC.OutputMode >= FORTRANMODE || AC.OutputMode >= PFORTRANMODE
+		 || AC.OutputMode >= REDUCEMODE ) && AC.OutputMode != CMODE )
 			*Out++ = '(';
 		*Out++ = '-';
 		if ( Power > -2*MAXPOWER )
@@ -773,7 +781,8 @@ UBYTE *WrtPower(UBYTE *Out, WORD Power)
 		else
 			Out = StrCopy(FindSymbol((WORD)((LONG)Power-2*MAXPOWER)),Out);
 /*			Out = StrCopy(VARNAME(symbols,(LONG)(-Power)-2*MAXPOWER),Out); */
-		if ( AC.OutputMode >= FORTRANMODE || AC.OutputMode >= PFORTRANMODE ) *Out++ = ')';
+		if ( AC.OutputMode >= FORTRANMODE || AC.OutputMode >= PFORTRANMODE
+		 || AC.OutputMode >= REDUCEMODE) *Out++ = ')';
 		*Out = 0;
 	}
 	return(Out);
@@ -827,6 +836,9 @@ VOID WriteLists()
 	OutScr = (UBYTE *)AT.WorkPointer + ( TOLONG(AT.WorkTop) - TOLONG(AT.WorkPointer) ) /2;
 	if ( AC.CodesFlag || AC.NamesFlag > 1 ) startvalue = 0;
 	else startvalue = FIRSTUSERSYMBOL;
+/*
+ 		#[ Symbols :
+*/
 	if ( ( j = NumSymbols ) > startvalue ) {
 		TokenToLine((UBYTE *)" Symbols");
 		*skip = 3;
@@ -867,6 +879,10 @@ VOID WriteLists()
 		*skip = 0;
 		FiniLine();
 	}
+/*
+ 		#] Symbols : 
+ 		#[ Indices :
+*/
 	if ( AC.CodesFlag || AC.NamesFlag > 1 ) startvalue = 0;
 	else startvalue = BUILTININDICES;
 	if ( ( j = NumIndices ) > startvalue ) {
@@ -897,6 +913,10 @@ VOID WriteLists()
 		*skip = 0;
 		FiniLine();
 	}
+/*
+ 		#] Indices : 
+ 		#[ Vectors :
+*/
 	if ( AC.CodesFlag || AC.NamesFlag > 1 ) startvalue = 0;
 	else startvalue = BUILTINVECTORS;
 	if ( ( j = NumVectors ) > startvalue ) {
@@ -912,7 +932,10 @@ VOID WriteLists()
 		*skip = 0;
 		FiniLine();
 	}
-
+/*
+ 		#] Vectors : 
+ 		#[ Functions :
+*/
 	if ( AC.CodesFlag || AC.NamesFlag > 1 ) startvalue = 0;
 	else startvalue = AM.NumFixedFunctions;
 	for ( k = 0; k < 2; k++ ) {
@@ -956,6 +979,10 @@ VOID WriteLists()
 		*skip = 0;
 		if ( first == 0 ) FiniLine();
 	}
+/*
+ 		#] Functions : 
+ 		#[ Sets :
+*/
 	if ( AC.CodesFlag || AC.NamesFlag > 1 ) startvalue = 0;
 	else startvalue = AM.NumFixedSets;
 	if ( ( j = AC.SetList.num ) > startvalue ) {
@@ -1086,6 +1113,10 @@ VOID WriteLists()
 		*skip = 0;
 		FiniLine();
 	}
+/*
+ 		#] Sets : 
+ 		#[ Expressions :
+*/
 	if ( AS.ExecMode ) {
 		e = Expressions;
 		j = NumExpressions;
@@ -1133,6 +1164,10 @@ VOID WriteLists()
 		*skip = 0;
 		FiniLine();
 	}
+/*
+ 		#] Expressions : 
+ 		#[ Dollars :
+*/
 
 	if ( AC.CodesFlag || AC.NamesFlag > 1 ) startvalue = 0;
 	else startvalue = BUILTINDOLLARS;
@@ -1185,6 +1220,9 @@ VOID WriteLists()
 		*skip = 0;
 		FiniLine();
 	}
+/*
+ 		#] Dollars : 
+*/
 
 	if ( AC.ncmod != 0 ) {
 		TokenToLine((UBYTE *)"All arithmetic is modulus ");
@@ -1235,7 +1273,7 @@ VOID WriteLists()
 			MesPrint("\nCurrently dictionary %s is active\n",
 				AO.Dictionaries[olddict-1]->name);
 		else
-			MesPrint("\nCurrently there is no actice dictionary\n");
+			MesPrint("\nCurrently there is no active dictionary\n");
 	}
 	if ( AC.CodesFlag ) {
 		if ( C->numlhs > 0 ) {
@@ -1757,7 +1795,15 @@ WORD WriteSubTerm(WORD *sterm, WORD first)
 			 || AC.OutputMode == REDUCEMODE )
 				TokenToLine((UBYTE *)")**(");
 			else if ( AC.OutputMode == CMODE ) TokenToLine((UBYTE *)",");
-			else TokenToLine((UBYTE *)")^(");
+			else {
+				UBYTE *Out1 = IsExponentSign();
+				if ( Out1 ) {
+					TokenToLine((UBYTE *)")");
+					TokenToLine(Out1);
+					TokenToLine((UBYTE *)"(");
+				}
+				else TokenToLine((UBYTE *)")^(");
+			}
 			NEXTARG(t)
 			WriteArgument(t);
 			TokenToLine((UBYTE *)")");
@@ -1787,7 +1833,9 @@ WORD WriteSubTerm(WORD *sterm, WORD first)
 			AC.outsidefun = oldoutsidefun;
 			TokenToLine((UBYTE *)")");
 			if ( sterm[3] != 1 ) {
-				TokenToLine((UBYTE *)"^");
+				UBYTE *Out1 = IsExponentSign();
+				if ( Out1 ) TokenToLine(Out1);
+				else TokenToLine((UBYTE *)"^");
 				Out = buffer;
 				NumCopy(sterm[3],Out);
 				TokenToLine(buffer);
@@ -1923,15 +1971,19 @@ WORD WriteInnerTerm(WORD *term, WORD first)
 	if ( AC.modpowers ) {
 		if ( n == 1 && *t == 1 && t > s ) first = 1;
 		else if ( ABS(AC.ncmod) == 1 ) {
+			UBYTE *Out1 = IsExponentSign();
 			LongToLine((UWORD *)AC.powmod,AC.npowmod);
-			TokenToLine((UBYTE *)"^");
+			if ( Out1 ) TokenToLine(Out1);
+			else TokenToLine((UBYTE *)"^");
 			TalToLine(AC.modpowers[(LONG)((UWORD)*t)]);
 			first = 0;
 		}
 		else {
 			LONG jj;
+			UBYTE *Out1 = IsExponentSign();
 			LongToLine((UWORD *)AC.powmod,AC.npowmod);
-			TokenToLine((UBYTE *)"^");
+			if ( Out1 ) TokenToLine(Out1);
+			else TokenToLine((UBYTE *)"^");
 			jj = (UWORD)*t;
 			if ( n == 2 ) jj += ((LONG)t[1])<<BITSINWORD;
 			if ( AC.modpowers[jj+1] == 0 ) {
@@ -1950,7 +2002,7 @@ WORD WriteInnerTerm(WORD *term, WORD first)
 				else IniLine(3);
 		}
 		if ( AO.CurrentDictionary > 0 ) TransformRational((UWORD *)t,n);
-		else                             RatToLine((UWORD *)t,n);
+		else                            RatToLine((UWORD *)t,n);
 		first = 0;
 	}
 	else first = 1;
@@ -2024,10 +2076,14 @@ WORD WriteInnerTerm(WORD *term, WORD first)
 						MesCall("WriteInnerTerm");
 						SETERROR(-1)
 					}
-					if ( AC.OutputMode == FORTRANMODE
-					 || AC.OutputMode == PFORTRANMODE ) { TokenToLine((UBYTE *)"**"); }
+					if ( AC.OutputMode == FORTRANMODE || AC.OutputMode == PFORTRANMODE
+					 || AC.OutputMode == REDUCEMODE ) { TokenToLine((UBYTE *)"**"); }
 					else if ( AC.OutputMode == CMODE ) { TokenToLine((UBYTE *)","); }
-					else { TokenToLine((UBYTE *)"^"); }
+					else {
+						UBYTE *Out1 = IsExponentSign();
+						if ( Out1 ) TokenToLine(Out1);
+						else TokenToLine((UBYTE *)"^");
+					}
 					TalToLine(pow);
 					if ( AC.OutputMode == CMODE ) TokenToLine((UBYTE *)")");
 				}
@@ -2077,7 +2133,7 @@ WORD WriteTerm(WORD *term, WORD *lbrac, WORD first, WORD prtf, WORD br)
 					b = AO.bracket + 1;
 					t = term + 1;
 					while ( n > 0 && ( *b++ == *t++ ) ) { n--; }
-					if ( n <= 0 && ( ( AO.InFbrack < AM.FortranCont )
+					if ( n <= 0 && ( ( AM.FortranCont <= 0 || AO.InFbrack < AM.FortranCont )
 					|| ( lowestlevel == 0 ) ) ) {
 /*
 						We continue inside a bracket.
@@ -2306,7 +2362,7 @@ WrtTmes:				t = term;
 		}
 	}
 	if ( !br ) AO.IsBracket = 0;
-	if ( ( AO.InFbrack >= AM.FortranCont ) && lowestlevel ) {
+	if ( ( AM.FortranCont > 0 && AO.InFbrack >= AM.FortranCont ) && lowestlevel ) {
 		if ( AC.OutputMode == CMODE ) TokenToLine((UBYTE *)";");
 		if ( ( AC.OutputMode == FORTRANMODE || AC.OutputMode == PFORTRANMODE )
 		 && !first ) {
