@@ -4,7 +4,7 @@
  */
 /* #[ License : */
 /*
- *   Copyright (C) 1984-2013 J.A.M. Vermaseren
+ *   Copyright (C) 1984-2023 J.A.M. Vermaseren
  *   When using this file you are requested to refer to the publication
  *   J.A.M.Vermaseren "New features of FORM" math-ph/0010025
  *   This is considered a matter of courtesy as the development was paid
@@ -66,9 +66,11 @@ SETUPPARAMETERS setupparameters[] =
 	,{(UBYTE *)"incdir",                     PATHVALUE, 0, (LONG)curdirp}
 	,{(UBYTE *)"indentspace",           NUMERICALVALUE, 0, (LONG)INDENTSPACE}
 	,{(UBYTE *)"insidefirst",               ONOFFVALUE, 0, (LONG)1}
+	,{(UBYTE *)"jumpratio",             NUMERICALVALUE, 0, (LONG)JUMPRATIO}
 	,{(UBYTE *)"largepatches",          NUMERICALVALUE, 0, (LONG)MAXPATCHES}
 	,{(UBYTE *)"largesize",             NUMERICALVALUE, 0, (LONG)LARGEBUFFER}
-	,{(UBYTE *)"maxnumbersize",         NUMERICALVALUE, 0, (LONG)MAXNUMBERSIZE}
+	,{(UBYTE *)"maxnumbersize",         NUMERICALVALUE, 0, (LONG)0}
+/*	,{(UBYTE *)"maxnumbersize",         NUMERICALVALUE, 0, (LONG)MAXNUMBERSIZE} */
 	,{(UBYTE *)"maxtermsize",           NUMERICALVALUE, 0, (LONG)MAXTER}
 	,{(UBYTE *)"maxwildcards",          NUMERICALVALUE, 0, (LONG)MAXWILDC}
 	,{(UBYTE *)"nospacesinnumbers",         ONOFFVALUE, 0, (LONG)0}
@@ -78,6 +80,7 @@ SETUPPARAMETERS setupparameters[] =
 	,{(UBYTE *)"nwritestatistics",          ONOFFVALUE, 0, (LONG)0}
 	,{(UBYTE *)"nwritethreadstatistics",    ONOFFVALUE, 0, (LONG)0}
 	,{(UBYTE *)"oldfactarg",                ONOFFVALUE, 0, (LONG)NEWFACTARG}
+	,{(UBYTE *)"oldgcd",                    ONOFFVALUE, 0, (LONG)1}
 	,{(UBYTE *)"oldorder",                  ONOFFVALUE, 0, (LONG)0}
 	,{(UBYTE *)"oldparallelstatistics",     ONOFFVALUE, 0, (LONG)0}
 	,{(UBYTE *)"parentheses",           NUMERICALVALUE, 0, (LONG)MAXPARLEVEL}
@@ -92,6 +95,7 @@ SETUPPARAMETERS setupparameters[] =
 	,{(UBYTE *)"smallsize",             NUMERICALVALUE, 0, (LONG)SMALLBUFFER}
 	,{(UBYTE *)"sortiosize",            NUMERICALVALUE, 0, (LONG)SORTIOSIZE}
 	,{(UBYTE *)"sorttype",                 STRINGVALUE, 0, (LONG)lowfirst}
+	,{(UBYTE *)"spectatorsize",         NUMERICALVALUE, 0, (LONG)SPECTATORSIZE}
 	,{(UBYTE *)"subfilepatches",        NUMERICALVALUE, 0, (LONG)SMAXFPATCHES}
 	,{(UBYTE *)"sublargepatches",       NUMERICALVALUE, 0, (LONG)SMAXPATCHES}
 	,{(UBYTE *)"sublargesize",          NUMERICALVALUE, 0, (LONG)SLARGEBUFFER}
@@ -110,6 +114,7 @@ SETUPPARAMETERS setupparameters[] =
     ,{(UBYTE *)"threadsortfilesynch",       ONOFFVALUE, 0, (LONG)0}
 	,{(UBYTE *)"totalsize",                 ONOFFVALUE, 0, (LONG)2}
 	,{(UBYTE *)"workspace",             NUMERICALVALUE, 0, (LONG)WORKBUFFER}
+	,{(UBYTE *)"wtimestats",                ONOFFVALUE, 0, (LONG)2}
 };
 
 /*
@@ -241,6 +246,7 @@ restart:;
 				if ( *s2 == 'K' ) { x = x * 1000; s2++; }
 				else if ( *s2 == 'M' ) { x = x * 1000000; s2++; }
 				else if ( *s2 == 'G' ) { x = x * 1000000000; s2++; }
+				else if ( *s2 == 'T' ) { x = x * 1000000000000; s2++; }
 				if ( *s2 && *s2 != ' ' && *s2 != '\t' ) {
 					MesPrint("%s: Numerical value expected for parameter %s"
 					,proop1[filetype],s1);
@@ -250,6 +256,7 @@ restart:;
 				sp->flags = USEDFLAG;
 				break;
 			case STRINGVALUE:
+				if ( StrICmp(s1,(UBYTE *)"tempsortdir") == 0 ) AM.havesortdir = 1;
 				s = s2; t = s2;
 				while ( *s ) {
 					if ( *s == ' ' || *s == '\t' ) break;
@@ -257,12 +264,29 @@ restart:;
 					*t++ = *s++;
 				}
 				*t = 0;
+				if ( sp->flags == USEDFLAG && sp->value != 0 )
+						M_free((VOID *)(sp->value),"Process option");
 				sp->value = (LONG)strDup1(s2,"Process option");
 				sp->flags = USEDFLAG;
 				break;
 			case PATHVALUE:
-				MesPrint("Setups: PATHVALUE not yet implemented");
-				error = 1; break;
+				if ( StrICmp(s1,(UBYTE *)"incdir") == 0 ) {
+					AM.IncDir = 0;
+				}
+				else if ( StrICmp(s1,(UBYTE *)"path") == 0 ) {
+					if ( AM.Path ) M_free(AM.Path,"path");
+					AM.Path = 0;
+				}
+				else {
+					MesPrint("Setups: %s not yet implemented",s1);
+					error = 1;
+					break;
+				}
+				if ( sp->flags == USEDFLAG && sp->value != 0 )
+					M_free((VOID *)(sp->value),"Process option");
+				sp->value = (LONG)strDup1(s2,"Process option");
+				sp->flags = USEDFLAG;
+				break;
 			case ONOFFVALUE:
 				if ( tolower(*s2) == 'o' && tolower(s2[1]) == 'n'
 				&& ( s2[2] == 0 || s2[2] == ' ' || s2[2] == '\t' ) )
@@ -375,15 +399,17 @@ int AllocSetups()
 {
 	SETUPPARAMETERS *sp;
 	LONG LargeSize, SmallSize, SmallEsize, TermsInSmall, IOsize;
-	int MaxPatches, MaxFpatches, error = 0;
+	int MaxPatches, MaxFpatches, error = 0, i, size;
 	UBYTE *s;
 #ifndef WITHPTHREADS
-	int j, size;
+	int j;
 #endif
 	sp = GetSetupPar((UBYTE *)"threads");
 	if ( sp->value > 0 ) AM.totalnumberofthreads = sp->value+1;
 
 	AM.OutBuffer = (UBYTE *)Malloc1(AM.OutBufSize+1,"OutputBuffer");
+	AP.PreAssignStack =(LONG *)Malloc1(AP.MaxPreAssignLevel*sizeof(LONG *),"PreAssignStack");
+	for ( i = 0; i < AP.MaxPreAssignLevel; i++ ) AP.PreAssignStack[i] = 0;
 	AC.iBuffer = (UBYTE *)Malloc1(AC.iBufferSize+1,"statement buffer");
 	AC.iStop = AC.iBuffer + AC.iBufferSize-2;
 	AP.preStart = (UBYTE *)Malloc1(AP.pSize,"instruction buffer");
@@ -455,19 +481,18 @@ int AllocSetups()
 	It seems better to fix it at its maximum value. This way we only worry
 	about maxtermsize. This can be understood better by the 'innocent' user.
 */
-	AM.MaxTal = (AM.MaxTer/sizeof(WORD)-2)/2;
-
+	if ( sp->value == 0 ) {
+		AM.MaxTal = (AM.MaxTer/sizeof(WORD)-2)/2;
+	}
+	else {
+		size = ( sp->value + 11 ) & (-4);
+		AM.MaxTal = size - 2;
+		if ( (size_t)AM.MaxTal > (size_t)((AM.MaxTer/sizeof(WORD)-2)/2) )
+					AM.MaxTal = (AM.MaxTer/sizeof(WORD)-2)/2;
+	}
 	AM.MaxTal &= -sizeof(WORD)*2;
+
 	sp->value = AM.MaxTal;
-/*
-	AT.n_coef = (WORD *)Malloc1(sizeof(WORD)*4*size+2,(char *)(sp->parameter));
-	AT.n_llnum = AT.n_coef + 2*AM.MaxTal;
-*/
-#ifdef WITHPTHREADS
-#else
-	AT.n_coef = (WORD *)Malloc1(sizeof(WORD)*4*AM.MaxTal+2,(char *)(sp->parameter));
-	AT.n_llnum = AT.n_coef + 2*AM.MaxTal;
-#endif
 	AC.cmod = (UWORD *)Malloc1(AM.MaxTal*4*sizeof(UWORD),(char *)(sp->parameter));
 	AM.gcmod = AC.cmod + AM.MaxTal;
 	AC.powmod = AM.gcmod + AM.MaxTal;
@@ -573,6 +598,8 @@ int AllocSetups()
 	AM.SMaxFpatches = sp->value;
 	sp = GetSetupPar((UBYTE *)"subsortiosize");
 	AM.SIOsize = sp->value;
+	sp = GetSetupPar((UBYTE *)"spectatorsize");
+	AM.SpectatorSize = sp->value;
 /*
 	The next code is just for the moment (26-jan-1997) because we have
 	the new parts combined with the old. Once the old parts are gone
@@ -599,6 +626,7 @@ int AllocSetups()
 	AM.CompressSize = sp->value;
 #ifndef WITHPTHREADS
 	AR.CompressBuffer = (WORD *)Malloc1((AM.CompressSize+10)*sizeof(WORD),"compresssize");
+	AR.CompressPointer = AR.CompressBuffer;
 	AR.ComprTop = AR.CompressBuffer + AM.CompressSize;
 #endif
 	sp = GetSetupPar((UBYTE *)"bracketindexsize");
@@ -637,7 +665,6 @@ int AllocSetups()
 
 	sp = GetSetupPar((UBYTE *)"continuationlines");
 	AM.FortranCont = sp->value;
-	if ( AM.FortranCont <= 0 ) AM.FortranCont = 1;
 	sp = GetSetupPar((UBYTE *)"oldorder");
 	AM.OldOrderFlag = sp->value;
 	sp = GetSetupPar((UBYTE *)"resettimeonclear");
@@ -646,6 +673,8 @@ int AllocSetups()
 	AO.NoSpacesInNumbers = AM.gNoSpacesInNumbers = AM.ggNoSpacesInNumbers = sp->value;
 	sp = GetSetupPar((UBYTE *)"indentspace");
 	AO.IndentSpace = AM.gIndentSpace = AM.ggIndentSpace = sp->value;
+	sp = GetSetupPar((UBYTE *)"jumpratio");
+	AM.jumpratio = sp->value;
 	sp = GetSetupPar((UBYTE *)"nwritestatistics");
 	AC.StatsFlag = AM.gStatsFlag = AM.ggStatsFlag = 1-sp->value;
 	sp = GetSetupPar((UBYTE *)"nwritefinalstatistics");
@@ -658,6 +687,11 @@ int AllocSetups()
 	AC.OldParallelStats = AM.gOldParallelStats = AM.ggOldParallelStats = sp->value;
 	sp = GetSetupPar((UBYTE *)"oldfactarg");
 	AC.OldFactArgFlag = AM.gOldFactArgFlag = AM.ggOldFactArgFlag = sp->value;
+	sp = GetSetupPar((UBYTE *)"oldgcd");
+	AC.OldGCDflag = AM.gOldGCDflag = AM.ggOldGCDflag = sp->value;
+	sp = GetSetupPar((UBYTE *)"wtimestats");
+	if ( sp->value == 2 ) sp->value = AM.ggWTimeStatsFlag;
+	AC.WTimeStatsFlag = AM.gWTimeStatsFlag = AM.ggWTimeStatsFlag = sp->value;
 	sp = GetSetupPar((UBYTE *)"sorttype");
 	if ( StrICmp((UBYTE *)"lowfirst",(UBYTE *)sp->value) == 0 ) {
 		AC.lSortType = SORTLOWFIRST;
@@ -764,10 +798,14 @@ VOID WriteSetup()
 					MesPrint("   %s: '%s'",sp->parameter,(UBYTE *)(AM.IncDir));
 					break;
 				}
+				/* fall through */
 			case STRINGVALUE:
 				if ( StrICmp(sp->parameter,(UBYTE *)"tempdir") == 0 && AM.TempDir ) {
 					MesPrint("   %s: '%s'",sp->parameter,(UBYTE *)(AM.TempDir));
 				}
+				else if ( StrICmp(sp->parameter,(UBYTE *)"tempsortdir") == 0 && AM.TempSortDir ) {
+					MesPrint("   %s: '%s'",sp->parameter,(UBYTE *)(AM.TempSortDir));
+				}
 				else {
 					MesPrint("   %s: '%s'",sp->parameter,(UBYTE *)(sp->value));
 				}
@@ -807,9 +845,8 @@ SORTING *AllocSort(LONG LargeSize, LONG SmallSize, LONG SmallEsize, LONG TermsIn
 	int i = 0, j = 0;
 	char *s;
 	if ( AM.S0 != 0 ) {
-		s = FG.fname; i = 0;
+		s = FG.fname2; i = 0;
 		while ( *s ) { s++; i++; }
-/*		i += 11; */
 		i += 16;
 	}
 	if ( MaxFpatches < 4 ) MaxFpatches = 4;
@@ -820,10 +857,12 @@ SORTING *AllocSort(LONG LargeSize, LONG SmallSize, LONG SmallEsize, LONG TermsIn
 	longerp += sizeof(WORD*) - (longerp%sizeof(WORD *));
 	longer++;
 	longer += sizeof(WORD*) - (longer%sizeof(WORD *));
+	if ( SmallSize < 16*AM.MaxTer ) SmallSize = 16*AM.MaxTer+16;
 	TermsInSmall = (TermsInSmall+15) & (-16L);
 	terms2insmall = 2*TermsInSmall;  /* Used to be just + 100 rather than *2 */
 	if ( SmallEsize < (SmallSize*3)/2 ) SmallEsize = (SmallSize*3)/2;
-	if ( SmallEsize < 3*AM.MaxTer ) SmallEsize = 3*AM.MaxTer;
+	if ( LargeSize > 0 && LargeSize < 2*SmallSize ) LargeSize = 2*SmallSize;
+/*	if ( SmallEsize < 3*AM.MaxTer ) SmallEsize = 3*AM.MaxTer; */
 	SmallEsize = (SmallEsize+15) & (-16L);
 	if ( LargeSize < 0 ) LargeSize = 0;
 	sortsize = sizeof(SORTING);
@@ -853,6 +892,9 @@ SORTING *AllocSort(LONG LargeSize, LONG SmallSize, LONG SmallEsize, LONG TermsIn
 		+2*sizeof(WORD *)*longer
 		+2*(longerp*(sizeof(WORD *)+sizeof(WORD)))
 		+(3*longerp+2)*sizeof(WORD)
+#ifdef WITHZLIB
+		+(2*longerp+4)*sizeof(WORD)
+#endif
 		+terms2insmall*sizeof(WORD *)
 		+terms2insmall*sizeof(WORD *)/2
 		+LargeSize
@@ -882,8 +924,8 @@ SORTING *AllocSort(LONG LargeSize, LONG SmallSize, LONG SmallEsize, LONG TermsIn
 	sort->used = sort->tree+longerp;
 #ifdef WITHZLIB
 	sort->fpcompressed = sort->used+longerp;
-	sort->fpincompressed = sort->fpcompressed+longerp;
-	sort->ktoi = sort->fpincompressed+longerp;
+	sort->fpincompressed = sort->fpcompressed+longerp+2;
+	sort->ktoi = sort->fpincompressed+longerp+2;
 	sort->zsparray = 0;
 #else
 	sort->ktoi = sort->used + longerp;
@@ -908,6 +950,7 @@ SORTING *AllocSort(LONG LargeSize, LONG SmallSize, LONG SmallEsize, LONG TermsIn
 #ifdef WITHZLIB
 /*	sort->file.ziosize = IOsize; */
 	sort->file.ziosize = IObuffersize*sizeof(WORD);
+	sort->file.ziobuffer = 0;
 #endif
 	if ( AM.S0 != 0 ) {
 		sort->file.name = (char *)(sort->file.PObuffer + IObuffersize);
@@ -917,6 +960,7 @@ SORTING *AllocSort(LONG LargeSize, LONG SmallSize, LONG SmallEsize, LONG TermsIn
 	sort->cBuffer = 0;
 	sort->cBufferSize = 0;
 	sort->f = 0;
+	sort->PolyWise = 0;
 
 	return(sort);
 }
@@ -932,9 +976,9 @@ VOID AllocSortFileName(SORTING *sort)
 	char *s, *t;
 /*
 		This is not the allocation before the tempfiles are determined.
-		Hence we can use the name in FG.fname and modify the tail
+		Hence we can use the name in FG.fname2 and modify the tail
 */
-	s = FG.fname; t = sort->file.name;
+	s = FG.fname2; t = sort->file.name;
 	while ( *s ) *t++ = *s++;
 #ifdef WITHPTHREADS
 	t[-2] = 'F';
@@ -951,34 +995,35 @@ VOID AllocSortFileName(SORTING *sort)
  		#[ AllocFileHandle :
 */
 
-FILEHANDLE *AllocFileHandle()
+FILEHANDLE *AllocFileHandle(WORD par,char *name)
 {
 	GETIDENTITY
-	LONG allocation;
+	LONG allocation, Ssize;
 	FILEHANDLE *fh;
 	int i = 0;
 	char *s, *t;
 
-	s = FG.fname; i = 0;
+	s = FG.fname2; i = 0;
 	while ( *s ) { s++; i++; }
-	i += 16;
-/*	i += 11; */
+	if ( par == 0 ) { i += 16; Ssize = AM.SIOsize; }
+	else { s = name; while ( *s ) { i++; s++; } i+= 2; Ssize = AM.SpectatorSize; }
 
-	allocation = sizeof(FILEHANDLE) + (AM.SIOsize+1)*sizeof(WORD) + i*sizeof(char);
+	allocation = sizeof(FILEHANDLE) + (Ssize+1)*sizeof(WORD) + i*sizeof(char);
 	fh = (FILEHANDLE *)Malloc1(allocation,"FileHandle");
 
 	fh->PObuffer = (WORD *)(fh+1);
-	fh->POstop = fh->PObuffer+AM.SIOsize;
-	fh->POsize = AM.SIOsize * sizeof(WORD);
+	fh->POstop = fh->PObuffer+Ssize;
+	fh->POsize = Ssize * sizeof(WORD);
 	fh->active = 0;
 	fh->handle = -1;
 	PUTZERO(fh->POposition);
 #ifdef WITHPTHREADS
 	fh->pthreadslock = dummylock;
 #endif
-	if ( AM.S0 != 0 ) {
+	if ( par == 0 ) {	/* sort file */
+	  if ( AM.S0 != 0 ) {
 		fh->name = (char *)(fh->POstop + 1);
-		s = FG.fname; t = fh->name;
+		s = FG.fname2; t = fh->name;
 		while ( *s ) *t++ = *s++;
 #ifdef WITHPTHREADS
 		t[-2] = 'F';
@@ -988,8 +1033,17 @@ FILEHANDLE *AllocFileHandle()
 		sprintf(t-1,"%d",AN.filenum);
 #endif
 		AN.filenum++;
+	  }
+	  else fh->name = 0;
+	}
+	else {	/* Spectator file */
+		fh->name = (char *)(fh->POstop + 1);
+		s = FG.fname; t = fh->name;
+		for ( i  = 0; i < FG.fnamebase; i++ ) *t++ = *s++;
+		s = name;
+		while ( *s ) *t++ = *s++;
+		*t = 0;
 	}
-	else fh->name = 0;
 	fh->POfill = fh->POfull = fh->PObuffer;
 	return(fh);
 }
@@ -1055,6 +1109,7 @@ int TryFileSetups()
 	eqnum = AC.CurrentStream->eqnum;
 	AC.NoShowInput = 1;
 	settop = buff + SETBUFSIZE;
+	if ( AC.CurrentStream->type == INPUTSTREAM && oldstreamposition == 0 ) AC.CurrentStream->fileposition = 0;
 	for(;;) {
 		c = GetInput();
 		if ( c == '*' || c == '\n' ) {
@@ -1089,11 +1144,18 @@ int TryFileSetups()
 		*s++ = '\n';
 		t = s = buff; /* name of the option */
 		while ( tolower(*s) >= 'a' && tolower(*s) <= 'z' ) s++;
-		*s++ = 0;
-		while ( *s == ' ' || *s == '\t' ) s++;
-		u = s; /* 'value' of the option */
-		while ( *s && *s != '\n' && *s != '\r' ) s++;
-		if ( *s ) *s++ = 0;
+		if ( *s != '\n' ) {
+			*s++ = 0;
+			while ( *s == ' ' || *s == '\t' ) s++;
+			u = s; /* 'value' of the option */
+			while ( *s && *s != '\n' && *s != '\r' ) s++;
+			if ( *s ) *s++ = 0;
+		}
+		else {
+			/* The value is empty. */
+			u = s;
+			*s++ = 0;
+		}
 		error += ProcessOption(t,u,1);
 nextline:;
 	}
@@ -1104,6 +1166,7 @@ nextline:;
 	AC.CurrentStream->prevline = prevline;
 	AC.CurrentStream->eqnum = eqnum;
 	ClearPushback();
+	if ( AC.CurrentStream->type == INPUTSTREAM ) AC.CurrentStream->fileposition = -1;
 	return(error);
 eoi:
 	MesPrint("Input file without a program.");
@@ -1139,4 +1202,3 @@ int TryEnvironment()
  		#] TryEnvironment : 
 	#] Setups :
 */
-
