@@ -2,12 +2,12 @@
  * 
  *  This file contains the main program.
  *  It also deals with the very early stages of the startup of FORM
- *	and the final stages when the program attemps some cleanup.
+ *	and the final stages when the program attempts some cleanup.
  *	Here is the routine that analyses the command tail.
  */
 /* #[ License : */
 /*
- *   Copyright (C) 1984-2013 J.A.M. Vermaseren
+ *   Copyright (C) 1984-2023 J.A.M. Vermaseren
  *   When using this file you are requested to refer to the publication
  *   J.A.M.Vermaseren "New features of FORM" math-ph/0010025
  *   This is considered a matter of courtesy as the development was paid
@@ -44,6 +44,12 @@
 #include <signal.h>
 #endif
 
+/*
+ * A macro for translating the contents of `x' into a string after expanding.
+ */
+#define STRINGIFY(x)  STRINGIFY__(x)
+#define STRINGIFY__(x) #x
+
 /*
  * FORMNAME = "FORM" or "TFORM" or "ParFORM".
  */
@@ -56,22 +62,141 @@
 #endif
 
 /*
- * VERSIONSTR = VERSION or VERSION"Beta".
+ * VERSIONSTR is the version information printed in the header line.
  */
-#ifdef BETAVERSION
-	#define VERSIONSTR VERSION"Beta"
+#ifdef HAVE_CONFIG_H
+	/* We have also version.h. */
+	#include "version.h"
+	#ifndef REPO_VERSION
+		#define REPO_VERSION STRINGIFY(REPO_MAJOR_VERSION) "." STRINGIFY(REPO_MINOR_VERSION)
+	#endif
+	#ifndef REPO_DATE
+		/* The build date, instead of the repo date. */
+		#define REPO_DATE __DATE__
+	#endif
+	#ifdef REPO_REVISION
+		#define VERSIONSTR FORMNAME " " REPO_VERSION " (" REPO_DATE ", " REPO_REVISION ")"
+	#else
+		#define VERSIONSTR FORMNAME " " REPO_VERSION " (" REPO_DATE ")"
+	#endif
+	#define MAJORVERSION REPO_MAJOR_VERSION
+	#define MINORVERSION REPO_MINOR_VERSION
 #else
-	#define VERSIONSTR VERSION
+	/*
+	 * Otherwise, form3.h defines MAJORVERSION, MINORVERSION and PRODUCTIONDATE,
+	 * possibly BETAVERSION.
+	 */
+	#ifdef BETAVERSION
+		#define VERSIONSTR__ STRINGIFY(MAJORVERSION) "." STRINGIFY(MINORVERSION) "Beta"
+	#else
+		#define VERSIONSTR__ STRINGIFY(MAJORVERSION) "." STRINGIFY(MINORVERSION)
+	#endif
+	#define VERSIONSTR FORMNAME " " VERSIONSTR__ " (" PRODUCTIONDATE ")"
 #endif
 
 /*
- * A macro for translating the contents of `x' into a string after expanding.
+ 		#] includes : 
+ 		#[ PrintHeader :
+*/
+
+/**
+ * Prints the header line of the output.
+ *
+ * @param  with_full_info  True for printing also runtime information.
  */
-#define STRINGIFY(x)  STRINGIFY__(x)
-#define STRINGIFY__(x) #x
+static void PrintHeader(int with_full_info)
+{
+#ifdef WITHMPI
+	if ( PF.me == MASTER && !AM.silent ) {
+#else
+	if ( !AM.silent ) {
+#endif
+		char buffer1[250], buffer2[80], *s = buffer1, *t = buffer2;
+		WORD length, n;
+		for ( n = 0; n < 250; n++ ) buffer1[n] = ' ';
+		/*
+		 * NOTE: we expect that the compiler optimizes strlen("string literal")
+		 * to just a number.
+		 */
+		if ( strlen(VERSIONSTR) <= 100 ) {
+			strcpy(s,VERSIONSTR);
+			s += strlen(VERSIONSTR);
+			*s = 0;
+		}
+		else {
+			/*
+			 * Truncate when it is too long.
+			 */
+			strncpy(s,VERSIONSTR,97);
+			s[97] = '.';
+			s[98] = '.';
+			s[99] = ')';
+			s[100] = '\0';
+			s += 100;
+		}
+
+		s += sprintf(s," %d-bits",(WORD)(sizeof(WORD)*16));
+/*		while ( *s ) s++; */
+		*s = 0;
+
+		if ( with_full_info ) {
+#if defined(WITHPTHREADS) || defined(WITHMPI)
+#if defined(WITHPTHREADS)
+			int nworkers = AM.totalnumberofthreads-1;
+#elif defined(WITHMPI)
+			int nworkers = PF.numtasks-1;
+#endif
+			s += sprintf(s," %d worker",nworkers);
+			*s = 0;
+/*			while ( *s ) s++; */
+			if ( nworkers != 1 ) {
+				*s++ = 's';
+				*s = '\0';
+			}
+#endif
+
+			sprintf(t,"Run: %s",MakeDate());
+			while ( *t ) t++;
+
+			/*
+			 * Align the date to the right, if it fits in a line.
+			 */
+			length = (s-buffer1) + (t-buffer2);
+			if ( length+2 <= AC.LineLength ) {
+				for ( n = AC.LineLength-length; n > 0; n-- ) *s++ = ' ';
+				*s = 0;
+				strcat(s,buffer2);
+				while ( *s ) s++;
+			}
+			else {
+				*s = 0;
+				strcat(s,"  ");
+				while ( *s ) s++;
+				*s = 0;
+				strcat(s,buffer2);
+				while ( *s ) s++;
+			}
+		}
+
+		/*
+		 * If the header information doesn't fit in a line, we need to extend
+		 * the line length temporarily.
+		 */
+		length = s-buffer1;
+		if ( length <= AC.LineLength ) {
+			MesPrint("%s",buffer1);
+		}
+		else {
+			WORD oldLineLength = AC.LineLength;
+			AC.LineLength = length;
+			MesPrint("%s",buffer1);
+			AC.LineLength = oldLineLength;
+		}
+	}
+}
 
 /*
- 		#] includes : 
+ 		#] PrintHeader : 
  		#[ DoTail :
 
 		Routine reads the command tail and handles the commandline options.
@@ -95,14 +220,13 @@ int DoTail(int argc, UBYTE **argv)
 	UBYTE *s, *t, *copy;
 	int threadnum = 0;
 	argc--; argv++;
+	AM.ClearStore = 0;
+	AM.TimeLimit = 0;
 	AM.LogType = -1;
 	AM.HoldFlag = AM.qError = AM.Interact = AM.FileOnlyFlag = 0;
 	AM.InputFileName = AM.LogFileName = AM.IncDir = AM.TempDir = AM.TempSortDir =
-#ifdef WITHMPI
-	AM.SetupDir = AM.SetupFile = 0;
-#else
 	AM.SetupDir = AM.SetupFile = AM.Path = 0;
-#endif
+	AM.FromStdin = 0;
 	if ( argc < 1 ) {
 		onlyversion = 0;
 		goto printversion;
@@ -114,6 +238,8 @@ int DoTail(int argc, UBYTE **argv)
 			switch (*s) {
 				case 'c': /* Error checking only */
 							AM.qError = 1;   break;
+				case 'C': /* Next arg is filename of log file */
+							TAKEPATH(AM.LogFileName)  break;
 				case 'D':
 				case 'd': /* Next arg is define preprocessor var. */
 							t = copy = strDup1(*argv,"Dotail");
@@ -165,6 +291,9 @@ int DoTail(int argc, UBYTE **argv)
 /*							if ( threadnum == 1 ) threadnum = 0; */
 							threadnum++;
 							break;
+				case 'W': /* Print the wall-clock time on the master. */
+							AM.ggWTimeStatsFlag = 1;
+							break;
 /*
 				case 'n':
 							Reserved for number of slaves without MPI
@@ -243,7 +372,15 @@ int DoTail(int argc, UBYTE **argv)
 				case 'S': /* Next arg is setup file */
 							TAKEPATH(AM.SetupFile) break;
 				case 't': /* Next arg is directory for temp files */
-							TAKEPATH(AM.TempDir)   break;
+							if ( s[1] == 's' ) {
+								s++;
+								AM.havesortdir = 1;
+								TAKEPATH(AM.TempSortDir)
+							}
+							else {
+								TAKEPATH(AM.TempDir)
+							}
+							break;
 				case 'T': /* Print the total size used at end of job */
 							AM.PrintTotalSize = 1; break;
 				case 'v':
@@ -251,18 +388,29 @@ printversion:;
 #ifdef WITHMPI
 							if ( PF.me == MASTER )
 #endif
-							{
-								char buffer[100], *s = buffer;
-								sprintf(s,"%s %s (%s)",FORMNAME,VERSIONSTR,PRODUCTIONDATE);
-								while ( *s ) s++;
-								sprintf(s," %d-bits",(WORD)(sizeof(WORD)*16));
-								while ( *s ) s++;
-								printf("%s\n",buffer);
-							}
+								PrintHeader(0);
 							if ( onlyversion ) return(1);
 							goto NoFile;
 				case 'y': /* Preprocessor dumps output. No compilation. */
 							AP.PreDebug = PREPROONLY;   break;
+				case 'z': /* The number following is a time limit in sec. */
+							t = s++;
+							AM.TimeLimit = 0;
+							while ( *s >= '0' && *s <= '9' )
+								AM.TimeLimit = 10*AM.TimeLimit + *s++ - '0';
+							break;
+				case 'Z': /* Removes the .str file on crash, no matter its contents */
+							AM.ClearStore = 1;   break;
+				case '\0': /* "-" to use STDIN for the input. */
+#ifdef WITHMPI
+							/* At the moment, ParFORM doesn't implement STDIN broadcasts. */
+							if ( PF.me == MASTER )
+								printf("Sorry, reading STDIN as input is currently not supported by ParFORM\n");
+							errorflag++;
+#endif
+							AM.FromStdin = 1;
+							AC.NoShowInput = 1; // disable input echoing by default
+							break;
 				default:
 						if ( FG.cTable[*s] == 1 ) {
 							AM.SkipClears = 0; t = s;
@@ -297,6 +445,10 @@ printversion:;
 	}
 	AM.totalnumberofthreads = threadnum;
 	if ( AM.InputFileName ) {
+		if ( AM.FromStdin ) {
+			printf("STDIN and the input filename cannot be specified simultaneously\n");
+			errorflag++;
+		}
 		s = AM.InputFileName;
 		while ( *s ) s++;
 		if ( s < AM.InputFileName+4 ||
@@ -307,7 +459,7 @@ printversion:;
 			while ( *s ) *t++ = *s++;
 			*t++ = '.'; *t++ = 'f'; *t++ = 'r'; *t++ = 'm'; *t = 0;
 		}
-		if ( AM.LogType >= 0 ) {
+		if ( AM.LogType >= 0 && AM.LogFileName == 0 ) {
 			AM.LogFileName = strDup1(AM.InputFileName,"name of logfile");
 			s = AM.LogFileName;
 			while ( *s ) s++;
@@ -325,6 +477,9 @@ printversion:;
 		}
 	}
 #endif
+	else if ( AM.FromStdin ) {
+		/* Do nothing. */
+	}
 	else {
 NoFile:
 #ifdef WITHMPI
@@ -334,6 +489,13 @@ NoFile:
 		errorflag++;
 	}
 	if ( AM.Path == 0 ) AM.Path = (UBYTE *)getenv("FORMPATH");
+	if ( AM.Path ) {
+		/*
+		 * AM.Path is taken from argv or getenv. Reallocate it to avoid invalid
+		 * frees when AM.Path has to be changed.
+		 */
+		AM.Path = strDup1(AM.Path,"DoTail Path");
+	}
 	return(errorflag);
 }
 
@@ -350,6 +512,13 @@ int OpenInput()
 	int oldNoShowInput = AC.NoShowInput;
 	UBYTE c;
 	if ( !AM.Interact ) {
+		if ( AM.FromStdin ) {
+			if ( OpenStream(0,INPUTSTREAM,0,PRENOACTION) == 0 ) {
+				Error0("Cannot open STDIN");
+				return(-1);
+			}
+		}
+		else {
 		if ( OpenStream(AM.InputFileName,FILESTREAM,0,PRENOACTION) == 0 ) {
 			Error1("Cannot open file",AM.InputFileName);
 			return(-1);
@@ -358,6 +527,7 @@ int OpenInput()
 			Error1("No input in file",AM.InputFileName);
 			return(-1);
 		}
+		}
 		AC.NoShowInput = 1;
 		while ( AM.SkipClears > 0 ) {
 			c = GetInput();
@@ -462,7 +632,7 @@ VOID ReserveTempFiles(int par)
 {
 	GETIDENTITY
 	SETUPPARAMETERS *sp;
-	UBYTE *s, *t, c;	
+	UBYTE *s, *t, *tenddir, *tenddir2, c;	
 	int i = 0;
 	WORD j;
 	if ( par == 0 || par == 1 ) {
@@ -474,6 +644,16 @@ VOID ReserveTempFiles(int par)
 		}
 		else AM.TempDir = (UBYTE *)(sp->value);
 	}
+	if ( AM.TempSortDir == 0 ) {
+		if ( AM.havesortdir ) {
+			sp = GetSetupPar((UBYTE *)"tempsortdir");
+			AM.TempSortDir = (UBYTE *)(sp->value);
+		}
+		else {
+			AM.TempSortDir = (UBYTE *)getenv("FORMTMPSORT");
+			if ( AM.TempSortDir == 0 ) AM.TempSortDir = AM.TempDir;
+		}
+	}
 /*
 	We have now in principle a path but we will use its first element only.
 	Later that should become more complicated. Then we will use a path and
@@ -482,11 +662,28 @@ VOID ReserveTempFiles(int par)
 	s = AM.TempDir; i = 200;   /* Some extra for VMS */
 	while ( *s && *s != ':' ) { if ( *s == '\\' ) s++; s++; i++; }
 
-	FG.fname = (char *)Malloc1(sizeof(UBYTE)*(i+10),"name for temporary files");
+	FG.fname = (char *)Malloc1(sizeof(UBYTE)*(i+14),"name for temporary files");
 	s = AM.TempDir; t = (UBYTE *)FG.fname;
 	while ( *s && *s != ':' ) { if ( *s == '\\' ) s++; *t++ = *s++; }
 	if ( (char *)t > FG.fname && t[-1] != SEPARATOR && t[-1] != ALTSEPARATOR )
 		*t++ = SEPARATOR;
+	*t = 0;
+	tenddir = t;
+	FG.fnamebase = t-(UBYTE *)(FG.fname);
+
+	s = AM.TempSortDir; i = 200;   /* Some extra for VMS */
+	while ( *s && *s != ':' ) { if ( *s == '\\' ) s++; s++; i++; }
+
+	FG.fname2 = (char *)Malloc1(sizeof(UBYTE)*(i+14),"name for sort files");
+	s = AM.TempSortDir; t = (UBYTE *)FG.fname2;
+	while ( *s && *s != ':' ) { if ( *s == '\\' ) s++; *t++ = *s++; }
+	if ( (char *)t > FG.fname2 && t[-1] != SEPARATOR && t[-1] != ALTSEPARATOR )
+		*t++ = SEPARATOR;
+	*t = 0;
+	tenddir2 = t;
+	FG.fname2base = t-(UBYTE *)(FG.fname2);
+
+	t = tenddir;
 	s = defaulttempfilename;
 #ifdef WITHMPI
 	{ 
@@ -502,6 +699,7 @@ VOID ReserveTempFiles(int par)
 		strcpy(FG.fname,"/formswap/xxxxxxxxxxxxxxxxxxxxx");
 		/*:[04nov2003 mt]*/
 		t = (UBYTE *)FG.fname + 10;
+		FG.fnamebase = t-FG.fname;
 	  }
 	  else{
 		/*[04nov2003 mt]:*/
@@ -511,6 +709,7 @@ VOID ReserveTempFiles(int par)
 		/*:[04nov2003 mt]*/
 		FG.fname[9] = '0' + PF.me;
 		t = (UBYTE *)FG.fname + 11;
+		FG.fnamebase = t-FG.fname;
 	  }
 #else
 	  iii = sprintf((char*)t,"%d",PF.me);
@@ -521,19 +720,17 @@ VOID ReserveTempFiles(int par)
 #endif
 	while ( *s ) *t++ = *s++;
 	*t = 0;
-#ifdef UNIX
 /*
 		There are problems when running many FORM jobs at the same time
 		from make or minos. If they start up simultaneously, occasionally
 		they can make the same .str file. We prevent this with first trying
-		a file that contains the last three digits of the pid. If this file
+		a file that contains the digits of the pid. If this file
 		has already been taken we fall back on the old scheme.
-		The whole is controled with the -M (MultiRun) parameter in the
+		The whole is controlled with the -M (MultiRun) parameter in the
 		command tail.
 */
 	if ( AM.MultiRun ) {
-		pid_t pp = getpid();
-		int num = ((int)pp)%100000;
+		int num = ((int)GetPID())%100000;
 		t += 2;
 		*t = 0;
 		t[-1] = 'r';
@@ -551,11 +748,8 @@ VOID ReserveTempFiles(int par)
 		}
 	}
 	else
-#endif
 	{
-#ifdef UNIX
 classic:;
-#endif
 	  for(;;) {
 		if ( ( AC.StoreHandle = OpenFile((char *)FG.fname) ) < 0 ) {
 			if ( ( AC.StoreHandle = CreateFile((char *)FG.fname) ) >= 0 ) break;
@@ -588,12 +782,19 @@ classic:;
 		else t[-5] = (UBYTE)(c+1);
 	  }
 	}
+/*
+	Now we should make sure that the tempsortdir cq tempsortfilename makes it
+	into a similar construction.
+*/
+	s = tenddir; t = tenddir2; while ( *s ) *t++ = *s++;
+	*t = 0;
+
 /*
 	Now we should asign a name to the main sort file and the two stage 4 files.
 */
-	AM.S0->file.name = (char *)Malloc1(sizeof(char)*i,"name for temporary files");
+	AM.S0->file.name = (char *)Malloc1(sizeof(char)*(i+14),"name for temporary files");
 	s = (UBYTE *)AM.S0->file.name;
-	t = (UBYTE *)FG.fname;
+	t = (UBYTE *)FG.fname2;
 	i = 1;
 	while ( *t ) { *s++ = *t++; i++; }
 	s[-2] = 'o'; *s = 0;
@@ -604,18 +805,22 @@ classic:;
 	are threads of course.
 */
 	if ( par == 0 ) {
-		s = (UBYTE *)Malloc1(sizeof(char)*i,"name for stage4 file a");
-		AR.FoStage4[0].name = (char *)s;
-		t = (UBYTE *)FG.fname;
+		s = (UBYTE *)((void *)(FG.fname2)); i = 0;
+		while ( *s ) { s++; i++; }
+		s = (UBYTE *)Malloc1(sizeof(char)*(i+1),"name for stage4 file a");
+		AR.FoStage4[1].name = (char *)s;
+		t = (UBYTE *)FG.fname2;
 		while ( *t ) *s++ = *t++;
 		s[-2] = '4'; s[-1] = 'a'; *s = 0;
-		s = (UBYTE *)Malloc1(sizeof(char)*i,"name for stage4 file b");
-		AR.FoStage4[1].name = (char *)s;
+		s = (UBYTE *)((void *)(FG.fname)); i = 0;
+		while ( *s ) { s++; i++; }
+		s = (UBYTE *)Malloc1(sizeof(char)*(i+1),"name for stage4 file b");
+		AR.FoStage4[0].name = (char *)s;
 		t = (UBYTE *)FG.fname;
 		while ( *t ) *s++ = *t++;
 		s[-2] = '4'; s[-1] = 'b'; *s = 0;
 		for ( j = 0; j < 3; j++ ) {
-			s = (UBYTE *)Malloc1(sizeof(char)*i,"name for scratch file");
+			s = (UBYTE *)Malloc1(sizeof(char)*(i+1),"name for scratch file");
 			AR.Fscr[j].name = (char *)s;
 			t = (UBYTE *)FG.fname;
 			while ( *t ) *s++ = *t++;
@@ -624,16 +829,18 @@ classic:;
 	}
 #ifdef WITHPTHREADS
 	else if ( par == 2 ) {
-		s = (UBYTE *)((void *)(FG.fname)); i = 0;
+		s = (UBYTE *)((void *)(FG.fname2)); i = 0;
 		while ( *s ) { s++; i++; }
 		s = (UBYTE *)Malloc1(sizeof(char)*(i+12),"name for stage4 file a");
-		sprintf((char *)s,"%s.%d",FG.fname,AT.identity);
+		sprintf((char *)s,"%s.%d",FG.fname2,AT.identity);
 		s[i-2] = '4'; s[i-1] = 'a';
-		AR.FoStage4[0].name = (char *)s;
+		AR.FoStage4[1].name = (char *)s;
+		s = (UBYTE *)((void *)(FG.fname)); i = 0;
+		while ( *s ) { s++; i++; }
 		s = (UBYTE *)Malloc1(sizeof(char)*(i+12),"name for stage4 file b");
 		sprintf((char *)s,"%s.%d",FG.fname,AT.identity);
 		s[i-2] = '4'; s[i-1] = 'b';
-		AR.FoStage4[1].name = (char *)s;
+		AR.FoStage4[0].name = (char *)s;
 		if ( AT.identity == 0 ) {
 			for ( j = 0; j < 3; j++ ) {
 				s = (UBYTE *)Malloc1(sizeof(char)*(i+1),"name for scratch file");
@@ -699,7 +906,8 @@ VOID StartVariables()
 	AC.exprfillwarning = 0;
 	AM.gLineLength = 79;
 	AM.OutBufSize = 80;
-	AM.MaxStreamSize = 1024;
+	AM.MaxStreamSize = MAXFILESTREAMSIZE;
+	AP.MaxPreAssignLevel = 4;
 	AC.iBufferSize = 512;
 	AP.pSize = 128;
 	AP.MaxPreIfLevel = 10;
@@ -713,7 +921,11 @@ VOID StartVariables()
 #endif
 	AC.tablefilling = 0;
 	AM.resetTimeOnClear = 1;
-
+	AM.gnumextrasym = AM.ggnumextrasym = 0;
+	AM.havesortdir = 0;
+	AM.SpectatorFiles = 0;
+	AM.NumSpectatorFiles = 0;
+	AM.SizeForSpectatorFiles = 0;
 /*
 	Information for the lists of variables. Part of error message and size:
 */
@@ -747,6 +959,8 @@ VOID StartVariables()
 	AC.DubiousList.size = sizeof(struct DuBiOuS);
 	AC.TableBaseList.message = "list of tablebases";
 	AC.TableBaseList.size = sizeof(DBASE);
+	AC.TestValue = 0;
+	AC.InnerTest = 0;
 
 	AC.AutoSymbolList.message = "autosymbol";
 	AC.AutoSymbolList.size = sizeof(struct SyMbOl);
@@ -804,6 +1018,9 @@ VOID StartVariables()
 #ifndef WITHPTHREADS
 	AT.ebufnum = inicbufs();		/* Buffer for extras during execution */
 	AT.fbufnum = inicbufs();		/* Buffer for caching in factorization */
+	AT.allbufnum = inicbufs();		/* Buffer for id,all */
+	AT.aebufnum = inicbufs();		/* Buffer for id,all */
+	AN.tryterm = 0;
 #else
 	AS.MasterSort = 0;
 #endif
@@ -813,10 +1030,14 @@ VOID StartVariables()
 	AM.zbufnum = inicbufs();		/* For very special values */
 	{
 		CBUF *C = cbuf+AM.zbufnum;
+		WORD one[5] = {4,1,1,3,0};
 		WORD zero = 0;
 		AddRHS(AM.zbufnum,1);
 		AM.zerorhs = C->numrhs;
-		AddNtoC(AM.zbufnum,1,&zero);
+		AddNtoC(AM.zbufnum,1,&zero,17);
+		AddRHS(AM.zbufnum,1);
+		AM.onerhs = C->numrhs;
+		AddNtoC(AM.zbufnum,5,one,17);
 	}
 	AP.inside.inscbuf = inicbufs();	/* For the #inside instruction */
 /*
@@ -849,7 +1070,7 @@ VOID StartVariables()
 	We do want to give them unique names though that the user cannot access.
 */
 	{
-		char dumstr[10];
+		char dumstr[20];
 		for ( ; i < FIRSTUSERSYMBOL; i++ ) {
 			sprintf(dumstr,":%d:",i);
 			AddSymbol((UBYTE *)dumstr,-MAXPOWER,MAXPOWER,VARTYPENONE,0);
@@ -860,20 +1081,23 @@ VOID StartVariables()
 	AddVector((UBYTE *)"parg_",VARTYPENONE,0);
 
 	AM.NumFixedFunctions = sizeof(fixedfunctions)/sizeof(struct fixedfun);
-	for ( i = 0; i < AM.NumFixedFunctions; i++ )
-		AddFunction((UBYTE *)fixedfunctions[i].name
-		                    ,fixedfunctions[i].commu
-		                    ,fixedfunctions[i].tensor
-		                    ,fixedfunctions[i].complx
-		                    ,fixedfunctions[i].symmetric
-							,0,-1,-1);
+	for ( i = 0; i < AM.NumFixedFunctions; i++ ) {
+		ii = AddFunction((UBYTE *)fixedfunctions[i].name
+		                         ,fixedfunctions[i].commu
+		                         ,fixedfunctions[i].tensor
+		                         ,fixedfunctions[i].complx
+		                         ,fixedfunctions[i].symmetric
+		                         ,0,-1,-1);
+		if ( fixedfunctions[i].tensor == GAMMAFUNCTION )
+							functions[ii].flags |= COULDCOMMUTE;
+	}
 /*
 	Next we add a number of dummy functions for ensuring that the user defined
 	functions start at a fixed given number FIRSTUSERFUNCTION.
 	We do want to give them unique names though that the user cannot access.
 */
 	{
-		char dumstr[10];
+		char dumstr[20];
 		for ( ; i < FIRSTUSERFUNCTION-FUNCTION; i++ ) {
 			sprintf(dumstr,"::%d::",i);
 			AddFunction((UBYTE *)dumstr,0,0,0,0,0,-1,-1);
@@ -918,6 +1142,23 @@ VOID StartVariables()
 	PutPreVar((UBYTE *)"optimminvar_",(UBYTE *)("0"),0,0);
 	PutPreVar((UBYTE *)"optimmaxvar_",(UBYTE *)("0"),0,0);
 	PutPreVar(AM.oldnumextrasymbols,(UBYTE *)("0"),0,0);
+	PutPreVar((UBYTE *)"optimvalue_",(UBYTE *)("0"),0,0);
+	PutPreVar((UBYTE *)"optimscheme_",(UBYTE *)("0"),0,0);
+	PutPreVar((UBYTE *)"tolower_",(UBYTE *)("0"),(UBYTE *)("?a"),0);
+	PutPreVar((UBYTE *)"toupper_",(UBYTE *)("0"),(UBYTE *)("?a"),0);
+	PutPreVar((UBYTE *)"SYSTEMERROR_",(UBYTE *)("0"),0,0);
+	{
+		char buf[41];  /* up to 128-bit */
+		LONG pid;
+#ifndef WITHMPI
+		pid = GetPID();
+#else
+		pid = ( PF.me == MASTER ) ? GetPID() : (LONG)0;
+		pid = PF_BroadcastNumber(pid);
+#endif
+		LongCopy(pid,buf);
+		PutPreVar((UBYTE *)"PID_",(UBYTE *)buf,0,0);
+	}
 	AM.atstartup = 0;
 	AP.MaxPreTypes = 10;
 	AP.NumPreTypes = 0;
@@ -956,7 +1197,11 @@ VOID StartVariables()
 	AC.ProcessStats = AM.gProcessStats = AM.ggProcessStats = 1;
 	AC.OldParallelStats = AM.gOldParallelStats = AM.ggOldParallelStats = 0;
 	AC.OldFactArgFlag = AM.gOldFactArgFlag = AM.ggOldFactArgFlag = NEWFACTARG;
+	AC.OldGCDflag = AM.gOldGCDflag = AM.ggOldGCDflag = 1;
+	AC.WTimeStatsFlag = AM.gWTimeStatsFlag = AM.ggWTimeStatsFlag = 0;
 	AM.gcNumDollars = AP.DollarList.num;
+	AC.SizeCommuteInSet = AM.gSizeCommuteInSet = 0;
+	AC.CommuteInSet = 0;
 
 	AM.PrintTotalSize = 0;
 
@@ -972,7 +1217,7 @@ VOID StartVariables()
 	AC.pfirstnum = 0;
 	AC.numpfirstnum = AC.sizepfirstnum = 0;
 #endif
-	AC.MemDebugFlag = 0;
+	AC.MemDebugFlag = 1;
 
 #ifdef WITHEXTERNALCHANNEL
 	AX.currentExternalChannel=0;
@@ -1019,65 +1264,11 @@ VOID StartMore()
 	PutPreVar((UBYTE *)"NPARALLELTASKS_",(UBYTE *)"1",0,0);
 #endif
 
-	PutPreVar((UBYTE *)"NAME_",AM.InputFileName,0,0);
+	PutPreVar((UBYTE *)"NAME_",AM.InputFileName ? AM.InputFileName : (UBYTE *)"STDIN",0,0);
 }
 
 /*
  		#] StartMore : 
- 		#[ PrintHeader :
-*/
-
-VOID PrintHeader()
-{
-#ifndef BETAVERSION
-/*
-	The header starting with the release of version 4.0
-*/
-#ifdef WITHMPI
-	if ( PF.me == MASTER && !AM.silent ) {
-#else
-	if ( !AM.silent ) {
-#endif
-		char buffer[100], *s = buffer;
-		sprintf(s,"%s %s (%s)",FORMNAME,VERSIONSTR,PRODUCTIONDATE);
-		while ( *s ) s++;
-		sprintf(s," %d-bits",(WORD)(sizeof(WORD)*16));
-		while ( *s ) s++;
-#if defined(WITHPTHREADS)
-		sprintf(s," %d worker",AM.totalnumberofthreads-1);
-		while ( *s ) s++;
-		if ( AM.totalnumberofthreads != 2 ) *s++ = 's';
-		else *s++ = ' ';
-#elif defined(WITHMPI)
-		sprintf(s," %d worker",PF.numtasks-1);
-		while ( *s ) s++;
-		if ( PF.numtasks != 2 ) *s++ = 's';
-		else *s++ = ' ';
-#else
-		sprintf(s,"           ");
-		while ( *s ) s++;
-#endif
-		while ( s-buffer < 48 ) *s++ = ' ';
-		sprintf(s," Run: %s",MakeDate());
-		MesPrint("%s",buffer);
-	}
-#else
-/*
-	The header till version 4.0 beta
-*/
-#ifdef WITHMPI
-	if ( PF.me == MASTER && !AM.silent ) {
-#else
-	if ( !AM.silent ) {
-#endif
-		MesPrint(FORMNAME" by J.Vermaseren %s(%s) Run at: %s",
-		         VERSIONSTR,PRODUCTIONDATE,MakeDate());
-	}
-#endif
-}
-
-/*
- 		#] PrintHeader : 
  		#[ IniVars :
 
 		This routine initializes the parameters that may change during the run.
@@ -1092,10 +1283,14 @@ WORD IniVars()
 #endif
 	WORD *fi, i, one = 1;
 	CBUF *C = cbuf+AC.cbufnum;
-	UBYTE buf[32];
 
+#ifdef WITHPTHREADS
+	UBYTE buf[32];
 	sprintf((char*)buf,"%d",AM.totalnumberofthreads);
-	PutPreVar((UBYTE *)"NTHREADS_",buf,0,0);
+	PutPreVar((UBYTE *)"NTHREADS_",buf,0,1);
+#else
+	PutPreVar((UBYTE *)"NTHREADS_",(UBYTE *)"1",0,1);
+#endif
 
 	AC.ShortStats = 0;
 	AC.WarnFlag = 1;
@@ -1110,7 +1305,11 @@ WORD IniVars()
 	AC.halfmod = 0; AC.nhalfmod = 0;
 	AC.modinverses = 0;
 	AC.lPolyFun = AM.gPolyFun = 0;
+	AC.lPolyFunInv = AM.gPolyFunInv = 0;
 	AC.lPolyFunType = AM.gPolyFunType = 0;
+	AC.lPolyFunExp = AM.gPolyFunExp = 0;
+	AC.lPolyFunVar = AM.gPolyFunVar = 0;
+	AC.lPolyFunPow = AM.gPolyFunPow = 0;
 	AC.DirtPow = 0;
 	AC.lDefDim = AM.gDefDim = 4;
 	AC.lDefDim4 = AM.gDefDim4 = 0;
@@ -1134,7 +1333,9 @@ WORD IniVars()
 	AC.OutNumberType = RATIONALMODE;
 	AM.gOutNumberType = RATIONALMODE;
 #ifdef WITHZLIB
-	AR.gzipCompress = 0;
+	AR.gzipCompress = GZIPDEFAULT;
+	AR.FoStage4[0].ziobuffer = 0;
+	AR.FoStage4[1].ziobuffer = 0;
 #endif
 	AR.BracketOn = 0;
 	AC.bracketindexflag = 0;
@@ -1264,10 +1465,13 @@ WORD IniVars()
 	AT.inprimelist = -1;
 	AT.sizeprimelist = 0;
 	AT.primelist = 0;
+	AT.LeaveNegative = 0;
+	AT.TrimPower = 0;
 	AN.SplitScratch = 0;
 	AN.SplitScratchSize = AN.InScratch = 0;
 	AN.SplitScratch1 = 0;
 	AN.SplitScratchSize1 = AN.InScratch1 = 0;
+	AN.idfunctionflag = 0;
 #endif
 	AO.OutputLine = AO.OutFill = BufferForOutput;
 	AO.FactorMode = 0;
@@ -1346,7 +1550,7 @@ VOID setSignalHandlers()
 	setNewSig(SIGHUP,onErrSig);
 	setNewSig(SIGALRM,onErrSig);
 	setNewSig(SIGVTALRM,onErrSig);
-	setNewSig(SIGPROF,onErrSig);
+/*	setNewSig(SIGPROF,onErrSig); */  /* Why did Tentukov forbid profilers?? */
 }
 
 #endif
@@ -1433,9 +1637,10 @@ int main(int argc, char **argv)
 	ReserveTempFiles(0);
 	IniFbuffer(AT.fbufnum);
 #endif
-	PrintHeader();
+	if ( !AM.FromStdin ) PrintHeader(1);
 	IniVars();
 	Globalize(1);
+	if ( AM.TimeLimit > 0 ) alarm(AM.TimeLimit);
 	TimeCPU(0);
 	TimeChildren(0);
 	TimeWallClock(0);
@@ -1460,7 +1665,7 @@ VOID CleanUp(WORD par)
 
 	if ( FG.fname ) {
 	CleanUpSort(0);
-	for ( i = 0; i < 2; i++ ) {
+	for ( i = 0; i < 3; i++ ) {
 		if ( AR.Fscr[i].handle >= 0 ) {
 			if ( AR.Fscr[i].name ) {
 /*
@@ -1491,18 +1696,19 @@ VOID CleanUp(WORD par)
 			}
 		}
 		CloseFile(AC.StoreHandle);
-		if ( par >= 0 || AR.StoreData.Handle < 0 ) {
+		if ( par >= 0 || AR.StoreData.Handle < 0 || AM.ClearStore ) {
 			remove(FG.fname);
 		}
 dontremove:;
 #else
 		CloseFile(AC.StoreHandle);
-		if ( par >= 0 || AR.StoreData.Handle < 0 ) {
+		if ( par >= 0 || AR.StoreData.Handle < 0 || AM.ClearStore > 0 ) {
 			remove(FG.fname);
 		}
 #endif
 	}
 	}
+	ClearSpectators(CLEARMODULE);
 /*
 	Remove recovery file on exit if everything went well
 */
@@ -1573,7 +1779,9 @@ VOID Terminate(int errorcode)
 	/*:[08may2006 mt]*/
 #endif
 #ifdef WITHPTHREADS
-	TerminateAllThreads();
+	if ( !WhoAmI() && !errorcode ) {
+		TerminateAllThreads();
+	}
 #endif
 	if ( AC.FinalStats ) {
 		if ( AM.PrintTotalSize ) {
@@ -1627,18 +1835,22 @@ VOID PrintRunningTime()
 		mastertime = AM.SumTime + TimeCPU(1);
 		wallclocktime = TimeWallClock(1);
 		totaltime = mastertime+workertime;
+		if ( !AM.silent ) {
 		MesPrint("  %l.%2i sec + %l.%2i sec: %l.%2i sec out of %l.%2i sec",
 			mastertime/1000,(WORD)((mastertime%1000)/10),
 			workertime/1000,(WORD)((workertime%1000)/10),
 			totaltime/1000,(WORD)((totaltime%1000)/10),
 			wallclocktime/100,(WORD)(wallclocktime%100));
+		}
 	}
 #else
 	LONG mastertime = AM.SumTime + TimeCPU(1);
 	LONG wallclocktime = TimeWallClock(1);
+	if ( !AM.silent ) {
 	MesPrint("  %l.%2i sec out of %l.%2i sec",
 		mastertime/1000,(WORD)((mastertime%1000)/10),
 		wallclocktime/100,(WORD)(wallclocktime%100));
+	}
 #endif
 }
 
@@ -1651,11 +1863,19 @@ LONG GetRunningTime()
 {
 #if defined(WITHPTHREADS) && (defined(WITHPOSIXCLOCK) || defined(WINDOWS))
 	LONG mastertime;
-	LONG workertime;
 	if ( AB[0] != 0 ) {
-		workertime = GetWorkerTimes();
+/* 
+#if ( defined(APPLE64) || defined(APPLE32) )
+		mastertime = AM.SumTime + TimeCPU(1);
+		return(mastertime);
+#else
+*/
+		LONG workertime = GetWorkerTimes();
 		mastertime = AM.SumTime + TimeCPU(1);
 		return(mastertime+workertime);
+/*
+#endif
+*/
 	}
 	else {
 		return(AM.SumTime + TimeCPU(1));
@@ -1676,4 +1896,3 @@ LONG GetRunningTime()
 /*
  		#] GetRunningTime : 
 */
-
