@@ -1,11 +1,11 @@
 /** @file tables.c
  * 
  *  Contains all functions that deal with the table bases on the 'FORM level'
- *  The low level databse routines are in minos.c
+ *  The low level database routines are in minos.c
  */
 /* #[ License : */
 /*
- *   Copyright (C) 1984-2013 J.A.M. Vermaseren
+ *   Copyright (C) 1984-2023 J.A.M. Vermaseren
  *   When using this file you are requested to refer to the publication
  *   J.A.M.Vermaseren "New features of FORM" math-ph/0010025
  *   This is considered a matter of courtesy as the development was paid
@@ -800,6 +800,8 @@ int CoTBaddto(UBYTE *s)
 			}
 			else if ( ExistsObject(d,basenumber,(char *)es) ) {}
 			else {
+			  int dict = AO.CurrentDictionary;
+			  AO.CurrentDictionary = 0;
 			  sum = i + T->numind;
 /*
 			  See also commentary below
@@ -824,6 +826,7 @@ int CoTBaddto(UBYTE *s)
 			  AO.OutInBuffer = 0;
 			  AddObject(d,basenumber,(char *)es,(char *)(AO.DollarOutBuffer));
 			  *s++ = c;
+			  AO.CurrentDictionary = dict;
 			}
 		}
 		else {
@@ -909,9 +912,12 @@ int CoTBenter(UBYTE *s)
 	int i, j, error = 0, error1 = 0, printall = 0;
 	TABLES T = 0;
 	WORD type, funnum;
+	int dict = AO.CurrentDictionary;
+	AO.CurrentDictionary = 0;
 	if ( ( d = FindTB(tablebasename) ) == 0 ) {
 		MesPrint("&No open tablebase with the name %s",tablebasename);
-		return(-1);
+		error = -1;
+		goto Endofall;
 	}
 	while ( *s == ',' || *s == ' ' || *s == '\t' ) s++;
 	if ( *s == '!' ) { printall = 1; s++; }
@@ -919,7 +925,7 @@ int CoTBenter(UBYTE *s)
 	if ( *s ) {
 	  while ( *s ) {
 		tablename = s;
-		if ( ( s = SkipAName(s) ) == 0 ) return(1);
+		if ( ( s = SkipAName(s) ) == 0 ) { error = 1; goto Endofall; }
 		c = *s; *s = 0;
 		if ( ( GetVar(tablename,&type,&funnum,CFUNCTION,NOAUTO) == NAMENOTFOUND )
 			|| ( T = functions[funnum].tabl ) == 0 ) {
@@ -970,7 +976,7 @@ int CoTBenter(UBYTE *s)
 
 						error1 = CoFill(buffer);
 
-						if ( error1 < 0 ) { return(error); }
+						if ( error1 < 0 ) goto Endofall;
 						if ( error1 != 0 ) error = error1;
 						M_free(buffer,"TableBase copy");
 					}
@@ -986,7 +992,8 @@ int CoTBenter(UBYTE *s)
 	  while ( *s ) {
 		basenumber++;
 		tablename = s; while ( *s ) s++; s++;
-		while ( *s ) s++; s++;
+		while ( *s ) s++;
+		s++;
 		if ( ( GetVar(tablename,&type,&funnum,CFUNCTION,NOAUTO) == NAMENOTFOUND )
 			|| ( T = functions[funnum].tabl ) == 0 ) {
 			MesPrint("&%s should be a previously declared table",tablename);
@@ -1031,7 +1038,7 @@ int CoTBenter(UBYTE *s)
 
 						error1 = CoFill(buffer);
 
-						if ( error1 < 0 ) { return(error); }
+						if ( error1 < 0 ) goto Endofall;
 						if ( error1 != 0 ) error = error1;
 						M_free(buffer,"TableBase copy");
 					}
@@ -1040,6 +1047,8 @@ int CoTBenter(UBYTE *s)
 		}
 	  }
 	}
+Endofall:;
+	AO.CurrentDictionary = dict;
 	return(error);
 }
 
@@ -1137,7 +1146,10 @@ int CheckTableDeclarations(DBASE *d)
 			 || ( T->sparse == 0 ) ) {
 			MesPrint("&%s has been declared previously, but not as a sparse table.",s);
 			error = 1;
-			while ( *s ) s++; s++; while ( *s ) s++; s++;
+			while ( *s ) s++;
+			s++;
+			while ( *s ) s++;
+			s++;
 		}
 		else {
 /*
@@ -1145,12 +1157,14 @@ int CheckTableDeclarations(DBASE *d)
 			We are not going to rename arguments when reading the elements.
 */
 			ss = s;
-			while ( *s ) s++; s++;
+			while ( *s ) s++;
+			s++;
 			if ( StrCmp(s,T->argtail) ) {
 				MesPrint("&Declaration of table %s in %s different from previous declaration",ss,d->name);
 				error = 1;
 			}
-			while ( *s ) s++; s++;
+			while ( *s ) s++;
+			s++;
 		}
 	}
 	if ( command ) { M_free(command,"table command"); }
@@ -1189,7 +1203,11 @@ int CoTBload(UBYTE *ss)
 		while ( *s ) {
 			num++;
 			if ( StrCmp(s,name) ) {
-				while ( *s ) s++; s++; while ( *s ) s++; s++; num++;
+				while ( *s ) s++;
+				s++; 
+				while ( *s ) s++;
+				s++;
+				num++;
 				continue;
 			}
 			name = s; while ( *s ) s++; ns = s-name; s++;
@@ -1202,7 +1220,8 @@ int CoTBload(UBYTE *ss)
 				for ( j = 0; j < NUMOBJECTS; j++ ) {
 					if ( d->iblocks[i]->objects[j].tablenumber == num ) {
 						t = arguments = (UBYTE *)(d->iblocks[i]->objects[j].element);
-						while ( *t ) t++; es = t - arguments;
+						while ( *t ) t++;
+						es = t - arguments;
 						cs = 2*es + 2*ns + ts + 10;
 						if ( cs > commandsize )	{
 							commandsize = 2*cs;
@@ -1262,7 +1281,8 @@ int CoTBload(UBYTE *ss)
 				for ( j = 0; j < NUMOBJECTS; j++ ) {
 					if ( d->iblocks[i]->objects[j].tablenumber == num ) {
 						t = arguments = (UBYTE *)(d->iblocks[i]->objects[j].element);
-						while ( *t ) t++; es = t - arguments;
+						while ( *t ) t++;
+						es = t - arguments;
 						cs = 2*es + 2*ns + ts + 10;
 						if ( cs > commandsize )	{
 							commandsize = 2*cs;
@@ -1613,8 +1633,11 @@ int CoTBuse(UBYTE *s)
 	  s = (UBYTE *)(d->tablenames); basenumber = 0;
 	  while ( *s ) {
 		basenumber++;
-		tablename = s; while ( *s ) s++; s++;
-		while ( *s ) s++; s++;
+		tablename = s;
+		while ( *s ) s++;
+		s++;
+		while ( *s ) s++;
+		s++;
 		if ( ( GetVar(tablename,&type,&funnum,CFUNCTION,NOAUTO) == NAMENOTFOUND )
 			|| ( T = functions[funnum].tabl ) == 0 ) {
 			MesPrint("&%s should be a previously declared table",tablename);
@@ -1934,7 +1957,7 @@ WORD Apply(WORD *term, WORD level)
 	rhs of table(indices,pattern). It does this up to maxtogo times
 	in the given term. It starts with the occurrences inside the
 	arguments of functions. If necessary it finishes at groundlevel.
-	An infite number of tries is indicates by maxtogo = 2^15-1 or 2^31-1.
+	An infinite number of tries is indicated by maxtogo = 2^15-1 or 2^31-1.
 	The occurrences are replaced by subexpressions. This allows TestSub
 	to finish the job properly.
 
