@@ -10,7 +10,7 @@
  */
 /* #[ License : */
 /*
- *   Copyright (C) 1984-2013 J.A.M. Vermaseren
+ *   Copyright (C) 1984-2023 J.A.M. Vermaseren
  *   When using this file you are requested to refer to the publication
  *   J.A.M.Vermaseren "New features of FORM" math-ph/0010025
  *   This is considered a matter of courtesy as the development was paid
@@ -35,10 +35,19 @@
  */
 /* #] License : */ 
 /*
-#define MALLOCDEBUG 1
-#define MALLOCDEBUGOUTPUT
   	#[ Includes :
+	Note: TERMMALLOCDEBUG tests part of the TermMalloc and NumberMalloc
+	      system. To work properly it needs MEMORYMACROS in declare.h
+	      not to be defined to make sure that all calls will be diverted
+	      to the routines here.
+#define TERMMALLOCDEBUG
+#define FILLVALUE 126
+#define MALLOCDEBUGOUTPUT
+#define MALLOCDEBUG 1
 */
+#ifndef FILLVALUE
+	#define FILLVALUE 0
+#endif
 
 /*
     The enhanced malloc debugger, see comments in the beginning of the
@@ -89,6 +98,10 @@ LONG numfrees = 0;
 #define M_alloc mprotectMalloc
     
 #endif
+ 
+#ifdef TERMMALLOCDEBUG
+WORD **DebugHeap1, **DebugHeap2;
+#endif
 
 /*
   	#] Includes : 
@@ -110,15 +123,23 @@ UBYTE *LoadInputFile(UBYTE *filename, int type)
 	filesize = BASEPOSITION(scrpos);
 	PUTZERO(scrpos);
 	SeekFile(handle,&scrpos,SEEK_SET);
-	buffer = (UBYTE *)Malloc1(filesize+1,"LoadInputFile");
+	buffer = (UBYTE *)Malloc1(filesize+2,"LoadInputFile");
 	if ( ReadFile(handle,buffer,filesize) != filesize ) {
 		Error1("Read error for file ",name);
 		M_free(buffer,"LoadInputFile");
+		if ( name != filename ) M_free(name,"FromLoadInputFile");
 		CloseFile(handle);
 		return(0);
 	}
 	CloseFile(handle);
-	buffer[filesize] = 0;
+	if ( type == PROCEDUREFILE || type == SETUPFILE ) {
+		buffer[filesize] = '\n';
+		buffer[filesize+1] = 0;
+	}
+	else {
+		buffer[filesize] = 0;
+	}
+	if ( name != filename ) M_free(name,"FromLoadInputFile");
 	return(buffer);
 }
 
@@ -205,6 +226,37 @@ UBYTE ReadFromStream(STREAM *stream)
 	}
 #endif /*ifdef WITHEXTERNALCHANNEL*/
 /*:[14apr2004 mt]*/
+	if ( stream->type == INPUTSTREAM ) {
+		if ( stream->pointer < stream->top ) {
+			c = *stream->pointer++;
+		}
+		else {
+			if ( ReadFile(stream->handle,&c,1) != 1 ) {
+				return(ENDOFSTREAM);
+			}
+			if ( stream->fileposition == 0 ) {
+				if ( !stream->buffer ) {
+					stream->buffersize = 32;
+					stream->buffer = (UBYTE *)Malloc1(stream->buffersize,"input stream buffer");
+					stream->pointer = stream->top = stream->buffer;
+				}
+				else {
+					if ( stream->top - stream->buffer >= stream->buffersize ) {
+						LONG oldsize = stream->buffersize;
+						DoubleBuffer((void**)&stream->buffer,(void**)&stream->top,sizeof(UBYTE),"double input stream buffer");
+						stream->buffersize = stream->top - stream->buffer;
+						stream->pointer = stream->top = stream->buffer + oldsize;
+					}
+				}
+				*stream->pointer = c;
+				stream->pointer = ++stream->top;
+				stream->inbuffer++;
+			}
+		}
+		if ( stream->eqnum == 1 ) { stream->eqnum = 0; stream->linenumber++; }
+		if ( c == LINEFEED ) stream->eqnum = 1;
+		return(c);
+	}
 	if ( stream->pointer >= stream->top ) {
 		if ( stream->type != FILESTREAM ) return(ENDOFSTREAM);
 		if ( stream->fileposition != stream->bufferposition+stream->inbuffer ) {
@@ -274,6 +326,7 @@ STREAM *OpenStream(UBYTE *name, int type, int prevarmode, int raiselow)
 	int handle, num;
 	LONG filesize;
 	switch ( type ) {
+		case REVERSEFILESTREAM:
 		case FILESTREAM:
 /*
 			Notice that FILESTREAM is only used for text files:
@@ -289,10 +342,20 @@ STREAM *OpenStream(UBYTE *name, int type, int prevarmode, int raiselow)
 			filesize = BASEPOSITION(scrpos);
 			PUTZERO(scrpos);
 			SeekFile(handle,&scrpos,SEEK_SET);
-			if ( filesize > AM.MaxStreamSize ) filesize = AM.MaxStreamSize;
+			if ( filesize > AM.MaxStreamSize && type == FILESTREAM )
+					filesize = AM.MaxStreamSize;
 			stream = CreateStream((UBYTE *)"filestream");
-			stream->buffer = (UBYTE *)Malloc1(filesize,"name of input stream");
+/*
+			The extra +1 in the Malloc1 is potentially needed in ReverseStatements!
+*/
+			stream->buffer = (UBYTE *)Malloc1(filesize+1,"name of input stream");
 			stream->inbuffer = ReadFile(handle,stream->buffer,filesize);
+			if ( type == REVERSEFILESTREAM ) {
+				if ( ReverseStatements(stream) ) {
+					M_free(stream->buffer,"name of input stream");
+					return(0);
+				}
+			}
 			stream->top = stream->buffer + stream->inbuffer;
 			stream->pointer = stream->buffer;
 			stream->handle = handle;
@@ -436,7 +499,7 @@ STREAM *OpenStream(UBYTE *name, int type, int prevarmode, int raiselow)
 			{/*Block*/
 				int n, *tmpn;
 				if( (n=getCurrentExternalChannel()) == 0 )
-					Error0("@No current extrenal channel");
+					Error0("@No current external channel");
 				stream = CreateStream((UBYTE *)"externalchannel");
 				stream->handle = CreateHandle();
 				tmpn = (int *)Malloc1(sizeof(int),"external channel handle");
@@ -453,6 +516,38 @@ STREAM *OpenStream(UBYTE *name, int type, int prevarmode, int raiselow)
 			break;
 #endif /*ifdef WITHEXTERNALCHANNEL*/
 /*:[14apr2004 mt]*/
+		case INPUTSTREAM:
+			/*
+			 * Assume that "name" stores a file descriptor (UNIX) or a FILE
+			 * pointer (Windows). We don't close it automatically on closing
+			 * the INPUTSTREAM stream (e.g., for stdin).
+			 */
+			stream = CreateStream((UBYTE *)"input stream");
+			stream->handle = CreateHandle();
+			{
+				FILES *f = (FILES *)Malloc1(sizeof(int),"input stream handle");
+				/* NOTE: in both cases name=0 indicates stdin. */
+#ifdef UNIX
+				f->descriptor = (int)(ssize_t)name;
+#else
+				f = name ? (FILES *)name : stdin;
+#endif
+				RWLOCKW(AM.handlelock);
+				filelist[stream->handle] = f;
+				UNRWLOCK(AM.handlelock);
+			}
+			stream->buffer = stream->pointer = stream->top = 0;
+			stream->inbuffer = 0;
+			stream->name = strDup1((UBYTE *)(name ? "INPUT" : "STDIN"),"input stream name");
+			stream->prevline = stream->linenumber = 1;
+			stream->eqnum = 0;
+			/*
+			 * fileposition == -1: default
+			 * fileposition ==  0: cache the input
+			 * See also: ReadFromStream, TryFileSetups
+			 */
+			stream->fileposition = -1;
+			break;
 		default:
 			return(0);
 	}
@@ -489,7 +584,7 @@ int LocateFile(UBYTE **name, int type)
 		if ( handle >= 0 ) return(handle);
 		MesPrint("Could not open setup file %s",(char *)(AM.SetupFile));
 	}
-	namesize = 2; s = *name;
+	namesize = 4; s = *name;
 	while ( *s ) { s++; namesize++; }
 	if ( type == SETUPFILE ) indir = AM.SetupDir;
 	else indir = AM.IncDir;
@@ -574,7 +669,7 @@ STREAM *CloseStream(STREAM *stream)
 		M_free(stream->FoldName,"stream->FoldName");
 		stream->FoldName = 0;
 	}
-	if ( stream->type == FILESTREAM ) {
+	if ( stream->type == FILESTREAM || stream->type == REVERSEFILESTREAM ) {
 		CloseFile(stream->handle);
 		if ( stream->buffer != 0 ) M_free(stream->buffer,"name of input stream");
 		stream->buffer = 0;
@@ -610,6 +705,15 @@ STREAM *CloseStream(STREAM *stream)
 	}
 #endif /*ifdef WITHEXTERNALCHANNEL*/
 /*:[14apr2004 mt]*/
+	else if ( stream->type == INPUTSTREAM ) {
+		FILES *f;
+		RWLOCKW(AM.handlelock);
+		f = filelist[stream->handle];
+		filelist[stream->handle] = 0;
+		numinfilelist--;
+		UNRWLOCK(AM.handlelock);
+		M_free(f,"input stream handle");
+	}
 	else if ( stream->type == PREVARSTREAM && (
 	stream->afterwards == PRERAISEAFTER || stream->afterwards == PRELOWERAFTER ) ) {
 		t = stream->buffer; x = 0; sgn = 1;
@@ -741,6 +845,109 @@ VOID PositionStream(STREAM *stream, LONG position)
 
 /*
  		#] PositionStream : 
+ 		#[ ReverseStatements :
+
+		Reverses the order of the statements in the buffer.
+		We allocate an extra buffer and copy a bit to and fro.
+		Note that there are some nasties that cannot be resolved.
+*/
+
+int ReverseStatements(STREAM *stream)
+{
+	UBYTE *spare = (UBYTE *)Malloc1((stream->inbuffer+1)*sizeof(UBYTE),"Reverse copy");
+	UBYTE *top = stream->buffer + stream->inbuffer, *in, *s, *ss, *out;
+	out = spare+stream->inbuffer+1;
+	in = stream->buffer;
+	while ( in < top ) {
+		s = in;
+		if ( *s == AP.ComChar ) {
+toeol:;
+			for(;;) {
+				if ( s == top ) { *--out = '\n'; break; }
+				if ( *s == '\\' ) {
+					s++;
+					if ( s >= top ) { /* This is an error! */
+irrend:					MesPrint("@Irregular end of reverse include file.");
+						return(1);
+					}
+				}
+				else if ( *s == '\n' ) {
+					s++; ss = s;
+					while ( ss > in ) *--out = *--ss;
+					in = s;
+					if ( out[0] == AP.ComChar && ss+6 < s && out[3] == '#' ) {
+/*
+						For folds we have to exchange begin and end
+*/
+						if ( out[4] == '[' ) out[4] = ']';
+						else if ( out[4] == ']' ) out[4] = '[';
+					}
+					break;
+				}
+				s++;
+			}
+			continue;
+		}
+		while ( s < top && ( *s == ' ' || *s == '\t' ) ) s++;
+		if ( *s == '#' ) {	/* preprocessor instruction */
+			goto toeol;		/* read to end of line */
+		}
+		if ( *s == '.' ) {	/* end-of-module instruction */
+			goto toeol;		/* read to end of line */
+		}
+/*
+		Here we have a regular statement. In principle we scan to ; and its \n
+		but there are special cases.
+		1: ; inside a string (in print "......;";)
+		2: multiple statements on one line.
+		3: ; + commentary after some blanks.
+		4: `var' can cause problems.....
+*/
+		while ( s < top ) {
+			if ( *s == ';' ) {
+				s++;
+				while ( s < top && ( *s == ' ' || *s == '\t' ) ) s++;
+				while ( s < top && *s == '\n' ) s++;
+				if ( s >= top && s[-1] != '\n' ) *s++ = '\n';
+				ss = s;
+				while ( ss > in ) *--out = *--ss;
+				in = s;
+				break;
+			}
+			else if ( *s == '"' ) {
+				s++;
+				while ( s < top ) {
+					if ( *s == '"' ) break;
+					if ( *s == '\\' ) { s++; }
+					s++;
+				}
+				if ( s >= top ) goto irrend;
+			}
+			else if ( *s == '\\' ) {
+				s++;
+				if ( s >= top ) goto irrend;
+			}
+			s++;
+		}
+		if ( in < top ) { /* Like blank lines at the end */
+			if ( s >= top && s[-1] != '\n' ) *s++ = '\n';
+			ss = s;
+			while ( ss > in ) *--out = *--ss;
+			in = s;
+		}
+	}
+	if ( out == spare ) stream->inbuffer++;
+	if ( out > spare+1 ) {
+		MesPrint("@Internal error in #reverseinclude instruction.");
+		return(1);
+	}
+	memcpy((void *)(stream->buffer),(void *)out,(size_t)(stream->inbuffer*sizeof(UBYTE)));
+	M_free(spare,"Reverse copy");
+	return(0);
+}
+
+/*
+ 		#] ReverseStatements : 
   	#] Streams : 
   	#[ Files :
  		#[ StartFiles :
@@ -1140,6 +1347,9 @@ LONG WriteFileToFile(int handle, UBYTE *buffer, LONG size)
 		if ( retval == 0 ) return(totalwritten);
 		totalwritten += retval;
 	}
+/*
+if ( handle == AC.LogHandle || handle == ERROROUT ) FlushFile(handle);
+*/
 	return(totalwritten);
 }
 #ifndef WITHMPI
@@ -1293,10 +1503,11 @@ VOID TruncateFile(int handle)
  		#[ GetChannel :
 
 		Checks whether we have this file already. If so, we return its
-		handle. If not, we open the file first and add it to the buffers.
+		handle. If not and mode == 0, we open the file first and add it
+		to the buffers.
 */
 
-int GetChannel(char *name)
+int GetChannel(char *name,int mode)
 {
 	CHANNEL *ch;
 	int i;
@@ -1305,6 +1516,11 @@ int GetChannel(char *name)
 		if ( channels[i].name == 0 ) continue;
 		if ( StrCmp((UBYTE *)name,(UBYTE *)(channels[i].name)) == 0 ) return(channels[i].handle);
 	}
+	if ( mode == 1 ) {
+		MesPrint("&File %s in print statement is not open",name);
+		MesPrint("   You should open it first with a #write or #append instruction");
+		return(-1);		
+	}
 	for ( i = 0; i < NumOutputChannels; i++ ) {
 		if ( channels[i].name == 0 ) break;
 	}
@@ -1523,6 +1739,20 @@ int StrICont(UBYTE *s1, UBYTE *s2)
 
 /*
  		#] StrICont : 
+ 		#[ CmpArray :
+*/
+
+int CmpArray(WORD *t1, WORD *t2, WORD n)
+{
+	int i,x;
+	for ( i = 0; i < n; i++ ) {
+		if ( ( x = (int)(t1[i]-t2[i]) ) != 0 ) return(x);
+	}
+	return(0);
+}
+
+/*
+ 		#] CmpArray : 
  		#[ ConWord :
 */
 
@@ -1630,6 +1860,40 @@ VOID WriteUnfinString(int type, UBYTE *str, int num)
 
 /*
  		#] WriteUnfinString : 
+ 		#[ AddToString :
+*/
+
+UBYTE *AddToString(UBYTE *outstring, UBYTE *extrastring, int par)
+{
+	UBYTE *s = extrastring, *t, *newstring;
+	int n, nn;
+	while ( *s ) { s++; }
+	n = s-extrastring;
+	if ( outstring == 0 ) {
+		s = extrastring;
+		t = outstring = (UBYTE *)Malloc1(n+1,"AddToString");
+		NCOPY(t,s,n)
+		*t++ = 0;
+		return(outstring);
+	}
+	else {
+		t = outstring;
+		while ( *t ) t++;
+		nn = t - outstring;
+		t = newstring = (UBYTE *)Malloc1(n+nn+2,"AddToString");
+		s = outstring;
+		NCOPY(t,s,nn)
+		if ( par == 1 ) *t++ = ',';
+		s = extrastring;
+		NCOPY(t,s,n)
+		*t = 0;
+		M_free(outstring,"AddToString");
+		return(newstring);
+	}
+}
+
+/*
+ 		#] AddToString : 
  		#[ strDup1 :
 
 		string duplication with message passing for Malloc1, allowing
@@ -1730,11 +1994,15 @@ WORD ReadSnum(UBYTE **p)
 
 */
 
-UBYTE *NumCopy(WORD x, UBYTE *to)
+UBYTE *NumCopy(WORD y, UBYTE *to)
 {
 	UBYTE *s;
 	WORD i = 0, j;
-	if ( x < 0 ) { x = -x; *to++ = '-'; }
+	UWORD x;
+	if ( y < 0 ) {
+		*to++ = '-';
+	}
+	x = WordAbs(y);
 	s = to;
 	do { *s++ = (UBYTE)((x % 10)+'0'); i++; } while ( ( x /= 10 ) != 0 );
 	*s-- = '\0';
@@ -1753,11 +2021,15 @@ UBYTE *NumCopy(WORD x, UBYTE *to)
 
 */
 
-char *LongCopy(LONG x, char *to)
+char *LongCopy(LONG y, char *to)
 {
 	char *s;
 	WORD i = 0, j;
-	if ( x < 0 ) { x = -x; *to++ = '-'; }
+	ULONG x;
+	if ( y < 0 ) {
+		*to++ = '-';
+	}
+	x = LongAbs(y);
 	s = to;
 	do { *s++ = (x % 10)+'0'; i++; } while ( ( x /= 10 ) != 0 );
 	*s-- = '\0';
@@ -1778,6 +2050,14 @@ char *LongCopy(LONG x, char *to)
 
 char *LongLongCopy(off_t *y, char *to)
 {
+	/*
+	 * This code fails to print the maximum negative value on systems with two's
+	 * complement. To fix this, we need the unsigned version of off_t with the
+	 * same size, but unfortunately it is undefined. On the other hand, if a
+	 * system is configured with a 64-bit off_t, in practice one never reaches
+	 * 2^63 ~ 10^18 as of 2016. If one really reach such a big number, then it
+	 * would be the time to move on a 128-bit off_t.
+	 */
 	off_t x = *y;
 	char *s;
 	WORD i = 0, j;
@@ -1935,7 +2215,7 @@ VOID iniTools(VOID)
 */
 #ifdef MALLOCDEBUG
 char *dummymessage = "Malloc";
-INILOCK(MallocLock);
+INILOCK(MallocLock)
 #endif
  
 VOID *Malloc(LONG size)
@@ -1987,7 +2267,7 @@ VOID *Malloc(LONG size)
 	}
 	t = (char *)mem;
 	u = t + size;
-	for ( i = 0; i < (int)BANNER; i++ ) { *t++ = 0; *--u = 0; }
+	for ( i = 0; i < (int)BANNER; i++ ) { *t++ = FILLVALUE; *--u = FILLVALUE; }
 	mem = (void *)t;
 	{
 		int j = nummalloclist-1, i;
@@ -1996,7 +2276,7 @@ VOID *Malloc(LONG size)
 			u = t + mallocsizes[j];
 			for ( i = 0; i < (int)BANNER; i++ ) {
 				u--;
-				if ( *t != 0 || *u != 0 ) {
+				if ( *t != FILLVALUE || *u != FILLVALUE ) {
 					MesPrint("Writing outside memory for %s",malloclist[i]);
 /*					MUNLOCK(ErrorMessageLock); */
 					UNLOCK(MallocLock);
@@ -2078,13 +2358,19 @@ VOID *Malloc1(LONG size, const char *messageifwrong)
 	
 	t = (char *)mem;
 	u = t + size;
-	for ( i = 0; i < (int)BANNER; i++ ) { *t++ = 0; *--u = 0; }
+	for ( i = 0; i < (int)BANNER; i++ ) { *t++ = FILLVALUE; *--u = FILLVALUE; }
 	mem = (void *)t;
 	M_check();
 /*	MUNLOCK(ErrorMessageLock); */
 	UNLOCK(MallocLock);
 #endif
-
+/* 
+	if ( size > 500000000L ) {
+		MLOCK(ErrorMessageLock);
+		MesPrint("Malloc1: %s, allocated %l bytes\n",messageifwrong,size);
+		MUNLOCK(ErrorMessageLock);
+	}
+*/
 	return(mem);
 }
 
@@ -2125,7 +2411,7 @@ void M_free(VOID *x, const char *where)
 	}
 	else {
 		for ( k = 0, j = 0; k < (int)BANNER; k++ ) {
-			if ( *--t ) j++;
+			if ( *--t != FILLVALUE ) j++;
 		}
 		if ( j ) {
 			LONG *tt = (LONG *)x;
@@ -2134,7 +2420,7 @@ void M_free(VOID *x, const char *where)
 		}
 		t += size;
 		for ( k = 0, j = 0; k < (int)BANNER; k++ ) {
-			if ( *--t ) j++;
+			if ( *--t != FILLVALUE ) j++;
 		}
 		if ( j ) {
 			LONG *tt = (LONG *)x;
@@ -2182,7 +2468,7 @@ void M_check()
 	for ( i = 0; i < nummalloclist; i++ ) {
 		t = (char *)(malloclist[i]);
 		for ( k = 0, j = 0; k < (int)BANNER; k++ ) {
-			if ( *t++ ) j++;
+			if ( *t++ != FILLVALUE ) j++;
 		}
 		if ( j ) {
 			tt = (LONG *)(malloclist[i]);
@@ -2193,7 +2479,7 @@ void M_check()
 		}
 		t = (char *)(malloclist[i]) + mallocsizes[i];
 		for ( k = 0, j = 0; k < (int)BANNER; k++ ) {
-			if ( *--t ) j++;
+			if ( *--t != FILLVALUE ) j++;
 		}
 		if ( j ) {
 			tt = (LONG *)t;
@@ -2272,11 +2558,20 @@ VOID TermMallocAddMemory(PHEAD0)
 	for ( i = 0; i < extra; i++ ) {
 		AT.TermMemHeap[i] = newbufs + i*(AM.MaxTer/sizeof(WORD)+TERMEXTRAWORDS);
 	}
+#ifdef TERMMALLOCDEBUG
+	DebugHeap2 = (WORD **)Malloc1((extra+AT.TermMemMax)*sizeof(WORD *),"TermMalloc");
+	for ( i = 0; i < AT.TermMemMax; i++ ) { DebugHeap2[i] = DebugHeap1[i]; }
+	for ( i = 0; i < extra; i++ ) {
+		DebugHeap2[i+AT.TermMemMax] = newbufs + i*(AM.MaxTer/sizeof(WORD)+TERMEXTRAWORDS);
+	}
+	if ( DebugHeap1 ) M_free(DebugHeap1,"TermMalloc");
+	DebugHeap1 = DebugHeap2;
+#endif
 	AT.TermMemTop = extra;
 	AT.TermMemMax += extra;
-/*
-MesPrint("AT.TermMemMax is now %l",AT.TermMemMax);
-*/
+#ifdef TERMMALLOCDEBUG
+	MesPrint("AT.TermMemMax is now %l",AT.TermMemMax);
+#endif
 }
 
 #ifndef MEMORYMACROS
@@ -2285,17 +2580,38 @@ WORD *TermMalloc2(PHEAD char *text)
 {
 	if ( AT.TermMemTop <= 0 ) TermMallocAddMemory(BHEAD0);
 
+#ifdef TERMMALLOCDEBUG
+	MesPrint("TermMalloc: %s, %d",text,(AT.TermMemMax-AT.TermMemTop));
+#endif
+
 #ifdef MALLOCDEBUGOUTPUT
 	MesPrint("TermMalloc: %s, %l/%l (%x)",text,AT.TermMemTop,AT.TermMemMax,AT.TermMemHeap[AT.TermMemTop-1]);
 #endif
-	
+
+	DUMMYUSE(text);
 	return(AT.TermMemHeap[--AT.TermMemTop]);
 }
  
 VOID TermFree2(PHEAD WORD *TermMem, char *text)
 {
+#ifdef TERMMALLOCDEBUG
+
+	int i;
+
+	for ( i = 0; i < AT.TermMemMax; i++ ) {
+		if ( TermMem == DebugHeap1[i] ) break;
+	}
+	if ( i >= AT.TermMemMax ) {
+		MesPrint(" ERROR: TermFree called with an address not given by TermMalloc.");
+		Terminate(-1);
+	}
+#endif
+	DUMMYUSE(text);
 	AT.TermMemHeap[AT.TermMemTop++] = TermMem;
 	
+#ifdef TERMMALLOCDEBUG
+	MesPrint("TermFree: %s, %d",text,(AT.TermMemMax-AT.TermMemTop));
+#endif
 #ifdef MALLOCDEBUGOUTPUT
 	MesPrint("TermFree: %s, %l/%l (%x)",text,AT.TermMemTop,AT.TermMemMax,TermMem);
 #endif
@@ -2330,6 +2646,10 @@ VOID TermFree2(PHEAD WORD *TermMem, char *text)
 #define NUMBERMEMSTARTNUM 16
 #define NUMBEREXTRAWORDS 10L
 
+#ifdef TERMMALLOCDEBUG
+UWORD **DebugHeap3, **DebugHeap4;
+#endif
+
 VOID NumberMallocAddMemory(PHEAD0)
 {
 	UWORD *newbufs;
@@ -2343,6 +2663,15 @@ VOID NumberMallocAddMemory(PHEAD0)
 	for ( i = 0; i < extra; i++ ) {
 		AT.NumberMemHeap[i] = newbufs + i*(LONG)(AM.MaxTal+NUMBEREXTRAWORDS);
 	}
+#ifdef TERMMALLOCDEBUG
+	DebugHeap4 = (UWORD **)Malloc1((extra+AT.NumberMemMax)*sizeof(WORD *),"NumberMalloc");
+	for ( i = 0; i < AT.NumberMemMax; i++ ) { DebugHeap4[i] = DebugHeap3[i]; }
+	for ( i = 0; i < extra; i++ ) {
+		DebugHeap4[i+AT.NumberMemMax] = newbufs + i*(LONG)(AM.MaxTal+NUMBEREXTRAWORDS);
+	}
+	if ( DebugHeap3 ) M_free(DebugHeap3,"NumberMalloc");
+	DebugHeap3 = DebugHeap4;
+#endif
 	AT.NumberMemTop = extra;
 	AT.NumberMemMax += extra;
 /*
@@ -2354,18 +2683,82 @@ MesPrint("AT.NumberMemMax is now %l",AT.NumberMemMax);
 
 UWORD *NumberMalloc2(PHEAD char *text)
 {
-	if ( AT.NumberMemTop <= 0 ) NumberMallocAddMemory(BHEAD0);
+	if ( AT.NumberMemTop <= 0 ) NumberMallocAddMemory(BHEAD text);
 
 #ifdef MALLOCDEBUGOUTPUT
+	if ( (AT.NumberMemMax-AT.NumberMemTop) > 10 )
 	MesPrint("NumberMalloc: %s, %l/%l (%x)",text,AT.NumberMemTop,AT.NumberMemMax,AT.NumberMemHeap[AT.NumberMemTop-1]);
 #endif
 
+	DUMMYUSE(text);
 	return(AT.NumberMemHeap[--AT.NumberMemTop]);
 }
  
 VOID NumberFree2(PHEAD UWORD *NumberMem, char *text)
 {
+#ifdef TERMMALLOCDEBUG
+	int i;
+	for ( i = 0; i < AT.NumberMemMax; i++ ) {
+		if ( NumberMem == DebugHeap3[i] ) break;
+	}
+	if ( i >= AT.NumberMemMax ) {
+		MesPrint(" ERROR: NumberFree called with an address not given by NumberMalloc.");
+		Terminate(-1);
+	}
+#endif
+	DUMMYUSE(text);
 	AT.NumberMemHeap[AT.NumberMemTop++] = NumberMem;
+
+#ifdef MALLOCDEBUGOUTPUT
+	if ( (AT.NumberMemMax-AT.NumberMemTop) > 10 )
+	MesPrint("NumberFree: %s, %l/%l (%x)",text,AT.NumberMemTop,AT.NumberMemMax,NumberMem);
+#endif
+}
+
+#endif
+
+/*
+ 		#] NumberMalloc : 
+ 		#[ CacheNumberMalloc :
+
+	Similar to NumberMalloc
+ */
+
+VOID CacheNumberMallocAddMemory(PHEAD0)
+{
+	UWORD *newbufs;
+	WORD extra;
+	int i;
+	if ( AT.CacheNumberMemMax == 0 ) extra = NUMBERMEMSTARTNUM;
+	else                             extra = AT.CacheNumberMemMax;
+	if ( AT.CacheNumberMemHeap ) M_free(AT.CacheNumberMemHeap,"NumberMalloc");
+	newbufs = (UWORD *)Malloc1(extra*(AM.MaxTal+NUMBEREXTRAWORDS)*sizeof(UWORD),"CacheNumberMalloc");
+	AT.CacheNumberMemHeap = (UWORD **)Malloc1((extra+AT.NumberMemMax)*sizeof(UWORD *),"CacheNumberMalloc");
+	for ( i = 0; i < extra; i++ ) {
+		AT.CacheNumberMemHeap[i] = newbufs + i*(LONG)(AM.MaxTal+NUMBEREXTRAWORDS);
+	}
+	AT.CacheNumberMemTop = extra;
+	AT.CacheNumberMemMax += extra;
+}
+
+#ifndef MEMORYMACROS
+
+UWORD *CacheNumberMalloc2(PHEAD char *text)
+{
+	if ( AT.CacheNumberMemTop <= 0 ) CacheNumberMallocAddMemory(BHEAD0);
+
+#ifdef MALLOCDEBUGOUTPUT
+	MesPrint("NumberMalloc: %s, %l/%l (%x)",text,AT.NumberMemTop,AT.NumberMemMax,AT.NumberMemHeap[AT.NumberMemTop-1]);
+#endif
+
+	DUMMYUSE(text);
+	return(AT.CacheNumberMemHeap[--AT.CacheNumberMemTop]);
+}
+ 
+VOID CacheNumberFree2(PHEAD UWORD *NumberMem, char *text)
+{
+	DUMMYUSE(text);
+	AT.CacheNumberMemHeap[AT.CacheNumberMemTop++] = NumberMem;
 	
 #ifdef MALLOCDEBUGOUTPUT
 	MesPrint("NumberFree: %s, %l/%l (%x)",text,AT.NumberMemTop,AT.NumberMemMax,NumberMem);
@@ -2375,7 +2768,7 @@ VOID NumberFree2(PHEAD UWORD *NumberMem, char *text)
 #endif
 
 /*
- 		#] NumberMalloc : 
+ 		#] CacheNumberMalloc : 
  		#[ FromList :
 
 	Returns the next object in a list.
@@ -2632,22 +3025,23 @@ void ExpandBuffer(void **buffer, LONG *oldsize, int type)
 LONG iexp(LONG x, int p)
 {
 	int sign;
-	LONG y;
+	ULONG y;
+	ULONG ux;
 	if ( x == 0 ) return(0);
 	if ( p == 0 ) return(1);
-	if ( x < 0 ) { sign = -1; x = -x; }
-	else sign = 1;
+	sign = x < 0 ? -1 : 1;
 	if ( sign < 0 && ( p & 1 ) == 0 ) sign = 1;
-	if ( x == 1 ) return(sign);
+	ux = LongAbs(x);
+	if ( ux == 1 ) return(sign);
 	if ( p < 0 ) return(0);
 	y = 1;
 	while ( p ) {
-		if ( ( p & 1 ) != 0 ) y *= x;
+		if ( ( p & 1 ) != 0 ) y *= ux;
 		p >>= 1;
-		x = x*x;
+		ux = ux*ux;
 	}
 	if ( sign < 0 ) y = -y;
-	return(y);
+	return ULongToLong(y);
 }
 
 /*
@@ -2663,10 +3057,11 @@ void ToGeneral(WORD *r, WORD *m, WORD par)
 {
 	WORD *mm = m, j, k;
 	if ( par ) m++;
-	else m += ARGHEAD + 1;
+	else { m[1] = 0; m += ARGHEAD + 1; }
 	j = -*r++;
 	k = 3;
-	if ( j >= FUNCTION ) { *m++ = j; *m++ = 2; }
+/*		JV: Bugfix 1-feb-2016. Old code assumed FUNHEAD to be 2 */
+	if ( j >= FUNCTION ) { *m++ = j; *m++ = FUNHEAD; FILLFUN(m) }
 	else {
 		switch ( j ) {
 			case SYMBOL: *m++ = j; *m++ = 4; *m++ = *r++; *m++ = 1; break;
@@ -2675,9 +3070,13 @@ void ToGeneral(WORD *r, WORD *m, WORD par)
 				else if ( *r == 0 ) { m--; }
 				else          { *m++ = -*r; *m++ = 1; *m++ = -3; }
 				goto MakeSize;
-			case MINVECTOR: k = -k;
+			case MINVECTOR:
+				k = -k;
+				/* fall through */
 			case INDEX:
-			case VECTOR: *m++ = INDEX; *m++ = 3; *m++ = *r++; break;
+			case VECTOR:
+				*m++ = INDEX; *m++ = 3; *m++ = *r++;
+				break;
 		}
 	}
 	*m++ = 1; *m++ = 1; *m++ = k;
@@ -2743,6 +3142,81 @@ int ToFast(WORD *r, WORD *m)
 
 /*
  		#] ToFast : 
+ 		#[ ToPolyFunGeneral :
+
+	Routine forces a polyratfun into general notation if needed.
+	If no action was needed, the return value is zero.
+	A positive return value indicates how many arguments were converted.
+	The new term overwrite the old.
+*/
+
+WORD ToPolyFunGeneral(PHEAD WORD *term)
+{
+	WORD *t = term+1, *tt, *to, *to1, *termout, *tstop, *tnext;
+	WORD numarg, i, change = 0;
+	tstop = term + *term; tstop -= ABS(tstop[-1]);
+	termout = to = AT.WorkPointer;
+	to++;
+	while ( t < tstop ) { /* go through the subterms */
+		if ( *t == AR.PolyFun ) {
+			tt = t+FUNHEAD; tnext = t + t[1];
+			numarg = 0;
+			while ( tt < tnext ) { numarg++; NEXTARG(tt); }
+			if ( numarg == 2 ) { /* this needs attention */
+				tt = t + FUNHEAD;
+				to1 = to;
+				i = FUNHEAD; NCOPY(to,t,i);
+				while ( tt < tnext ) { /* Do the arguments */
+					if ( *tt > 0 ) {
+						i = *tt; NCOPY(to,tt,i);
+					}
+					else if ( *tt == -SYMBOL ) {
+						to1[1] += 6+ARGHEAD; to1[2] |= MUSTCLEANPRF; change++;
+						*to++ = 8+ARGHEAD; *to++ = 0; FILLARG(to);
+						*to++ = 8; *to++ = SYMBOL; *to++ = 4; *to++ = tt[1];
+						*to++ = 1; *to++ = 1; *to++ = 1; *to++ = 3;
+						tt += 2;
+					}
+					else if ( *tt == -SNUMBER ) {
+						if ( tt[1] > 0 ) {
+							to1[1] += 2+ARGHEAD; to1[2] |= MUSTCLEANPRF; change++;
+							*to++ = 4+ARGHEAD; *to++ = 0; FILLARG(to);
+							*to++ = 4; *to++ = tt[1]; *to++ = 1; *to++ = 3;
+							tt += 2;
+						}
+						else if ( tt[1] < 0 ) {
+							to1[1] += 2+ARGHEAD; to1[2] |= MUSTCLEANPRF; change++;
+							*to++ = 4+ARGHEAD; *to++ = 0; FILLARG(to);
+							*to++ = 4; *to++ = -tt[1]; *to++ = 1; *to++ = -3;
+							tt += 2;
+						}
+						else {
+							MLOCK(ErrorMessageLock);
+							MesPrint("Internal error: Zero in PolyRatFun");
+							MUNLOCK(ErrorMessageLock);
+							Terminate(-1);
+						}
+					}
+				}
+				t = tnext;
+				continue;
+			}
+		}
+		i = t[1]; NCOPY(to,t,i)
+	}
+	if ( change ) {
+		tt = term + *term;
+        while ( t < tt ) *to++ = *t++;
+		*termout = to - termout;
+		t = term; i = *termout; tt = termout;
+		NCOPY(t,tt,i)
+		AT.WorkPointer = term + *term;
+	}
+	return(change);
+}
+
+/*
+ 		#] ToPolyFunGeneral : 
  		#[ IsLikeVector :
 
 		Routine determines whether a function argument is like a vector.
@@ -2951,7 +3425,7 @@ int CompArg(WORD *s1, WORD *s2)
 		s1 += ARGHEAD; s2 += ARGHEAD;
 docompare:
 		while ( s1 < st1 && s2 < st2 ) {
-			if ( ( k = CompareTerms(BHEAD s1,s2,(WORD)2) ) != 0 ) {
+			if ( ( k = CompareTerms(s1,s2,(WORD)2) ) != 0 ) {
 				AT.comsym[3] = x[1];
 				AT.comnum[1] = x[2];
 				AT.comnum[3] = x[3];
@@ -2983,12 +3457,60 @@ argerror:
  		#[ TimeWallClock :
 */
 
-#include <sys/timeb.h>
+#ifdef HAVE_CLOCK_GETTIME
+#include <time.h> /* for clock_gettime() */
+#else
+#ifdef HAVE_GETTIMEOFDAY
+#include <sys/time.h> /* for gettimeofday() */
+#else
+#include <sys/timeb.h> /* for ftime() */
+#endif
+#endif
 
+/**
+ * Returns the wall-clock time.
+ *
+ * @param   par  If zero, the wall-clock time will be reset to 0.
+ * @return       The wall-clock time in centiseconds.
+ */
 LONG TimeWallClock(WORD par)
 {
+	/*
+	 * NOTE: this function is not thread-safe. Operations on tp are not atomic.
+	 */
+
+#ifdef HAVE_CLOCK_GETTIME
+	struct timespec ts;
+	clock_gettime(CLOCK_MONOTONIC, &ts);
+
+	if ( par ) {
+		return(((LONG)(ts.tv_sec)-AM.OldSecTime)*100 +
+			((LONG)(ts.tv_nsec / 1000000)-AM.OldMilliTime)/10);
+	}
+	else {
+		AM.OldSecTime   = (LONG)(ts.tv_sec);
+		AM.OldMilliTime = (LONG)(ts.tv_nsec / 1000000);
+		return(0L);
+	}
+#else
+#ifdef HAVE_GETTIMEOFDAY
+	struct timeval t;
+	LONG sec, msec;
+	gettimeofday(&t, NULL);
+	sec = (LONG)t.tv_sec;
+	msec = (LONG)(t.tv_usec/1000);
+	if ( par ) {
+		return (sec-AM.OldSecTime)*100 + (msec-AM.OldMilliTime)/10;
+	}
+	else {
+		AM.OldSecTime   = sec;
+		AM.OldMilliTime = msec;
+		return(0L);
+	}
+#else
 	struct timeb tp;
 	ftime(&tp);
+
 	if ( par ) {
 		return(((LONG)(tp.time)-AM.OldSecTime)*100 + 
 			((LONG)(tp.millitm)-AM.OldMilliTime)/10);
@@ -2998,6 +3520,8 @@ LONG TimeWallClock(WORD par)
 		AM.OldMilliTime = (LONG)(tp.millitm);
 		return(0L);
 	}
+#endif
+#endif
 }
 
 /*
@@ -3017,6 +3541,12 @@ LONG TimeChildren(WORD par)
  		#[ TimeCPU :
 */
 
+/**
+ * Returns the CPU time.
+ *
+ * @param   par  If zero, the CPU time will be reset to 0.
+ * @return       The CPU time in milliseconds.
+ */
 LONG TimeCPU(WORD par)
 {
 	GETIDENTITY
