@@ -11,7 +11,7 @@
  */
 /* #[ License : */
 /*
- *   Copyright (C) 1984-2013 J.A.M. Vermaseren
+ *   Copyright (C) 1984-2023 J.A.M. Vermaseren
  *   When using this file you are requested to refer to the publication
  *   J.A.M.Vermaseren "New features of FORM" math-ph/0010025
  *   This is considered a matter of courtesy as the development was paid
@@ -1089,6 +1089,8 @@ DEBUG(MesPrint("Thread %w(b): s[3] = %d, w=(%d,%d,%d,%d)",s[3],w[0],w[1],w[2],w[
 								for ( j = 0; j < i; j++ ) {
 									if ( *t == s[2] && *s <= SYMTOSUB ) {
 										dirty = 1; v[2] |= DIRTYFLAG;
+										if ( AR.PolyFunType == 2 && v[0] == AR.PolyFun )
+											v[2] |= MUSTCLEANPRF;
 										if ( *s == SYMTOSYM ) *m = s[3];
 										else if ( *s == SYMTONUM ) {
 											m[-1] = -SNUMBER;
@@ -1240,6 +1242,8 @@ ss10:							*m++ = *t++;
 							*w = WORDDIF(m,w);	/* Fill parameter length */
 							if ( adirt ) {
 								dirty = w[1] = 1; v[2] |= DIRTYFLAG;
+								if ( AR.PolyFunType == 2 && v[0] == AR.PolyFun )
+									v[2] |= MUSTCLEANPRF;
 								AN.WildDirt = adirt;
 							}
 							else {
@@ -1537,7 +1541,7 @@ DEBUG(WORD *mm;)
 DEBUG(mm = m;)
 				n = *w - ARGHEAD;
 				w += ARGHEAD;
-				while ( (m + n + 10) > C->Top ) m = DoubleCbuffer(AT.ebufnum,m);
+				while ( (m + n + 10) > C->Top ) m = DoubleCbuffer(AT.ebufnum,m,4);
 				while ( --n >= 0 ) *m++ = *w++;
 				*m++ = 0;
 				C->rhs[C->numrhs+1] = m;
@@ -1567,7 +1571,7 @@ DEBUG(mm=m;)
 					w = AN.argaddress;
 					*m++ = 0;
 				}
-				while ( (m + n + 10) > C->Top ) m = DoubleCbuffer(AT.ebufnum,m);
+				while ( (m + n + 10) > C->Top ) m = DoubleCbuffer(AT.ebufnum,m,5);
 DEBUG(if ( mm != m-1 ) MesPrint("Thread %w(e): Alarm!"); mm = m-1;)
 				while ( --n >= 0 ) *m++ = *w++;
 				*m++ = 0;
@@ -1595,7 +1599,7 @@ DEBUG(mm=m;)
 					else n += 2;
 				}
 				*m++ = 0;
-				while ( (m + n + 10) > C->Top ) m = DoubleCbuffer(AT.ebufnum,m);
+				while ( (m + n + 10) > C->Top ) m = DoubleCbuffer(AT.ebufnum,m,6);
 DEBUG(if ( mm != m-1 ) MesPrint("Thread %w(f): Alarm!"); mm = m-1;)
 				a = (WORD **)(AN.argaddress);
 				while ( --k >= 0 ) {
@@ -1627,7 +1631,7 @@ DEBUG(MesPrint("Thread %w(f): w=(%d,%d,%d,%d)(%d)",mm[0],mm[1],mm[2],mm[3],C->nu
 				w = AN.argaddress;
 				n = *w - ARGHEAD;
 				w += ARGHEAD;
-				while ( (m + n + 10) > C->Top ) m = DoubleCbuffer(AT.ebufnum,m);
+				while ( (m + n + 10) > C->Top ) m = DoubleCbuffer(AT.ebufnum,m,7);
 				while ( --n >= 0 ) *m++ = *w++;
 				*m++ = 0;
 				C->rhs[C->numrhs+1] = m;
@@ -1636,7 +1640,7 @@ DEBUG(MesPrint("Thread %w(f): w=(%d,%d,%d,%d)(%d)",mm[0],mm[1],mm[2],mm[3],C->nu
 				w = AN.argaddress;
 				n = *w - ARGHEAD;
 				w += ARGHEAD;
-				while ( (m + n + 10) > C->Top ) m = DoubleCbuffer(AT.ebufnum,m);
+				while ( (m + n + 10) > C->Top ) m = DoubleCbuffer(AT.ebufnum,m,8);
 				sstop = w + n;
 				while ( w < sstop ) {			/* Run over terms */
 					tt = w + *w; ttstop = tt - ABS(tt[-1]);
@@ -1686,7 +1690,8 @@ DEBUG(MesPrint("Thread %w(f): w=(%d,%d,%d,%d)(%d)",mm[0],mm[1],mm[2],mm[3],C->nu
 	else {
 		do {
 			if ( w[2] == oldnumber && ( *w == type || ( type == VECTOVEC
-			&& ( *w == VECTOMIN || *w == VECTOSUB ) )
+			&& ( *w == VECTOMIN || *w == VECTOSUB ) ) || ( type == VECTOMIN
+			&& ( *w == VECTOVEC || *w == VECTOSUB ) )
 			|| ( type == INDTOIND && *w == INDTOSUB ) ) ) {
 				if ( n > 1 && w[4] == SETTONUM ) i = w[7];
 				*w = type;
@@ -1756,7 +1761,7 @@ FlipOn:
 WORD CheckWild(PHEAD WORD oldnumber, WORD type, WORD newnumber, WORD *newval)
 {
 	GETBIDENTITY
-	WORD *w, *m, *s, n, old2;
+	WORD *w, *m, *s, n, old2, inset;
 	WORD n2, oldval, dirty, i, j, notflag = 0, retblock = 0;
 	CBUF *C = cbuf+AT.ebufnum;
 	m = AT.WildMask;
@@ -2077,6 +2082,12 @@ WORD CheckWild(PHEAD WORD oldnumber, WORD type, WORD newnumber, WORD *newval)
 						old2 = *w;
 						goto TestSet;
 					}
+					else if ( type == VECTOMIN &&
+							( *w == VECTOSUB || *w == VECTOVEC ) ) {
+						if ( *m ) goto NoMatch;
+						old2 = *w;
+						goto TestSet;
+					}
 				}
 				m++; w += w[1];
 				if ( n > 1 && ( *w == FROMSET
@@ -2105,6 +2116,10 @@ TestSet:
 		WORD k;
 		s = w;							
 		j = w[2]; n2 = w[3];
+/*
+			if SETTONUM:  x?j[n2]
+			if FROMSET:   x?j?n2    or x?j and n2 = -WOLDOFFSET.
+*/
 		if ( j > WILDOFFSET ) {
 			j -= 2*WILDOFFSET;
 			notflag = 1;
@@ -2172,6 +2187,9 @@ TestSet:
 					if ( type != INDTOIND ||
 					newnumber < AM.IndDum || newnumber >= AM.IndDum+MAXDUMMIES ) goto NoMnot;
 					break;
+				case VECTOR_:
+					if ( type != VECTOVEC ) goto NoMnot;
+					break;
 				default:
 					goto NoMnot;
 			}
@@ -2202,10 +2220,39 @@ NoMnot:
 			}
 			goto NoMnot;
 		}
+/*
+		Now we have to determine which set element
+*/
 		w = SetElements + Sets[j].first;
 		m = SetElements + Sets[j].last;
+		if ( ( Sets[j].flags & ORDEREDSET ) == ORDEREDSET ) {
+/*
+			We search first and ask questions later
+*/
+			i = BinarySearch(w,Sets[j].last-Sets[j].first,newnumber);
+			if ( i < 0 ) {	/* no matter what, it is not in the set. */
+				goto NoMnot;
+			}
+			else {
+/*
+				We can set the proper parameters now to make only the
+				checks for the given set element.
+				After that we jump into the appropriate loop.
+*/
+				w = m = SetElements + i;
+				i++;
+				if ( Sets[j].type == -1 || Sets[j].type == CNUMBER ) {
+					goto insideloop1;
+				}
+				else {
+					goto insideloop2;
+				}
+			}
+		}
 		i = 1;
-		if ( Sets[j].type == -1 || Sets[j].type == CNUMBER ) { do {
+		if ( Sets[j].type == -1 || Sets[j].type == CNUMBER ) {
+		  do {
+			insideloop1:
 			if ( notflag ) {
 				switch ( type ) {
 					case SYMTOSYM:
@@ -2261,8 +2308,12 @@ NoMnot:
 				return(0);
 			}
 			i++;
-		} while ( ++w < m ); }
-		else { do {
+		  } while ( ++w < m );
+		}
+		else {
+		  do {
+			insideloop2:
+			inset = *w;
 			if ( notflag ) {
 			switch ( type ) {
 				case SYMTONUM:
@@ -2271,6 +2322,7 @@ NoMnot:
 					|| ( type == SYMTONUM && *w-2*MAXPOWER == newnumber ) ) {
 						goto NoMatch;
 					}
+					/* fall through */
 				case SYMTOSUB:
 					if ( *w < 0 ) {
 						WORD *mm = AT.WildMask, *mmm, *part;
@@ -2302,14 +2354,21 @@ NoMnot:
 						}
 					}
 					break;
+				case VECTOMIN:
+					if ( type == VECTOMIN ) {
+						if ( inset >= AM.OffsetVector ) { i++; continue; }
+						inset += WILDMASK;
+					}
+					/* fall through */
 				case VECTOVEC:
-					if ( *w == newnumber ) goto NoMatch;
+					if ( inset == newnumber ) goto NoMatch;
+					/* fall through */
 				case VECTOSUB:
-					if ( *w - WILDOFFSET >= AM.OffsetVector ) {
+					if ( inset - WILDOFFSET >= AM.OffsetVector ) {
 						WORD *mm = AT.WildMask, *mmm, *part;
 						WORD *ww = AN.WildValue;
 						WORD nn = AN.NumWild;
-						k = *w - WILDOFFSET;
+						k = inset - WILDOFFSET;
 						while ( --nn >= 0 ) {
 							if ( *mm && ww[2] == k && ww[0] == type ) {
 								if ( type == VECTOVEC ) {
@@ -2337,6 +2396,7 @@ NoMnot:
 					break;
 				case INDTOIND:
 					if ( *w == newnumber ) goto NoMatch;
+					/* fall through */
 				case INDTOSUB:
 					if ( *w - (WORD)WILDMASK >= AM.OffsetIndex ) {
 						WORD *mm = AT.WildMask, *mmm, *part;
@@ -2387,8 +2447,13 @@ NoMnot:
 					break;
 			}
 			}
-			else if ( ( *w == newnumber && type != SYMTONUM ) ||
-			( type == SYMTONUM && *w-2*MAXPOWER == newnumber ) ) {
+			else {
+			  if ( type == VECTOMIN ) {
+				if ( inset >= AM.OffsetVector ) { i++; continue; }
+				inset += WILDMASK;
+			  }
+			  if ( ( inset == newnumber && type != SYMTONUM ) ||
+			  ( type == SYMTONUM && inset-2*MAXPOWER == newnumber ) ) {
 				if ( *s == SETTONUM ) {
 					if ( n2 == oldnumber && type
 					<= SYMTOSUB ) goto NoMatch;
@@ -2422,9 +2487,11 @@ NoMnot:
 					}
 				}
 				return(0);
+			  }
 			}
 			i++;
-		} while ( ++w < m ); }
+		  } while ( ++w < m );
+		}
 		if ( notflag ) return(0);
 		AN.oldtype = old2; AN.oldvalue = oldval; goto NoMatch;
 	}
@@ -2435,7 +2502,7 @@ NoM:
 }
 
 /*
- 		#] CheckWild :
+ 		#] CheckWild : 
  	#] Wildcards :
   	#[ DenToFunction :
 
@@ -2479,4 +2546,3 @@ int DenToFunction(WORD *term, WORD numfun)
 /*
   	#] DenToFunction : 
 */
-
